/*
								+----------------------------------+
								|                                  |
								|  ***  Expression evaluator  ***  |
								|                                  |
								|   Copyright   -tHE SWINe- 2006  |
								|                                  |
								|         GPU_ExpEval.cpp          |
								|                                  |
								+----------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include "GPU_Lib.h"
#include "GPU_ExpEval.h"
#include "GPU_ExpFuncs.h"
#include "GPU_ExpParser.h"

#if defined(_MSC_VER) && !defined(__MWERKS__)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/*
 *								--- lexical analyzer ---
 */

enum {
	char_Space,	//  \t\n\r
	char_Alpha, // _a-z
	char_Digit, // 0-9
	char_DecPt, // .
	char_Oper,  // +-*/(),{}
	char_Id,	// %#$@
	char_TypeCount
};

enum {
	state_Error = -1,
	state_Init = 0,
	state_Final = 0x10000,
	state_Final2 = 0x10001
};


typedef struct {
	int n_next_state;
	const char *(*p_emit_func)(TToken *p_token, const char *p_s_buffer);
} TLexerState;

#define SF state_Final
#define SF2 state_Final2
#define SE state_Error

const TLexerState p_lexer_table[][char_TypeCount] = {
	{{0, NULL}, {1, NULL}, {2, NULL}, {2, NULL}, {SF2, p_s_Emit_Operator}, {4, NULL}}, // initial
	{{SF, p_s_Emit_Keyword}, {1, NULL}, {1, NULL}, {SF, p_s_Emit_Keyword}, {SF, p_s_Emit_Keyword}, {SF, p_s_Emit_Keyword}}, // keyword
	{{SF, p_s_Emit_Constant}, {SE, NULL}, {2, NULL}, {3, NULL}, {SF, p_s_Emit_Constant}, {SF, p_s_Emit_Constant}}, // number constant
	{{SF, p_s_Emit_Constant}, {SE, NULL}, {3, NULL}, {SE, NULL}, {SF, p_s_Emit_Constant}, {SF, p_s_Emit_Constant}}, // decimal number constant
	{{SE, NULL}, {SE, NULL}, {5, NULL}, {SE, NULL}, {SE, NULL}, {SE, NULL}}, // id
	{{SF, p_s_Emit_Ident}, {SF, p_s_Emit_Ident}, {5, NULL}, {SF, p_s_Emit_Ident}, {SF, p_s_Emit_Ident}, {SF, p_s_Emit_Ident}},
};
// lexicar analyzer transition table
// anything can be followed by anything, except % must be followed by number
// and number literal musn't be followed by alpha

const struct {
	const char *p_s_keyword;
	int n_token_type;
} p_keyword_list[] = {
	{"true", token_True},
	{"false", token_False},
	{"square", token_Square},
	{"diamond", token_Diamond},
	{"blend", token_Blend},
	{"decal", token_Decal},
	{"invert", token_Invert},
	{"abs", token_Abs},
	{"min", token_Min},
	{"max", token_Max},
	{"thresh", token_Thresh},
	{"log", token_Log},
	{"exp", token_Exp},
	{"sqrt", token_Sqrt},
	{"pow", token_Pow},
	{"scale", token_Scale},
	{"cmatrix", token_CMatrix},
	{"grey", token_Grey},
	{"resize", token_Resize},
	{"rotate", token_Rotate},
	{"lerp", token_Lerp},
	{"erode", token_Erode},
	{"dilate", token_Dilate},
	{"open", token_Open},
	{"close", token_Close},
	{"hitmiss", token_HitMiss},
	{"mean", token_Mean},
	{"median", token_Median},
	{"convolve", token_Convolve}
}, p_operator_list[] = {
	{"(", token_LeftPar},
	{")", token_RightPar},
	{"{", token_Begin},
	{"}", token_End},
	{",", token_Comma},
	{"+", token_Add},
	{"-", token_Sub},
	{"*", token_Mul},
	{"/", token_Div}
};
// keywords and operators

/**
 *	int n_ClassifyChar(char n_char)
 *		- return character group where n_char belongs
 */
int n_ClassifyChar(char n_char)
{
	if((n_char >= 'a' && n_char <= 'z') || n_char == '|') // _a-z
		return char_Alpha;
	if(n_char >= '0' && n_char <= '9') // 0-9
		return char_Digit;
	if(isspace(n_char) || !n_char) //  \t\n\r
		return char_Space;
	switch(n_char) {
	case '.':
		return char_DecPt;
	case '%':
	case '#':
	case '$':
	case '@':
		return char_Id;
	case '+':
	case '-':
	case '*':
	case '/':
	case '(':
	case ')':
	case '{':
	case '}':
	case ',':
		return char_Oper;
	}
	return -1;
}

/**
 *	char *p_s_GetToken(const char *p_s_expr, TToken *p_token, unsigned int *p_string_pos)
 *		- gets token from p_s_expr, starting at position p_string_pos
 *		- fills p_token and updates p_string_pos
 *		- returns NULL on success or error description string on failure
 */
char *p_s_GetToken(const char *p_s_expr, TToken *p_token, unsigned int *p_string_pos)
{
	p_s_expr += *p_string_pos;
	// move to the beginning of given string

	p_token->n_string_pos = *p_string_pos;

	if(!*p_s_expr) {
		p_token->n_type = token_EndOfLine;
		return NULL;
	}
	// we were given empty string

	char *p_s_buffer = NULL, *p_s_buffer_ptr;
	int n_buffer_size = 0, n_buffer_space = 0;
	// string buffer

	int n_state = state_Init;
	for(unsigned int n_length = strlen(p_s_expr) + 1; n_length;
	   ++ *p_string_pos, ++ p_s_expr, -- n_length) {
		if(n_buffer_space < 2) {
			if(n_buffer_size) {
				n_buffer_size *= 2;
				int n_off = p_s_buffer_ptr - p_s_buffer;
				p_s_buffer = (char*)realloc(p_s_buffer, n_buffer_size * sizeof(char));
				p_s_buffer_ptr = p_s_buffer + n_off;
				n_buffer_space = (p_s_buffer + n_buffer_size) - p_s_buffer_ptr;
			} else {
				p_s_buffer_ptr = p_s_buffer = (char*)malloc((n_buffer_space =
					n_buffer_size = 16) * sizeof(char));
			}
			if(p_s_buffer == NULL)
				return "not enough memory";
		}
		// make space in the buffer

		int n_char_class;
		if((n_char_class = n_ClassifyChar(*p_s_expr)) < 0)
			return "invalid character";
		// find out which character was that

		int n_next_state;
		switch(n_next_state = p_lexer_table[n_state][n_char_class].n_next_state) {
		case state_Init:
			p_token->n_string_pos = *p_string_pos + 1;
			break;
		case state_Error:
			return "unexpected character";
		case state_Final:
			{
				*p_s_buffer_ptr = 0;

				char *p_s_result = (char*)p_lexer_table[n_state][n_char_class].p_emit_func(p_token,
					p_s_buffer);
				free(p_s_buffer);
				return p_s_result;
			}
		case state_Final2:
			{
				++ *p_string_pos; // !!
				*p_s_buffer_ptr ++ = *p_s_expr;
				*p_s_buffer_ptr = 0;
				// accumulate current character as well!

				char *p_s_result = (char*)p_lexer_table[n_state][n_char_class].p_emit_func(p_token,
					p_s_buffer);
				free(p_s_buffer);
				return p_s_result;
			}
		default:
			-- n_buffer_space;
			*p_s_buffer_ptr ++ = *p_s_expr;
			// accumulate current character
			break;
		}
		n_state = n_next_state;
		// proceed
	}

	if(n_state == state_Init) {
		p_token->n_type = token_EndOfLine;
		return NULL;
	}
	// we were given empty string

	return "unexpected end";
}

/**
 *	const char *p_s_Emit_Ident(TToken *p_token, const char *p_s_buffer)
 *		- emits variable identifier token, writes it to p_token
 *		- p_s_buffer contains token string
 */
const char *p_s_Emit_Ident(TToken *p_token, const char *p_s_buffer)
{
	switch(*p_s_buffer) {
	case '%':
		p_token->n_type = token_Image;
		break;
	case '#':
		p_token->n_type = token_Matrix;
		break;
	case '$':
		p_token->n_type = token_Scalar;
		break;
	case '@':
		p_token->n_type = token_Vector;
		break;
	}
	p_token->t_data.n_id = atol(p_s_buffer + 1);
	return NULL;
}

/**
 *	const char *p_s_Emit_Constant(TToken *p_token, const char *p_s_buffer)
 *		- emits number literal token, writes it to p_token
 *		- p_s_buffer contains token string
 */
const char *p_s_Emit_Constant(TToken *p_token, const char *p_s_buffer)
{
	p_token->n_type = token_Number;
	p_token->t_data.f_value = (float)atof(p_s_buffer);
	return NULL;
}

/**
 *	const char *p_s_Emit_Operator(TToken *p_token, const char *p_s_buffer)
 *		- emits operator token, writes it to p_token
 *		- p_s_buffer contains token string
 */
const char *p_s_Emit_Operator(TToken *p_token, const char *p_s_buffer)
{
	for(int i = 0; i < sizeof(p_operator_list) / sizeof(p_operator_list[0]); ++ i) {
		if(!strcmp(p_s_buffer, p_operator_list[i].p_s_keyword)) {
			p_token->n_type = p_operator_list[i].n_token_type;
			return NULL;
		}
	}
	return "invalid operator";
}

/**
 *	const char *p_s_Emit_Keyword(TToken *p_token, const char *p_s_buffer)
 *		- emits keyword token, writes it to p_token
 *		- p_s_buffer contains token string
 */
const char *p_s_Emit_Keyword(TToken *p_token, const char *p_s_buffer)
{
	for(int i = 0; i < sizeof(p_keyword_list) / sizeof(p_keyword_list[0]); ++ i) {
		if(!strcmp(p_s_buffer, p_keyword_list[i].p_s_keyword)) {
			p_token->n_type = p_keyword_list[i].n_token_type;
			return NULL;
		}
	}
	return "invalid keyword";
}

/*
 *								--- ~lexical analyzer ---
 */

/*
 *								--- parser ---
 */

static int n_global_handle_num = 0, n_global_handle_list_size = 0;
static TExpressionHandle **p_handle_list;

/**
 *	TExpressionHandle *p_AllocHandle(const char *p_s_expression)
 *		- returns new handle named p_s_expression or NULL in case there was not enough memory
 */
TExpressionHandle *p_AllocHandle(const char *p_s_expression)
{
	if(n_global_handle_num == n_global_handle_list_size) {
		if(!n_global_handle_list_size) {
			p_handle_list = (TExpressionHandle**)malloc((n_global_handle_list_size = 16) *
				sizeof(TExpressionHandle*));
		} else {
			n_global_handle_list_size *= 2;
			p_handle_list = (TExpressionHandle**)realloc(p_handle_list,
				n_global_handle_list_size * sizeof(TExpressionHandle*));
		}
		if(p_handle_list == NULL) {
			n_global_handle_list_size = 0;
			return NULL;
		}
	}

	TExpressionHandle *p_tmp;
	if((p_tmp = (TExpressionHandle*)malloc(sizeof(TExpressionHandle))) == NULL)
		return NULL;
	p_tmp->n_reference_num = 1;
	p_tmp->n_temp_image_num = 0;
	p_tmp->n_temp_matrix_num = 0;
	p_tmp->n_temp_vector_num = 0;
	p_tmp->p_value = NULL;
	p_tmp->n_program_length = 0;
	if((p_tmp->p_s_expression = (char*)malloc((strlen(p_s_expression) + 1) *
	   sizeof(char))) == NULL) {
		free(p_tmp);
		return NULL;
	}
	strcpy(p_tmp->p_s_expression, p_s_expression);

	p_handle_list[n_global_handle_num] = p_tmp;
	++ n_global_handle_num;
	// add handle to global list

	return p_tmp;
}

/**
 *	TExpression *p_AllocExpression(int n_type, int n_return_type, int n_subexpr_num)
 *		- returns new expression of type n_type, with return value n_return_type and
 *		  n_subexpr_num subnode places or NULL in case there was not enough memory
 */
TExpression *p_AllocExpression(int n_type, int n_return_type, int n_subexpr_num)
{
	TExpression *p_tmp;
	if((p_tmp = (TExpression*)malloc(sizeof(TExpression))) == NULL)
		return NULL;
	if(n_subexpr_num) {
		if((p_tmp->p_sub_expr = (TExpression**)malloc(n_subexpr_num *
		   sizeof(TExpression*))) == NULL) {
			free(p_tmp);
			return NULL;
		}
		for(int i = 0; i < n_subexpr_num; ++ i)
			p_tmp->p_sub_expr[i] = NULL;
	} else
		p_tmp->p_sub_expr = NULL;
	p_tmp->n_type = n_type;
	p_tmp->n_return_type = n_return_type;
	p_tmp->n_sub_expr_num = n_subexpr_num;
	p_tmp->p_exec_func = NULL;
	return p_tmp;
}

/**
 *	void FreeExpression_r(TExpression *p_expr)
 *		- recursively deletes p_expr and it's subnodes
 */
void FreeExpression_r(TExpression *p_expr)
{
	if(p_expr->n_sub_expr_num) {
		for(int i = 0; i < p_expr->n_sub_expr_num; ++ i) {
			if(p_expr->p_sub_expr[i] != NULL)
				FreeExpression_r(p_expr->p_sub_expr[i]);
		}
		free(p_expr->p_sub_expr);
	}
	free(p_expr);
}

/**
 *	void FreeHandle(TExpressionHandle *p_handle)
 *		- "deletes" expression handle
 *		- decreases it's reference counter, but doesn't free anything until it reaches zero
 */
void FreeHandle(TExpressionHandle *p_handle)
{
	if(-- p_handle->n_reference_num > 0)
		return;

	for(int i = 0; i < n_global_handle_num; ++ i) {
		if(p_handle_list[i] == p_handle) {
			for(int j = -- n_global_handle_num - 1; j >= i; -- j)
				p_handle_list[j + 1] = p_handle_list[j];
			break;
		}
	}
	if(!n_global_handle_num) {
		free(p_handle_list);
		n_global_handle_list_size = 0;
	}
	// remove handle from the global list

	if(p_handle->n_temp_image_num) {
		for(int i = 0; i < p_handle->n_temp_image_num; ++ i) {
			if(p_handle->p_temp_image[i] != NULL)
				p_handle->p_temp_image[i]->Optional.Regular.Free(p_handle->p_temp_image[i]);
		}
		free(p_handle->p_temp_image);
	}
	if(p_handle->n_temp_vector_num) {
		for(int i = 0; i < p_handle->n_temp_vector_num; ++ i) {
			if(p_handle->p_temp_vector[i] != NULL)
				free(p_handle->p_temp_vector[i]);
		}
		free(p_handle->p_temp_vector);
	}
	if(p_handle->n_temp_matrix_num) {
		for(int i = 0; i < p_handle->n_temp_matrix_num; ++ i) {
			if(p_handle->p_temp_matrix[i] != NULL)
				free(p_handle->p_temp_matrix[i]);
		}
		free(p_handle->p_temp_matrix);
	}
	if(p_handle->p_value != NULL)
		FreeExpression_r(p_handle->p_value);
	if(p_handle->n_program_length)
		free(p_handle->p_program);
	if(p_handle->p_s_expression != NULL)
		free(p_handle->p_s_expression);
	free(p_handle);
}

/**
 *	ImageStruct *p_GetTempImage(TExpressionHandle *p_handle,
 *		int n_width, int n_height, int n_format)
 *		- allocates new temporary image in p_handle nad returns its pointer
 *		- returns NULL in case there was not enough memory
 */
ImageStruct *p_GetTempImage(TExpressionHandle *p_handle, int n_width, int n_height, int n_format)
{
	if(p_handle->n_temp_image_num) {
		p_handle->p_temp_image = (ImageStruct**)realloc(p_handle->p_temp_image,
			++ p_handle->n_temp_image_num * sizeof(ImageStruct*));
	} else {
		p_handle->p_temp_image = (ImageStruct**)malloc(sizeof(ImageStruct*));
		p_handle->n_temp_image_num = 1;
	}
	if(p_handle->p_temp_image == NULL)
		return NULL;
	return p_handle->p_temp_image[p_handle->n_temp_image_num - 1] =
		p_Alloc_Image(n_width, n_height, n_format);
}

/**
 *	TMatrix4x4 *p_GetTempMatrix(TExpressionHandle *p_handle)
 *		- allocates new temporary matrix in p_handle nad returns its pointer
 *		- returns NULL in case there was not enough memory
 */
TMatrix4x4 *p_GetTempMatrix(TExpressionHandle *p_handle)
{
	if(p_handle->n_temp_matrix_num) {
		p_handle->p_temp_matrix = (TMatrix4x4**)realloc(p_handle->p_temp_matrix,
			++ p_handle->n_temp_matrix_num * sizeof(TMatrix4x4*));
	} else {
		p_handle->p_temp_matrix = (TMatrix4x4**)malloc(sizeof(TMatrix4x4*));
		p_handle->n_temp_matrix_num = 1;
	}
	if(p_handle->p_temp_matrix == NULL)
		return NULL;
	return p_handle->p_temp_matrix[p_handle->n_temp_matrix_num - 1] =
		(TMatrix4x4*)malloc(sizeof(TMatrix4x4));
}

/**
 *	TVector4 *p_GetTempVector(TExpressionHandle *p_handle)
 *		- allocates new temporary vector in p_handle nad returns its pointer
 *		- returns NULL in case there was not enough memory
 */
TVector4 *p_GetTempVector(TExpressionHandle *p_handle)
{
	if(p_handle->n_temp_vector_num) {
		p_handle->p_temp_vector = (TVector4**)realloc(p_handle->p_temp_vector,
			++ p_handle->n_temp_vector_num * sizeof(TVector4*));
	} else {
		p_handle->p_temp_vector = (TVector4**)malloc(sizeof(TVector4*));
		p_handle->n_temp_vector_num = 1;
	}
	if(p_handle->p_temp_vector == NULL)
		return NULL;
	return p_handle->p_temp_vector[p_handle->n_temp_vector_num - 1] =
		(TVector4*)malloc(sizeof(TVector4));
}

inline int max(int a, int b)
{
	return (a > b)? a : b;
}

static const TFunctionBinding p_binding_table[] = {
	{token_Scalar, token_Scalar, 0, {0}, CopyVarFloat},
	{token_Vector, token_Vector, 0, {0}, CopyVarPtr},
	{token_Matrix, token_Matrix, 0, {0}, CopyVarPtr},
	{token_Image, token_Image, 0, {0}, CopyImgPtr},	// just copy value pointer from p_params
	{token_Number, token_Scalar, 0, {0}, NULL},
	{token_Number, token_Vector, 0, {0}, NULL},
	{token_Number, token_Matrix, 0, {0}, NULL}, // data is ready
	{token_Invert, token_Image, 1, {token_Image}, CallInverti},
	{token_Abs, token_Image, 1, {token_Image}, CallAbsi},
	{token_Log, token_Image, 1, {token_Image}, CallLogi},
	{token_Exp, token_Image, 1, {token_Image}, CallExpi},
	{token_Sqrt, token_Image, 1, {token_Image}, CallSqrti},
	{token_Grey, token_Image, 1, {token_Image}, CallGreyi},
	{token_Blend, token_Image, 4, {token_Image, token_Image, token_Scalar, token_Scalar}, CallBlendiinn},
	{token_Min, token_Vector, 1, {token_Image}, CallMini},
	{token_Min, token_Image, 2, {token_Image, token_Image}, CallMinii},
	{token_Min, token_Image, 2, {token_Image, token_Scalar, token_Scalar}, CallMininn},
	{token_Max, token_Vector, 1, {token_Image}, CallMaxi},
	{token_Max, token_Image, 2, {token_Image, token_Image}, CallMaxii},
	{token_Max, token_Image, 2, {token_Image, token_Scalar, token_Scalar}, CallMaxinn},
	{token_Thresh, token_Image, 2, {token_Image, token_Scalar}, CallThreshin},
	{token_Pow, token_Image, 2, {token_Image, token_Scalar}, CallPowin},
	{token_Median, token_Image, 2, {token_Image, token_Scalar}, CallMedianin},
	{token_Scale, token_Image, 3, {token_Image, token_Scalar, token_Scalar}, CallScaleinn},
	{token_Mean, token_Image, 3, {token_Image, token_Scalar, token_Scalar}, CallMeaninn},
	{token_CMatrix, token_Image, 2, {token_Image, token_Matrix}, CallCmatrixim},
	{token_Decal, token_Image, 2, {token_Image, token_Image}, CallDecalii},
	{token_HitMiss, token_Image, 2, {token_Image, token_Image}, CallHitMisii},
	{token_Erode, token_Image, 2, {token_Image, token_Image}, CallErodeii},
	{token_Dilate, token_Image, 2, {token_Image, token_Image}, CallDilateii},
	{token_Open, token_Image, 2, {token_Image, token_Image}, CallOpenii},
	{token_Close, token_Image, 2, {token_Image, token_Image}, CallCloseii},
	{token_Erode, token_Image, 4, {token_Image, token_Square, token_Scalar, token_Scalar}, CallErodeiSquarenn},
	{token_Dilate, token_Image, 4, {token_Image, token_Square, token_Scalar, token_Scalar}, CallDilateiSquarenn},
	{token_Open, token_Image, 4, {token_Image, token_Square, token_Scalar, token_Scalar}, CallOpeniSquarenn},
	{token_Close, token_Image, 4, {token_Image, token_Square, token_Scalar, token_Scalar}, CallCloseiSquarenn},
	{token_Erode, token_Image, 4, {token_Image, token_Diamond, token_Scalar, token_Scalar}, CallErodeiDiamondnn},
	{token_Dilate, token_Image, 4, {token_Image, token_Diamond, token_Scalar, token_Scalar}, CallDilateiDiamondnn},
	{token_Open, token_Image, 4, {token_Image, token_Diamond, token_Scalar, token_Scalar}, CallOpeniDiamondnn},
	{token_Close, token_Image, 4, {token_Image, token_Diamond, token_Scalar, token_Scalar}, CallCloseiDiamondnn},
	{token_Convolve, token_Image, 2, {token_Image, token_Image}, CallConvolveii},
	{token_Convolve, token_Image, 3, {token_Image, token_Image, token_Image}, CallConvolveiii},
	{token_Resize, token_Image, 1, {token_Image}, CallResizei},
	{token_Resize, token_Image, 3, {token_Image, token_Scalar, token_Scalar}, CallResizeinn},
	{token_Lerp, token_Image, 9, {token_Image, token_Scalar, token_Scalar, token_Scalar,
		token_Scalar, token_Scalar, token_Scalar, token_Scalar, token_Scalar}, CallLerpi8n},
	{token_Lerp, token_Image, 11, {token_Image, token_Scalar, token_Scalar, token_Scalar,
		token_Scalar, token_Scalar, token_Scalar, token_Scalar, token_Scalar, token_Scalar,
		token_Scalar}, CallLerpinn8n},
	{token_Rotate, token_Image, 4, {token_Image, token_Scalar, token_True, token_Scalar}, CallRotateinbn},
	{token_Rotate, token_Image, 4, {token_Image, token_Scalar, token_False, token_Scalar}, CallRotateinbn},
	{token_Rotate, token_Image, 6, {token_Image, token_Scalar, token_Scalar, token_True, token_Scalar}, CallRotateinnnbn},
	{token_Rotate, token_Image, 6, {token_Image, token_Scalar, token_Scalar, token_False, token_Scalar}, CallRotateinnnbn},
	{token_Add, token_Image, 2, {token_Image, token_Scalar}, CallAddin},
	{token_Add, token_Image, 2, {token_Image, token_Vector}, CallAddiv},
	{token_Add, token_Image, 2, {token_Image, token_Image}, CallAddii},
	{token_Sub, token_Image, 2, {token_Image, token_Scalar}, CallSubin},
	{token_Sub, token_Image, 2, {token_Image, token_Vector}, CallSubiv},
	{token_Sub, token_Image, 2, {token_Image, token_Image}, CallSubii},
	{token_Mul, token_Image, 2, {token_Image, token_Scalar}, CallMulin},
	{token_Mul, token_Image, 2, {token_Image, token_Vector}, CallMuliv},
	{token_Mul, token_Image, 2, {token_Image, token_Matrix}, CallCmatrixim},
	{token_Mul, token_Image, 2, {token_Image, token_Image}, CallMulii},
	{token_Div, token_Image, 2, {token_Image, token_Scalar}, CallDivin},
	{token_Div, token_Image, 2, {token_Image, token_Vector}, CallDiviv},
	{token_Div, token_Image, 2, {token_Image, token_Image}, CallDivii}
};
// function binding table

static TToken t_token;
// current token for parser

/**
 *	TExpression *p_Parse_L1_r(const char *p_s_expr, unsigned int *p_expr_pos,
 *		TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
 *		unsigned int *p_error_pos)
 *		- recursive descent parsing routine, level 1
 *		- returns expression tree or NULL in case errors occured
 *		- parses p_s_expr at p_expr_pos which is updated. temporary objects are allocated
 *		  inside p_handle, first call parameters p_params are used to determine temporary
 *		  image formats
 *		- p_error_string_ptr and p_error_pos are error string pointer and error position pointer.
 *		  filled when error occurs.
 */
TExpression *p_Parse_L1_r(const char *p_s_expr, unsigned int *p_expr_pos,
	TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
	unsigned int *p_error_pos);

/**
 *	bool FindExecFunction(TExpression *p_expression, int n_param_num,
 *		TExpressionHandle *p_handle, char **p_error_string_ptr)
 *		- finds execution function for p_expression with n_param_num parameters
 *		  (function is always allocated with number of parameters of it's largest overload
 *		  so some might be unused)
 *		- returns true on success, false on failure
 */
bool FindExecFunction(TExpression *p_expression, int n_param_num,
	TExpressionHandle *p_handle, char **p_error_string_ptr)
{
	for(unsigned int i = 0; i < sizeof(p_binding_table) / sizeof(p_binding_table[0]); ++ i) {
		if(p_expression->n_type != p_binding_table[i].n_function_token)
			continue;
		// see if it's the same function

		if(n_param_num != p_binding_table[i].n_param_num)
			continue;
		bool b_equal = true;
		for(int j = 0; j < n_param_num; ++ j) {
			if(p_binding_table[i].p_param_type[j] !=
			   p_expression->p_sub_expr[j]->n_return_type) {
				b_equal = false;
				break;
			}
		}
		if(!b_equal)
			continue;
		// see if param count and types match

		p_expression->n_return_type = p_binding_table[i].n_return_type;
		p_expression->p_exec_func = p_binding_table[i].p_exec_func;
		// set return type and function pointer

		if(p_expression->n_return_type == token_Vector) {
			if((p_expression->t_result.p_vector = p_GetTempVector(p_handle)) == NULL) {
				*p_error_string_ptr = "not enough memory";
				break;
			}
		} else if(p_expression->n_return_type == token_Matrix) {
			if((p_expression->t_result.p_matrix = p_GetTempMatrix(p_handle)) == NULL) {
				*p_error_string_ptr = "not enough memory";
				break;
			}
		} else if(p_expression->n_return_type == token_Image) {
			int n_result_width, n_result_height;
			ImageStruct *p_src_image = NULL;
			for(int j = 0; j < n_param_num; ++ j) {
				if(p_expression->p_sub_expr[j]->n_return_type == token_Image) {
					p_src_image = p_expression->p_sub_expr[j]->t_result.p_image;
					break;
				}
			}
			if(p_src_image == NULL) {
				*p_error_string_ptr = "cannot determine result pixel format";
				break;
			}
			if((p_expression->n_type == token_Resize && n_param_num == 3) ||
			   (p_expression->n_type == token_Rotate && n_param_num == 6) ||
			   (p_expression->n_type == token_Lerp && n_param_num == 11)) {
				n_result_width = (int)p_expression->p_sub_expr[1]->t_result.f_scalar;
				n_result_height = (int)p_expression->p_sub_expr[2]->t_result.f_scalar;
			} else {
				n_result_width = p_src_image->XSize;
				n_result_height = p_src_image->YSize;
			}
			if((p_expression->t_result.p_image = p_GetTempImage(p_handle,
			   n_result_width, n_result_height, p_src_image->PixelType)) == NULL) {
				*p_error_string_ptr = "not enough memory";
				break;
			}
		}
		// alloc intermediate (here are fuction nodes only,
		// variables process elsewhere)

		return true;
	}
	// find proper exec function, determine return value, alloc intermediates

	*p_error_string_ptr = "overloaded function not found";

	return false;
}

/**
 *	TExpression *p_Parse_L3_r(const char *p_s_expr, unsigned int *p_expr_pos,
 *		TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
 *		unsigned int *p_error_pos)
 *		- recursive descent parsing routine, level 3
 *		- returns expression tree or NULL in case errors occured
 *		- parses p_s_expr at p_expr_pos which is updated. temporary objects are allocated
 *		  inside p_handle, first call parameters p_params are used to determine temporary
 *		  image formats
 *		- p_error_string_ptr and p_error_pos are error string pointer and error position pointer.
 *		  filled when error occurs.
 */
TExpression *p_Parse_L3_r(const char *p_s_expr, unsigned int *p_expr_pos,
	TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
	unsigned int *p_error_pos)
{
	switch(t_token.n_type) {
	case token_Scalar:
	case token_Vector:
	case token_Matrix:
	case token_Image:
		{
			TExpression *p_result;
			if((p_result = p_AllocExpression(t_token.n_type, t_token.n_type, 0)) == NULL) {
				*p_error_string_ptr = "not enough memory";
				*p_error_pos = t_token.n_string_pos;
				return NULL;
			}
			p_result->n_variable_id = t_token.t_data.n_id;
			p_result->t_result.p_void_ptr = p_params[p_result->n_variable_id];
			// variable

			if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
				FreeExpression_r(p_result);
				return NULL;
			}
			// get next token

			p_result->p_exec_func = (p_result->n_type == token_Scalar)? CopyVarFloat :
				((p_result->n_type == token_Image)? CopyImgPtr : CopyVarPtr);

			return p_result;
		}
	case token_Number:
		{
			TExpression *p_result;
			if((p_result = p_AllocExpression(token_Number, token_Scalar, 0)) == NULL) {
				*p_error_string_ptr = "not enough memory";
				*p_error_pos = t_token.n_string_pos;
				return NULL;
			}
			p_result->n_variable_id = -1;
			p_result->t_result.f_scalar = t_token.t_data.f_value;
			// variable

			if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
				FreeExpression_r(p_result);
				return NULL;
			}
			// get next token

			p_result->p_exec_func = NULL;

			return p_result;
		}
	case token_LeftPar:
		{
			if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos))
				return NULL;
			// skip (

			TExpression *p_result;
			if((p_result = p_Parse_L1_r(p_s_expr, p_expr_pos,
			   p_handle, p_params, p_error_string_ptr, p_error_pos)) == NULL)
				return NULL;
			// parse result

			if(t_token.n_type != token_RightPar) {
				*p_error_string_ptr = "missing: \')\'";
				*p_error_pos = t_token.n_string_pos;
				FreeExpression_r(p_result);
				return NULL;
			}
			if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
				FreeExpression_r(p_result);
				return NULL;
			}
			// skip )

			return p_result;
		}
	case token_RightPar:
		*p_error_string_ptr = "unexpected token: \')\'";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	case token_Begin:
		{
			if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos))
				return NULL;
			// skip {

			TExpression *p_result;
			if(t_token.n_type == token_Begin) {
				if((p_result = p_AllocExpression(token_Number, token_Matrix, 0)) == NULL ||
				   (p_result->t_result.p_matrix = p_GetTempMatrix(p_handle)) == NULL) {
					if(p_result)
						FreeExpression_r(p_result);
					*p_error_string_ptr = "not enough memory";
					return NULL;
				}
				// alloc matrix

				for(int j = 0; j < 4; ++ j) {
					if(j) {
						if(t_token.n_type != token_Comma) {
							*p_error_string_ptr = "missing: \',\'";
							*p_error_pos = t_token.n_string_pos;
							FreeExpression_r(p_result);
							return NULL;
						}
						if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
							FreeExpression_r(p_result);
							return NULL;
						}
					}
					// skip ,

					if(t_token.n_type != token_End) {
						*p_error_string_ptr = "missing: \'}\'";
						*p_error_pos = t_token.n_string_pos;
						FreeExpression_r(p_result);
						return NULL;
					}
					if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
						FreeExpression_r(p_result);
						return NULL;
					}
					// skip {

					for(int i = 0; i < 4; ++ i) {
						if(i) {
							if(t_token.n_type != token_Comma) {
								*p_error_string_ptr = "missing: \',\'";
								*p_error_pos = t_token.n_string_pos;
								FreeExpression_r(p_result);
								return NULL;
							}
							if(*p_error_string_ptr = p_s_GetToken(p_s_expr,
							   &t_token, p_expr_pos)) {
								FreeExpression_r(p_result);
								return NULL;
							}
						}
						// skip ,

						if(t_token.n_type != token_Number) {
							*p_error_string_ptr = "constant matrix must contain "
								"number literals only";
							*p_error_pos = t_token.n_string_pos;
							FreeExpression_r(p_result);
							return NULL;
						}

						(*p_result->t_result.p_matrix)[i][j] = t_token.t_data.f_value;

						if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
							FreeExpression_r(p_result);
							return NULL;
						}
						// read number
					}

					if(t_token.n_type != token_End) {
						*p_error_string_ptr = "missing: \'}\'";
						*p_error_pos = t_token.n_string_pos;
						FreeExpression_r(p_result);
						return NULL;
					}
					if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
						FreeExpression_r(p_result);
						return NULL;
					}
					// skip }
				}
				// matrix
			} else {
				if((p_result = p_AllocExpression(token_Number, token_Vector, 0)) == NULL ||
				   (p_result->t_result.p_vector = p_GetTempVector(p_handle))) {
					if(p_result)
						FreeExpression_r(p_result);
					*p_error_string_ptr = "not enough memory";
					return NULL;
				}
				// alloc vector

				for(int i = 0; i < 4; ++ i) {
					if(i) {
						if(t_token.n_type != token_Comma) {
							*p_error_string_ptr = "missing: \',\'";
							*p_error_pos = t_token.n_string_pos;
							FreeExpression_r(p_result);
							return NULL;
						}
						if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
							FreeExpression_r(p_result);
							return NULL;
						}
					}
					// skip ,

					if(t_token.n_type != token_Number) {
						*p_error_string_ptr = "constant vector must contain number literals only";
						*p_error_pos = t_token.n_string_pos;
						FreeExpression_r(p_result);
						return NULL;
					}

					(*p_result->t_result.p_vector)[i] = t_token.t_data.f_value;

					if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
						FreeExpression_r(p_result);
						return NULL;
					}
					// read number
				}
				// vector
			}

			if(t_token.n_type != token_End) {
				*p_error_string_ptr = "missing: \'}\'";
				*p_error_pos = t_token.n_string_pos;
				FreeExpression_r(p_result);
				return NULL;
			}
			if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
				FreeExpression_r(p_result);
				return NULL;
			}
			// skip }

			p_result->p_exec_func = NULL;

			return p_result;
		}
	case token_End:
		*p_error_string_ptr = "unexpected token: \'}\'";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	case token_Comma:
		*p_error_string_ptr = "unexpected token: \',\'";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	case token_Add:
		*p_error_string_ptr = "unexpected token: \'+\'";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	case token_Sub:
		*p_error_string_ptr = "unexpected token: \'-\'";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	case token_Mul:
		*p_error_string_ptr = "unexpected token: \'*\'";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	case token_Div:
		*p_error_string_ptr = "unexpected token: \'/\'";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	case token_True:
	case token_False:
	case token_Square:
	case token_Diamond:
		{
			TExpression *p_result;
			if((p_result = p_AllocExpression(t_token.n_type, t_token.n_type, 0)) == NULL) {
				*p_error_string_ptr = "not enough memory";
				return NULL;
			}
			// variable

			if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
				FreeExpression_r(p_result);
				return NULL;
			}
			// get next token

			p_result->p_exec_func = NULL;

			return p_result;
		}

	case token_Invert: // invert(i)
	case token_Abs: // abs(i)
	case token_Log: // log(i)
	case token_Exp: // exp(i)
	case token_Sqrt: // sqrt(i)
	case token_Grey: // grey(i)
	case token_Blend: // blend(i, i, n, n)
	case token_Min: // min(i), min(i, i), min(i, n, n)
	case token_Max: // max(i), max(i, i), max(i, n, n)
	case token_Thresh: // thresh(i, n)
	case token_Pow: // pow(i, n)
	case token_Median: // median(i, n)
	case token_Scale: // scale(i, n, n)
	case token_Mean: // mean(i, n, n)
	case token_CMatrix: // cmatrix(i, m)
	case token_Decal: // decal(i, i)
	case token_HitMiss: // hitmiss(i, i)
	case token_Erode: // close(i, i), close(i, square|diamond, n, n)
	case token_Dilate: // close(i, i), close(i, square|diamond, n, n)
	case token_Open: // close(i, i), close(i, square|diamond, n, n)
	case token_Close: // close(i, i), close(i, square|diamond, n, n)
	case token_Resize: // resize(i[, n, n])
	case token_Convolve: // convolve(i, i), convolve(i, i, i)
	case token_Lerp: // lerp(i[, n, n],  n, n,  n, n,  n, n,  n, n)
	case token_Rotate: // rotate(i[, n, n], n, true|false, n) // angle, scaledown?, zoom
		{
			TToken t_func_token = t_token;
			int n_max_param_num;
			switch(t_token.n_type) {
			case token_Invert: // invert(i)
			case token_Abs: // abs(i)
			case token_Log: // log(i)
			case token_Exp: // exp(i)
			case token_Sqrt: // sqrt(i)
			case token_Grey: // grey(i)
				n_max_param_num = 1;
				break;
			case token_Thresh: // thresh(i, n)
			case token_Pow: // pow(i, n)
			case token_Median: // median(i, n)
			case token_CMatrix: // cmatrix(i, m)
			case token_Decal: // decal(i, i)
			case token_HitMiss: // hitmiss(i, i)
				n_max_param_num = 2;
				break;
			case token_Scale: // scale(i, n, n)
			case token_Mean: // mean(i, n, n)
			case token_Min: // min(i), min(i, i), min(i, n, n)
			case token_Max: // max(i), max(i, i), max(i, n, n)
			case token_Resize: // resize(i[, n, n])
			case token_Convolve: // convolve(i, i), convolve(i, i, i)
				n_max_param_num = 3;
				break;
			case token_Blend: // blend(i, i, n, n)
			case token_Erode: // close(i, i), close(i, square|diamond, n, n)
			case token_Dilate: // close(i, i), close(i, square|diamond, n, n)
			case token_Open: // close(i, i), close(i, square|diamond, n, n)
			case token_Close: // close(i, i), close(i, square|diamond, n, n)
				n_max_param_num = 4;
				break;
			case token_Lerp: // lerp(i[, n, n],  n, n,  n, n,  n, n,  n, n)
				n_max_param_num = 11;
				break;
			case token_Rotate: // rotate(i[, n, n], n, true|false, n) // angle, scaledown?, zoom
				n_max_param_num = 6;
				break;
			}

			TExpression *p_result;
			if((p_result = p_AllocExpression(t_token.n_type,
			   token_Image, n_max_param_num)) == NULL) {
				*p_error_string_ptr = "not enough memory";
				return NULL;
			}
			// variable

			do {
				if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos))
					break;
				if(t_token.n_type != token_LeftPar) {
					*p_error_string_ptr = "missing: \'(\'";
					*p_error_pos = t_token.n_string_pos;
					break;
				}
				if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos))
					break;
				// skip func(

				int n_param_num = 0;
				for(int i = 0; i < n_max_param_num; ++ i, ++ n_param_num) {
					if(t_token.n_type == token_RightPar)
						break;

					if(i) {
						if(t_token.n_type != token_Comma) {
							*p_error_string_ptr = "missing: \',\'";
							*p_error_pos = t_token.n_string_pos;
							break;
						}
						if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos))
							break;
						// skip ,
					}

					if((p_result->p_sub_expr[i] = p_Parse_L1_r(p_s_expr, p_expr_pos, p_handle,
					   p_params, p_error_string_ptr, p_error_pos)) == NULL) {
						if(*p_error_string_ptr == NULL)
							*p_error_string_ptr = "failed to parse function parameter";
						break;
					}
				}
				if(*p_error_string_ptr != NULL)
					break;
				// parse paramters

				if(t_token.n_type != token_RightPar) {
					*p_error_string_ptr = "missing: \')\'";
					*p_error_pos = t_token.n_string_pos;
					break;
				}
				if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos))
					break;
				// skip )

				if(FindExecFunction(p_result, n_param_num, p_handle, p_error_string_ptr))
					return p_result;
				// find proper exec function, determine return value, alloc intermediates

				*p_error_pos = t_func_token.n_string_pos;
			} while(0);

			FreeExpression_r(p_result);
			return NULL;
		}

	case token_EndOfLine:
		*p_error_string_ptr = "unexpected end of line";
		return NULL;
	default:
		*p_error_string_ptr = "unknown token";
		*p_error_pos = t_token.n_string_pos;
		return NULL;
	}
}

/**
 *	TExpression *p_Parse_L2_r(const char *p_s_expr, unsigned int *p_expr_pos,
 *		TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
 *		unsigned int *p_error_pos)
 *		- recursive descent parsing routine, level 2
 *		- returns expression tree or NULL in case errors occured
 *		- parses p_s_expr at p_expr_pos which is updated. temporary objects are allocated
 *		  inside p_handle, first call parameters p_params are used to determine temporary
 *		  image formats
 *		- p_error_string_ptr and p_error_pos are error string pointer and error position pointer.
 *		  filled when error occurs.
 */
TExpression *p_Parse_L2_r(const char *p_s_expr, unsigned int *p_expr_pos,
	TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
	unsigned int *p_error_pos)
{
	TExpression *p_left = p_Parse_L3_r(p_s_expr, p_expr_pos,
		p_handle, p_params, p_error_string_ptr, p_error_pos);
	// parse left side

	while(1) {
		TToken t_oper = t_token;
		switch(t_token.n_type) {
		case token_Mul:
		case token_Div:
			{
				if(p_left == NULL) {
					if(*p_error_string_ptr == NULL) {
						*p_error_string_ptr = (t_token.n_type == token_Mul)?
							"unexpected token: \'*\'" : "unexpected token: \'/\'";
						*p_error_pos = t_token.n_string_pos;
					}
					return NULL;
				}

				if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
					FreeExpression_r(p_left);
					return NULL;
				}
				// skip operator token

				TExpression *p_right;
				if((p_right = p_Parse_L3_r(p_s_expr, p_expr_pos,
				   p_handle, p_params, p_error_string_ptr, p_error_pos)) == NULL) {
					FreeExpression_r(p_left);
					return NULL;
				}
				// parse right side

				TExpression *p_result;
				if((p_result = p_AllocExpression(t_oper.n_type,
				   max(p_left->n_return_type, p_left->n_return_type), 2)) == NULL) {
					FreeExpression_r(p_right);
					FreeExpression_r(p_left);
					*p_error_string_ptr = "not enough memory";
					return NULL;
				}
				// alloc expression (return type is the more complex of two expressions)

				p_result->p_sub_expr[0] = p_left;
				p_result->p_sub_expr[1] = p_right;
				// set pointers

				if(!FindExecFunction(p_result, 2, p_handle, p_error_string_ptr)) {
					*p_error_pos = t_oper.n_string_pos;
					FreeExpression_r(p_result);
					return NULL;
				}
				// find proper exec function, determine return value, alloc intermediates

				p_left = p_result;
				break;
			}
		default:
			return p_left;
		}
	}
}

/**
 *	TExpression *p_Parse_L1_r(const char *p_s_expr, unsigned int *p_expr_pos,
 *		TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
 *		unsigned int *p_error_pos)
 *		- recursive descent parsing routine, level 1
 *		- returns expression tree or NULL in case errors occured
 *		- parses p_s_expr at p_expr_pos which is updated. temporary objects are allocated
 *		  inside p_handle, first call parameters p_params are used to determine temporary
 *		  image formats
 *		- p_error_string_ptr and p_error_pos are error string pointer and error position pointer.
 *		  filled when error occurs.
 */
TExpression *p_Parse_L1_r(const char *p_s_expr, unsigned int *p_expr_pos,
	TExpressionHandle *p_handle, void **p_params, char **p_error_string_ptr,
	unsigned int *p_error_pos)
{
	TExpression *p_left = p_Parse_L2_r(p_s_expr, p_expr_pos,
		p_handle, p_params, p_error_string_ptr, p_error_pos);
	// parse left side

	while(1) {
		TToken t_oper = t_token;
		switch(t_token.n_type) {
		case token_Add:
		case token_Sub:
			{
				if(p_left == NULL) {
					if(*p_error_string_ptr == NULL) {
						*p_error_string_ptr = (t_token.n_type == token_Add)?
							"unexpected token: \'+\'" : "unexpected token: \'-\'";
						*p_error_pos = t_token.n_string_pos;
					}
					return NULL;
				}

				if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, p_expr_pos)) {
					FreeExpression_r(p_left);
					return NULL;
				}
				// skip operator token

				TExpression *p_right;
				if((p_right = p_Parse_L2_r(p_s_expr, p_expr_pos,
				   p_handle, p_params, p_error_string_ptr, p_error_pos)) == NULL) {
					FreeExpression_r(p_left);
					return NULL;
				}
				// parse right side

				TExpression *p_result;
				if((p_result = p_AllocExpression(t_oper.n_type,
				   max(p_left->n_return_type, p_left->n_return_type), 2)) == NULL) {
					FreeExpression_r(p_right);
					FreeExpression_r(p_left);
					*p_error_string_ptr = "not enough memory";
					return NULL;
				}
				// alloc expression (return type is the more complex of two expressions)

				p_result->p_sub_expr[0] = p_left;
				p_result->p_sub_expr[1] = p_right;
				// set pointers

				if(!FindExecFunction(p_result, 2, p_handle, p_error_string_ptr)) {
					*p_error_pos = t_oper.n_string_pos;
					FreeExpression_r(p_result);
					return NULL;
				}
				// find proper exec function, determine return value, alloc intermediates

				p_left = p_result;
				break;
			}
		default:
			return p_left;
		}
	}
}

/**
 *	int GetProgramOps_r(TExpression *p_expression, TOperation *p_program, bool b_count)
 *		- gets progam operations from p_expression tree
 *		- returns number of operations, in case b_count is false then operations
 *		  are written into p_program array which must have sufficient length
 */
int GetProgramOps_r(TExpression *p_expression, TOperation *p_program, bool b_count)
{
	int n_size = 0;
	for(int i = 0; i < p_expression->n_sub_expr_num; ++ i) {
		if(p_expression->p_sub_expr[i] != NULL) {
			n_size += GetProgramOps_r(p_expression->p_sub_expr[i],
				p_program + n_size, b_count);
		}
	}
	// get operations from sub-expressions

	if(p_expression->p_exec_func) {
		if(!b_count) {
			p_program[n_size].p_exec_func = p_expression->p_exec_func;
			p_program[n_size].p_expression = p_expression;
		}
		++ n_size;
	}
	// get operation from self

	return n_size;
}

/**
 *	bool CreateProgram(TExpressionHandle *p_handle)
 *		- creates linear array, containing expression execution program
 *		- returns true on success, false on failure
 */
bool CreateProgram(TExpressionHandle *p_handle)
{
	if(p_handle->p_value == NULL)
		return false;
	// incorrectly parsed handle

	if(p_handle->n_program_length) {
		free(p_handle->p_program);
		p_handle->n_program_length = 0;
	}
	// free previous program if present

	int n_program_length = GetProgramOps_r(p_handle->p_value, NULL, true);
	// count program operations

	if((p_handle->p_program = (TOperation*)malloc(n_program_length * sizeof(TOperation))) == NULL)
		return false;
	// alloc program operations

	p_handle->n_program_length = n_program_length;
	GetProgramOps_r(p_handle->p_value, p_handle->p_program, false);
	// get program operations

	return true;
}

/**
 *	TExpressionHandle *p_ParseExpr(const char *p_s_expr, void **p_params,
 *		char **p_error_string_ptr, unsigned int *p_error_pos)
 *		- parses expression p_s_expr, accompanied by p_params
 *		  (pointers to objects referenced in equation)
 *		- note temporary image sizes are taken from input image sizes
 *		- in case error occurs, p_error_string_ptr is written constant pointer with error
 *		  description and p_error_pos is filled with char position where error occured
 *		- returns expression handle or NULL
 */
TExpressionHandle *p_ParseExpr(const char *p_s_expr, void **p_params,
	char **p_error_string_ptr, unsigned int *p_error_pos)
{
	for(int i = 0; i < n_global_handle_num; ++ i) {
		if(!strcmp(p_handle_list[i]->p_s_expression, p_s_expr)) {
			++ p_handle_list[i]->n_reference_num;
			return p_handle_list[i];
		}
	}
	// try to find existing handle to such an expression

	unsigned int n_expr_pos = 0;
	if(*p_error_string_ptr = p_s_GetToken(p_s_expr, &t_token, &n_expr_pos))
		return NULL;
	// try to get first token

	TExpressionHandle *p_handle;
	if((p_handle = p_AllocHandle(p_s_expr)) == NULL) {
		*p_error_string_ptr = "not enough memory";
		*p_error_pos = 0;
		return NULL;
	}
	// alloc

	*p_error_pos = -1;

	if((p_handle->p_value = p_Parse_L1_r(p_s_expr, &n_expr_pos,
	   p_handle, p_params, p_error_string_ptr, p_error_pos)) == NULL) {
		if(*p_error_string_ptr == NULL)
			*p_error_string_ptr = "parsing failed";
		if(*p_error_pos == -1)
			*p_error_pos = n_expr_pos;
		FreeHandle(p_handle);
		return NULL;
	}
	// parse

	if(t_token.n_type != token_EndOfLine) {
		*p_error_string_ptr = "unexpected token";
		*p_error_pos = t_token.n_string_pos;
		FreeHandle(p_handle);
		return NULL;
	}
	// did parsing go to the end?

	if(p_handle->p_value->n_return_type != token_Image) {
		*p_error_string_ptr = "top-level expression must return image";
		*p_error_pos = strlen(p_s_expr);
		FreeHandle(p_handle);
		return NULL;
	}
	// does it return image?

	if(!CreateProgram(p_handle)) {
		*p_error_string_ptr = "not enough memory";
		*p_error_pos = strlen(p_s_expr);
		FreeHandle(p_handle);
		return NULL;
	}
	// create program for expression execution

	return p_handle;
}

/**
 *	bool ImageExpression(ImageStruct *p_dest, TExpressionHandle *p_expr, void **p_params)
 *		- executes expression p_expr with parameters p_params, result is written to p_dest
 *		- returns true on success, false on failure
 */
bool ImageExpression(ImageStruct *p_dest, TExpressionHandle *p_expr, void **p_params)
{
	if(p_expr == NULL || !p_expr->n_program_length || p_expr->p_value == NULL)
		return false;

	bool b_downloads = b_AutoDownloads_Enabled();
	if(b_downloads)
		Disable_AutoDownloads();
	bool b_uploads = b_AutoUploads_Enabled();
	if(b_uploads)
		Disable_AutoUploads();

	if(p_expr->p_value->n_type != token_Image)
		p_expr->p_value->t_result.p_image = p_dest;
	// in case expression root is operation, set dest image as it's intermediate result

	for(int i = 0, n = p_expr->n_program_length; i < n; ++ i) {
		if(!p_expr->p_program[i].p_exec_func(p_expr->p_program[i].p_expression, p_params)) {
			if(b_downloads)
				Enable_AutoDownloads();
			if(b_uploads)
				Enable_AutoUploads();
			return false;
		}
	}
	// call GPU lib functions

	if(p_expr->p_value->n_type == token_Image)
		ImageCopy(p_dest, p_expr->p_value->t_result.p_image);
	// in case expression is just image alone, it's necessary to copy it

	if(b_downloads)
		Enable_AutoDownloads();
	if(b_uploads)
		Enable_AutoUploads();

	return Download_ImageStruct(p_dest);
}

/*
 *								--- ~parser ---
 */
