/*
								+----------------------------------+
								|                                  |
								|  ***  Expression evaluator  ***  |
								|                                  |
								|   Copyright   -tHE SWINe- 2006  |
								|                                  |
								|          GPU_ExpEval.h           |
								|                                  |
								+----------------------------------+
*/

#ifndef __EXPRESSION_EVALUATOR_INCLUDED
#define __EXPRESSION_EVALUATOR_INCLUDED

/*
 *								--- parser ---
 */

typedef float TMatrix4x4[4][4];
typedef float TVector4[4];

/**
 *	struct TExpression
 *		- expression tree node
 *		- holds it's subnodes, result value and execution function
 */
typedef struct _TExpression {
	int n_type;
	int n_return_type; // token_Image, token_Scalar, token_Vector, token_Matrix

	int n_variable_id; // in case it's variable reference. otherwise -1
	union {
		ImageStruct *p_image; // points to temp images, top-level expression points to output
		float f_scalar;
		TVector4 *p_vector; // poitns to temp vectors
		TMatrix4x4 *p_matrix; // points to temp matrices
		void *p_void_ptr; // to fill pointer regardless of type
	} t_result;

	int n_sub_expr_num;
	struct _TExpression **p_sub_expr;

	bool (*p_exec_func)(struct _TExpression *p_expr, void **p_params); // execution function (calls GPU_Lib)
} TExpression;

/**
 *	struct TOperation
 *		- single image-processing operation
 *		- because of effectiveness, operations are not executed on tree but on linear array
 *		  so this contains execution function and associated expression, containing data.
 */
typedef struct {
	TExpression *p_expression;
	bool (*p_exec_func)(struct _TExpression *p_expr, void **p_params);
} TOperation;

/**
 *	struct TExpressionHandle
 *		- expression handle
 *		- contains information for execution, temporary data structures and
 *		  original expression string
 *		- when compiling same expression multiple times, existing handle is returned and it's
 *		  reference counter is increased
 */
typedef struct {
	int n_reference_num;
	// number of references

	TExpression *p_value;
	// expression value

	char *p_s_expression;
	// copy of expression string

	ImageStruct **p_temp_image;
	int n_temp_image_num;
	TVector4 **p_temp_vector;
	int n_temp_vector_num;
	TMatrix4x4 **p_temp_matrix;
	int n_temp_matrix_num;
	// temporary data storage

	TOperation *p_program;
	int n_program_length;
	// program to execute expression
} TExpressionHandle;

/**
 *	void FreeHandle(TExpressionHandle *p_handle)
 *		- "deletes" expression handle
 *		- decreases it's reference counter, but doesn't free anything until it reaches zero
 */
void FreeHandle(TExpressionHandle *p_handle);

/**
 *	TExpressionHandle *p_ParseExpr(const char *p_s_expr, void **p_params,
 *		char **p_error_string_ptr, unsigned int *p_error_pos)
 *		- parses expression p_s_expr, accompanied by p_params
 *		  (pointers to objects referenced in equation)
 *		- note temporary image sizes are taken from input image sizes
 *		- in case error occurs, p_error_string_ptr is written constant pointer with error
 *		  description and p_error_pos is filled with char position where error occured
 *		- returns expression handle or NULL
 */
TExpressionHandle *p_ParseExpr(const char *p_s_expr, void **p_params,
	char **p_error_string_ptr, unsigned int *p_error_pos);

/**
 *	bool ImageExpression(ImageStruct *p_dest, TExpressionHandle *p_expr, void **p_params)
 *		- executes expression p_expr with parameters p_params, result is written to p_dest
 *		- returns true on success, false on failure
 */
bool ImageExpression(ImageStruct *p_dest, TExpressionHandle *p_expr, void **p_params);

/*
 *								--- ~parser ---
 */

#endif //__EXPRESSION_EVALUATOR_INCLUDED
