/*
								+----------------------------------+
								|                                  |
								|    ***  Image processing  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|          ImageProc.cpp           |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glext.h>
#include <math.h>
#include <algorithm>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLDrv.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/VertexBufferObject.h"
#include "../berLame_src/Vector.h"
#include "GPU_Lib.h"
#include "TempTex_Manager.h"
#include "Shader_Manager.h"
#include "RenderBuffer_Manager.h"
#include "GPU_Lib_Private.h"
#include "ImageTransfer.h"
#include "ImageBind.h"
#include "ImageProc.h"

/**
 *	bool _ImageImageFixed(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
 *		ImageStruct *p_src_image1, GLenum n_image1_env_mode)
 *		- processes two images p_src_image0, p_src_image1 using fixed pipeline
 *		- result is rendered to p_dest_image
 *		- p_state is OpenGL state cache
 *		- second texture unit is set to n_image1_env_mode
 *		- returns true on success, false on failure
 */
bool _ImageImageFixed(CGLState *p_state, ImageStruct *p_dest_image, ImageStruct *p_src_image0,
	ImageStruct *p_src_image1, GLenum n_image1_env_mode)
{
	CGLNamedRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = p_Bind_DestImage(p_state, p_dest_image)))
		return false;
	// bind destination image as render-buffer

	p_state->ActiveTextureUnit(0);
	if(!Bind_SrcImage(p_state, p_src_image0)) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	// bind source images as textures

	p_state->ActiveTextureUnit(1);
	if(!Bind_SrcImage(p_state, p_src_image1)) {
		Release_DestImage(p_state, p_dest_image, p_render_buffer);
		return false;
	}
	if(n_image1_env_mode != GL_MODULATE)
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, n_image1_env_mode);

	Set_NoShaderState(p_state);
	// set required GL state

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	FS_Quad_2Texcoords();
	// render fullscreen quad

	if(!Release_DestImage(p_state, p_dest_image, p_render_buffer))
		return false;
	// unbind render-buffer

	//p_state->ActiveTextureUnit(1);
	if(n_image1_env_mode != GL_MODULATE)
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); // return default
	if(!Release_SrcImage(p_state, p_src_image1))
		return false;
	p_state->ActiveTextureUnit(0);
	if(!Release_SrcImage(p_state, p_src_image0))
		return false;
	// refresh memory mapping for source images if necessary

	return true;
}

/**
 *	int _n_Prepare_ImagingRendering(CGLState *p_state, CGLPixelBufferObject &r_tmp_imaging_buffer,
 *		CGLNamedRenderBuffer_FBO *&r_p_render_buffer, ImageStruct *p_src_image, int &r_n_format,
 *		int &r_n_data_type, unsigned int &r_n_pixel_size)
 *		- prepares imaging rendering which needs to have data in memory area, accessible for
 *		  glDrawPixels(). copies data from p_src_image to r_tmp_imaging_buffer,
 *		  creates r_p_render_buffer for rendering p_src_image (needs it even if pixels are sinked)
 *		  and fills texture format informations
 *		- returns texture id (texture name for manager, not OpenGL id) or -1 on failure
 */
int _n_Prepare_ImagingRendering(CGLState *p_state, CGLPixelBufferObject &r_tmp_imaging_buffer,
	CGLNamedRenderBuffer_FBO *&r_p_render_buffer, ImageStruct *p_src_image, int &r_n_format,
	int &r_n_data_type, unsigned int &r_n_pixel_size)
{
	int n_format, n_data_type, n_internal_format;
	if(!Get_GL_PixelType(p_src_image->PixelType, n_format, n_data_type, n_internal_format))
		return -1;
	r_n_format = n_format;
	r_n_data_type = n_data_type;
	// get pixel format

	p_state->ActiveTextureUnit(0);
	TImageStructInfo *p_image_info;
	if(!Bind_SrcImage(p_state, p_src_image, &p_image_info))
		return -1;
	// bind source image as texture

	p_state->BindPixelUnPackBuffer(0);
	// explicitly bind no unpack buffer otherwise buffer data would be read from there !!!

	unsigned int n_pixel_size = n_GL_PixelSize(n_internal_format);
	unsigned int n_image_size = p_src_image->XSize * p_src_image->YSize * n_pixel_size;
	if(!r_tmp_imaging_buffer.Alloc(p_state, n_image_size, 0, GL_STREAM_COPY_ARB))
		return -1;
	r_n_pixel_size = n_pixel_size;
	// make sure there is enough space in the buffer to store texture; reset buffer data everytime

	glPixelStorei(GL_PACK_SKIP_ROWS, 0);
	glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_PACK_ROW_LENGTH, 0);
	glPixelStorei(GL_PACK_ALIGNMENT, n_pixel_size);
	// set proper transfer mode

	r_tmp_imaging_buffer.Bind(p_state);
	glGetTexImage(GL_TEXTURE_2D,
				0, // LOD pyramid level
				n_format,
				n_data_type,
				r_tmp_imaging_buffer.p_OffsetPointer(0));
	//p_state->BindPixelPackBuffer(0); // imaging subset requires this
	// fast block copy in graphics card (texture to buffer)

	if(!Release_SrcImage(p_state, p_src_image))
		return -1;
	// release src image

	p_state->BindPixelPackBuffer(0); // imaging subset requires this
	// moved here because Release_SrcImage may need to map pixel buffer

	Set_NoShaderState(p_state);
	// no shaders

	CGLNamedTexture_2D *p_temp_texture;
	if(!(p_temp_texture = p_GetTemporaryTexture2D(TTextureFormat(p_src_image->XSize,
	   p_src_image->YSize, n_internal_format))))
		return -1;
	// get unused / create new temporary texture

	if(!(r_p_render_buffer = p_Bind_Texture_Raw(p_state, p_temp_texture))) {
		ReturnTemporaryTexture(p_temp_texture->n_Id());
		return -1;
	}
	// bind the texture for rendering

	glRasterPos2i(-1, -1);
	// create the right viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// make sure matrices are unit

	return p_temp_texture->n_Id();
}

/*
 *		-end-of-file-
 */
