/*
								+----------------------------------+
								|                                  |
								|  ***  Reference rasterizer  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|           PixelConv.h            |
								|                                  |
								+----------------------------------+
*/

#ifndef __REFRAST_PIXEL_CONVERSION_INCLUDED
#define __REFRAST_PIXEL_CONVERSION_INCLUDED

#include <limits.h>
#include "GPU_Lib.h" // pixel types, ImageStruct

#ifndef SHORT_MIN
#define SHORT_MIN (-32768)
#endif
#ifndef SHORT_MAX
#define SHORT_MAX 32767
#endif
// MSVC++ doesn't know this

/**
 *	#define RGBA(r,g,b,a)
 *		- simple RGBA macro
 */
#define RGBA(r,g,b,a) (((r) << 16) | ((g) << 8) | (b) | ((a) << 24))

inline Image8Value GetPixel8(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y)
{
	return *(Image8Value*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y);
}

inline Image16Value GetPixel16(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y)
{
	return *(Image16Value*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y);
}

inline Image32Value GetPixel32(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y)
{
	return *(Image32Value*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y);
}

inline ImageFloatValue GetPixelFloat(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y)
{
	return *(ImageFloatValue*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y);
}

inline ImageRGBValueRGB GetPixelRGB(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y)
{
	return *(ImageRGBValueRGB*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y);
}

inline ImageComplexValue GetPixelComplex(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y)
{
	return *(ImageComplexValue*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y);
}

inline void GetPixel4f(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, float *p_dest)
{
	switch(p_image->PixelType) {
	case Image8:
	case Image8Linear:
		p_dest[0] = p_dest[1] = p_dest[2] = GetPixel8(p_image, n_x, n_y) / 255.0f;
		p_dest[3] = 1.0f;
		break;
	case Image16:
	case Image16Linear:
		p_dest[0] = p_dest[1] = p_dest[2] = GetPixel16(p_image, n_x, n_y) / (float)SHORT_MAX;
		p_dest[3] = 1.0f;
		break;
	case Image32:
	case Image32Linear:
		p_dest[0] = p_dest[1] = p_dest[2] = GetPixel32(p_image, n_x, n_y) / (float)LONG_MAX;
		p_dest[3] = 1.0f;
		break;
	case ImageFloat:
	case ImageFloatLinear:
		p_dest[0] = p_dest[1] = p_dest[2] = GetPixelFloat(p_image, n_x, n_y);
		p_dest[3] = 1.0f;
		break;
	case ImageRGB:
	case ImageRGBLinear:
		{
			ImageRGBValueRGB n_pixel = GetPixelRGB(p_image, n_x, n_y);
			p_dest[0] = ((n_pixel >> 16) & 0xff) / 255.0f;
			p_dest[1] = ((n_pixel >> 8) & 0xff) / 255.0f;
			p_dest[2] = (n_pixel & 0xff) / 255.0f;
			p_dest[3] = (n_pixel >> 24) / 255.0f;
		}
		break;
	case ImageComplex:
	case ImageComplexLinear:
		{
			ImageComplexValue t_pixel = GetPixelComplex(p_image, n_x, n_y);
			p_dest[0] = p_dest[1] = p_dest[2] = t_pixel.Re;
			p_dest[3] = t_pixel.Im;
		}
		break;
	}
}

inline void SetPixel8(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, Image8Value n_value)
{
	*(Image8Value*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y) = n_value;
}

inline void SetPixel16(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, Image16Value n_value)
{
	*(Image16Value*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y) = n_value;
}

inline void SetPixel32(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, Image32Value n_value)
{
	*(Image32Value*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y) = n_value;
}

inline void SetPixelFloat(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, ImageFloatValue f_value)
{
	*(ImageFloatValue*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y) = f_value;
}

inline void SetPixelRGB(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, ImageRGBValueRGB n_value)
{
	*(ImageRGBValueRGB*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y) = n_value;
}

inline void SetPixelComplex(const ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, ImageComplexValue t_value)
{
	*(ImageComplexValue*)(p_image->Raster + p_image->XOffset * n_x + p_image->YOffset * n_y) = t_value;
}

inline void SetPixel4f(ImageStruct *p_image,
	unsigned int n_x, unsigned int n_y, const float *p_dest)
{
	switch(p_image->PixelType) {
	case Image8:
	case Image8Linear:
		{
			float f_r = p_dest[0] * 255;
			SetPixel8(p_image, n_x, n_y, (f_r >= 0)? ((f_r <= 255)? (unsigned char)f_r : 0xff) : 0);
		}
		break;
	case Image16:
	case Image16Linear:
		{
			float f_r = p_dest[0] * SHORT_MAX;
			SetPixel16(p_image, n_x, n_y, (f_r >= SHORT_MIN)?
				((f_r <= SHORT_MAX)? (Image16Value)f_r : SHORT_MAX) : SHORT_MIN);
		}
		break;
	case Image32:
	case Image32Linear:
		{
			float f_r = p_dest[0] * LONG_MAX;
			SetPixel32(p_image, n_x, n_y, (f_r >= LONG_MIN)?
				((f_r <= LONG_MAX)? (Image32Value)f_r : LONG_MAX) : LONG_MIN);
		}
		break;
	case ImageFloat:
	case ImageFloatLinear:
		SetPixelFloat(p_image, n_x, n_y, p_dest[0]);
		break;
	case ImageRGB:
	case ImageRGBLinear:
		{
			float f_r = p_dest[0] * 255;
			float f_g = p_dest[1] * 255;
			float f_b = p_dest[2] * 255;
			float f_a = p_dest[3] * 255;
			unsigned char n_r = (f_r >= 0)? ((f_r <= 255)? (unsigned char)f_r : 0xff) : 0;
			unsigned char n_g = (f_g >= 0)? ((f_g <= 255)? (unsigned char)f_g : 0xff) : 0;
			unsigned char n_b = (f_b >= 0)? ((f_b <= 255)? (unsigned char)f_b : 0xff) : 0;
			unsigned char n_a = (f_a >= 0)? ((f_a <= 255)? (unsigned char)f_a : 0xff) : 0;
			SetPixelRGB(p_image, n_x, n_y, RGBA(n_r, n_g, n_b, n_a));
		}
		break;
	case ImageComplex:
	case ImageComplexLinear:
		{
			ImageComplexValue t_pixel = {p_dest[0], p_dest[3]};
			SetPixelComplex(p_image, n_x, n_y, t_pixel);
		}
		break;
	}
}

#endif //__REFRAST_PIXEL_CONVERSION_INCLUDED
