/*
								+----------------------------------+
								|                                  |
								|  ***  Reference rasterizer  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|           RefRaster.h            |
								|                                  |
								+----------------------------------+
*/

#ifndef __REFERENCE_RASTERZER_INCLUDED
#define __REFERENCE_RASTERZER_INCLUDED

/**
 *	bool rrImageCopy(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- copies source image to dest image. converts pixel types and
 *		  has free bilinear resampling, same as all the others
 *		- returns true on success, false on failure
 */
bool rrImageCopy(ImageStruct *p_dest, const ImageStruct *p_src);

/**
 *	bool rrImageImageAdd(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image addition point operator
 */
bool rrImageImageAdd(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageSub(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image subtraction point operator
 */
bool rrImageImageSub(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageMult(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image multiplication point operator
 */
bool rrImageImageMult(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageDiv(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image division point operator
 */
bool rrImageImageDiv(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageBlend(ImageStruct *p_dest, const ImageStruct *p_src0,
 *		const ImageStruct *p_src1, float f_weight0, float f_weight1)
 *		- calculates following equation: p_src_image0 * f_weight0 + p_src_image1 * f_weight1
 *		- weight magnitudes are unlimited, results are clamped based on dest image pixelformat
 */
bool rrImageImageBlend(ImageStruct *p_dest, const ImageStruct *p_src0,
	const ImageStruct *p_src1, float f_weight0, float f_weight1);

/**
 *	bool rrImageImageDecal(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- alpha-blends source image 1 onto source image 0, ie:
 *		  p_src_image0 * (1 - p_src_image1_alpha) + p_src_image1 * p_src_image1_alpha
 */
bool rrImageImageDecal(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageInvert(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise 1 - p_src_image
 */
bool rrImageInvert(ImageStruct *p_dest, const ImageStruct *p_src);

/**
 *	bool rrImageAbs(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise absolute value of p_src_image
 */
bool rrImageAbs(ImageStruct *p_dest, const ImageStruct *p_src);

/**
 *	bool rrImageImageMin(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- pointwise minimum of p_src_image0 and p_src_image1
 */
bool rrImageImageMin(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageMax(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- pointwise maximum of p_src_image0 and p_src_image1
 */
bool rrImageImageMax(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageThresh(ImageStruct *p_dest, const ImageStruct *p_src, float f_threshold)
 *		- thresholds image to f_threshold
 *		- note threshold value is relative to image pixel-type value range, for example
 *		  threshold 0.5 corresponds with both 128 for Image8 and 32767 for Image16
 *		- floating-point images are compared without any value rescaling
 */
bool rrImageThresh(ImageStruct *p_dest, const ImageStruct *p_src, float f_threshold);

/**
 *	bool rrImageLog(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise log(1 + p_src_image)
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool rrImageLog(ImageStruct *p_dest, const ImageStruct *p_src);

/**
 *	bool rrImageExp(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise exp(p_src_image) - 1
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool rrImageExp(ImageStruct *p_dest, const ImageStruct *p_src);

/**
 *	bool rrImageSqrt(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise square root of p_src_image
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageLog
 */
bool rrImageSqrt(ImageStruct *p_dest, const ImageStruct *p_src);

/**
 *	bool rrImagePow(ImageStruct *p_dest, const ImageStruct *p_src, float f_power)
 *		- pointwise raises p_src_image to f_power
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageExp
 */
bool rrImagePow(ImageStruct *p_dest, const ImageStruct *p_src, float f_power);

/**
 *	bool rrImageScale(ImageStruct *p_dest, const ImageStruct *p_src, float f_scale, float f_bias)
 *		- pointwise image scaling as p_src_image * f_scale + f_bias
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- scale and bias values are unlimited
 */
bool rrImageScale(ImageStruct *p_dest, const ImageStruct *p_src, float f_scale, float f_bias);

/**
 *	bool rrImageColorMatrix(ImageStruct *p_dest, const ImageStruct *p_src,
 *		const float *p_mat4x4, bool b_transpose)
 *		- treats each point of p_src_image as four-dimensional vector (RGBA),
 *		  multiplies it with matrix p_mat4x4 which can be applied as is or transposed
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- matrix elements values are unlimited
 *		- for example this matrix can be used to convert RGB images to YCbCr colorspace:
 *		  (note alpha channel of such an image must contain value 1 (respectively 255))
 *
 *		float p_mat4x4[16] = { // RGB to YCbCr (ITU-R BT.601)
 *			    .299f,     .587f,     .114f,   0, // R' = .299R + .587G + .114B
 *			-.168736f, -.331264f,       .5f, .5f, // G' = .5 - .168736R - .331264G - .5B
 *			      .5f, -.418688f, -.081312f, .5f, // B' = .5 + .5R - .418688G - .081312B
 *			        0,         0,         0,   1  // A' = 1
 *		};
 */
bool rrImageColorMatrix(ImageStruct *p_dest, const ImageStruct *p_src,
	const float *p_mat4x4, bool b_transpose);

/**
 *	bool rrImageGreyscale(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- turns source image into greyscale, using std ".299, .587, .144" equation
 *		- in case dest is RGBA image, RGB are set to gray value, A is set to 1
 */
bool rrImageGreyscale(ImageStruct *p_dest, const ImageStruct *p_src);

/**
 *	bool rrImageAffineTransform(ImageStruct *p_dest, const ImageStruct *p_src,
 *		float x0, float y0, float z0, float x1, float y1, float z1,
 *		float x2, float y2, float z2, float x3, float y3, float z3)
 *		- arbitrary image interpolation over convex quad given as (x0, y0, z0), (x1, y1, z1), ...
 *		- coordinates in OpenGL-ish fashion, orthographic projection
 */
bool rrImageAffineTransform(ImageStruct *p_dest, const ImageStruct *p_src,
	float x0, float y0, float z0, float x1, float y1, float z1,
	float x2, float y2, float z2, float x3, float y3, float z3);

/**
 *	bool rrImageSquareErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageSquareDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageSquareOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageSquareClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageDiamondErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageDiamondDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageDiamondOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageDiamondClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageImageErode(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- erosion morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageErode(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageDilate(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- dilation morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageDilate(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageOpen(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- opening morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageOpen(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageClose(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- closing morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageClose(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageImageHitMiss(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- hit-and-miss morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- structuring element is thresholded as follows: values below .25 are "zero",
 *		  values above .75 are "one", values between .25 and .75 are "don't care"
 */
bool rrImageImageHitMiss(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1);

/**
 *	bool rrImageMean(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- separable (two-pass) mean image filter
 *		- averaging window size is n_kernel_w  n_kernel_h pixels
 */
bool rrImageMean(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageMedian(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_size)
 *		- separable (two-pass) median image filter
 *		- averaging window size is n_kernel_size  n_kernel_size pixels, the only supported
 *		  sizes are 3  3, 5  5, 7  7 and 15  15.
 */
bool rrImageMedian(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_size);

/**
 *	bool rrImageLocalMin(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- finds local minimum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool rrImageLocalMin(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageLocalMax(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- finds local maximum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool rrImageLocalMax(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h);

/**
 *	bool rrImageConvolveSeparable(ImageStruct *p_dest, const ImageStruct *p_src,
 *		ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
 *		- separable convolution filter
 *		- this implementation imposes strict kernel alignment
 *		  (p_kernel_h->YSize = 1, p_kernel_v->XSize = 1)
 *		- note this handles convolution with complex kernels and / or complex image
 */
bool rrImageConvolveSeparable(ImageStruct *p_dest, const ImageStruct *p_src,
	ImageStruct *p_kernel_h, ImageStruct *p_kernel_v);

/**
 *	bool rrImageConvolve2D(ImageStruct *p_dest, const ImageStruct *p_src, ImageStruct *p_kernel)
 *		- two-dimensional convolution filter
 *		- note this handles convolution with complex kernel and / or complex image
 */
bool rrImageConvolve2D(ImageStruct *p_dest, const ImageStruct *p_src, ImageStruct *p_kernel);

/**
 *	bool rrImageGetHistogram(const ImageStruct *p_src, int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image histogram
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool rrImageGetHistogram(const ImageStruct *p_src, int n_histogram_width, unsigned int *p_histogram);

/**
 *	bool rrImageGetLuminosityHistogram(const ImageStruct *p_src,
 *		int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image luminosity histogram (greyscale conversion applied)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool rrImageGetLuminosityHistogram(const ImageStruct *p_src,
	int n_histogram_width, unsigned int *p_histogram);

/**
 *	bool rrImageGetRGBHistogram(const ImageStruct *p_src, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b)
 *		- retrieves RGB image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g and p_histo_b will contain histogram data, they doen't have
 *		  to be zeroed prior calling this
 */
bool rrImageGetRGBHistogram(const ImageStruct *p_src, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b);

/**
 *	bool rrImageGetRGBAHistogram(const ImageStruct *p_src, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g,
 *		unsigned int *p_histo_b, unsigned int *p_histo_a)
 *		- retrieves RGBA image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g, p_histo_b and p_histo_a will contain histogram data, they
 *		  doen't have to be zeroed prior calling this
 */
bool rrImageGetRGBAHistogram(const ImageStruct *p_src, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g,
	unsigned int *p_histo_b, unsigned int *p_histo_a);

/**
 *	bool rrImageGetMinMax(const ImageStruct *p_src, void *p_minmax)
 *		- get image minimal and maximal values
 *		- values are written to p_minmax in image pixel-type format in order min, then max
 */
bool rrImageGetMinMax(const ImageStruct *p_src, void *p_minmax);

#endif //__REFERENCE_RASTERZER_INCLUDED
