/*
								+----------------------------------+
								|                                  |
								|      *** Shader manager ***      |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|         Shader_Manager.h         |
								|                                  |
								+----------------------------------+
*/

#ifndef __GL_SHADER_MANAGER_INCLUDED
#define __GL_SHADER_MANAGER_INCLUDED

#include "Manager.h"

/**
 *	class CGLShaderObject
 *		- extension of shader object with automatically assigned id
 */
class CGLNamedShaderObject : public CGLShaderObject {
protected:
	static int m_n_id_space;
	int m_n_id;

public:
	inline int n_Id() const { return m_n_id; }

	inline CGLNamedShaderObject(const TShaderInfo *m_p_vertex_shader, const TShaderInfo *m_p_fragment_shader)
		:CGLShaderObject(m_p_vertex_shader, m_p_fragment_shader)
	{
		m_n_id = ++ m_n_id_space;
	}
};

//int CGLNamedShaderObject::m_n_id_space = 0;
// needs to be declared once

/**
 *	class CFindShader_Name
 *		- finds shader by shader name
 */
class CFindShader_Name {
protected:
	const char *m_p_s_shader_name;

public:
	inline CFindShader_Name(const char *p_s_shader_name)
		:m_p_s_shader_name(p_s_shader_name)
	{}

	inline bool operator ()(const char *p_s_name) const
	{
		return !strcmp(p_s_name, m_p_s_shader_name);
	}
};

/**
 *	class CShaderFactory
 *		- shader factory
 */
class CShaderFactory {
protected:
	struct TShaderSource {
		const char *p_s_shader_name;
		const TShaderInfo *p_vert_shader_src, *p_frag_shader_src;

		inline bool operator ==(const char *p_s_name)
		{
			return !strcmp(p_s_shader_name, p_s_name);
		}
	};
	static std::vector<TShaderSource> m_shader_source_list;

	CGLState *m_p_state;

public:
	inline CShaderFactory(CGLState *p_state)
		:m_p_state(p_state)
	{}

	inline CGLNamedShaderObject *Load(const char *p_s_name)
	{
		std::vector<TShaderSource>::const_iterator p_src_iter = std::find(m_shader_source_list.begin(),
			m_shader_source_list.end(), p_s_name);
		if(p_src_iter != m_shader_source_list.end()) {
			CGLNamedShaderObject *p_shader;
			if(!(p_shader = new CGLNamedShaderObject((*p_src_iter).p_vert_shader_src, (*p_src_iter).p_frag_shader_src)))
				return 0;

#ifdef GPU_LIB_REPORT_SHADER_COMPILE
			printf("compiling shader \'%s\' ... ", p_s_name);
#endif //GPU_LIB_REPORT_SHADER_COMPILE

			char *p_s_compile_log, *p_s_link_log;
			p_shader->Compile_Link(m_p_state, p_s_compile_log, p_s_link_log);
			// compile and link shader

			if(p_s_compile_log) {
#ifdef GPU_LIB_REPORT_SHADER_COMPILE_ERRORS
				fprintf(stderr, "\nwarning: shader \'%s\' compile log:\n%s\n",
					p_s_name, p_s_compile_log);
#endif //GPU_LIB_REPORT_SHADER_COMPILE_ERRORS
				delete[] p_s_compile_log;
			}
			if(p_s_link_log) {
#ifdef GPU_LIB_REPORT_SHADER_COMPILE_ERRORS
				fprintf(stderr, "\nwarning: shader \'%s\' link log:\n%s\n",
					p_s_name, p_s_link_log);
#endif //GPU_LIB_REPORT_SHADER_COMPILE_ERRORS
				delete[] p_s_link_log;
			}

			p_shader->Release(m_p_state);
			// make sure shader doesn't remain bound

#ifdef GPU_LIB_REPORT_SHADER_COMPILE
			printf("done\n");
#endif //GPU_LIB_REPORT_SHADER_COMPILE

			return p_shader;
		}
		return 0;
	}

	static inline bool b_Valid_Object(const CGLNamedShaderObject *p_shader)
	{
		return p_shader != 0 && p_shader->b_Status();
	}

	/**
	 *	static bool Register_ShaderSource(const char *p_s_shader_name,
	 *		const TShaderInfo *p_vert_shader_src, const TShaderInfo *p_frag_shader_src)
	 *		- registers shader source code, associated to shader name
	 *		- p_s_shader_name is shader name ("vert-name|frag-name"), p_vert_shader_src is
	 *		  vertex shader source code and p_frag_shader_src is fragment shader source code.
	 *		  one of source codes (typically vertex shader) can be 0
	 *		- doesn't check shader name uniqueness; shaders registered earlier will
	 *		  shadow any shaders with the same name inserted later
	 *		- returns true on success, false on failure
	 */
	static bool Register_ShaderSource(const char *p_s_shader_name,
		const TShaderInfo *p_vert_shader_src, const TShaderInfo *p_frag_shader_src)
	{
		_ASSERTE(p_s_shader_name && (p_frag_shader_src || p_vert_shader_src));

		m_shader_source_list.reserve(m_shader_source_list.size() + 1);
		if(m_shader_source_list.capacity() < m_shader_source_list.size() + 1)
			return false;
		TShaderSource t_tmp = {p_s_shader_name, p_vert_shader_src, p_frag_shader_src};
		m_shader_source_list.push_back(t_tmp);
		return true;
	}
};

//std::vector<CShaderFactory::TShaderSource> CShaderFactory::m_shader_source_list;
// needs to be declared once

/**
 *	class CGLShader_Manager
 *		- OpenGL shader manager class
 */
typedef CResourceManager<CGLNamedShaderObject, CUnitCost<CGLNamedShaderObject>, int, CFindIntId<CGLNamedShaderObject>,
	const char*, CFindShader_Name, CShaderFactory> CGLShader_Manager;

#endif //__GL_SHADER_MANAGER_INCLUDED
