/*
 *		--- speed-test program utility functions ---
 */

#include "../../berLame_src/Jpeg.h"
#include "../../berLame_src/Tga.h"
#include <math.h>

#ifdef WIN32
#include <windows.h>
#else
#pragma pack(2)

typedef short WORD;
typedef int DWORD;

typedef struct {
    WORD           bfType;
    DWORD          bfSize;
    WORD           bfReserved1;
    WORD           bfReserved2;
    DWORD          bfOffBits;
} BITMAPFILEHEADER;

typedef struct {
    DWORD          biSize;
    DWORD          biWidth;
    DWORD          biHeight;
    WORD           biPlanes;
    WORD           biBitCount;
    DWORD          biCompression;
    DWORD          biSizeImage;
    DWORD          biXPelsPerMeter;
    DWORD          biYPelsPerMeter;
    DWORD          biClrUsed;
    DWORD          biClrImportant;
} BITMAPINFOHEADER;

#pragma pack()
#define BI_RGB       0
#define BI_RLE8      1
#define BI_RLE4      2
#define BI_BITFIELDS 3

#endif //WIN32

bool Save_RGBLinearImageStruct(const char *p_s_file, const ImageStruct *p_imagestruct)
{
	if(p_imagestruct->PixelType != ImageRGBLinear)
		return false;
	// optimised for ImageRGBLinear images

	FILE *p_fw;
	if(!(p_fw = fopen(p_s_file, "wb")))
		return false;
	// open file

	BITMAPINFOHEADER t_info_header = {0};
	t_info_header.biSize = sizeof(BITMAPINFOHEADER);
	t_info_header.biWidth = p_imagestruct->XSize;
	t_info_header.biHeight = p_imagestruct->YSize;
	t_info_header.biPlanes = 1;
	t_info_header.biBitCount = 24;
	t_info_header.biCompression = BI_RGB;
	t_info_header.biSizeImage = p_imagestruct->XSize * p_imagestruct->YSize * 3;
	t_info_header.biXPelsPerMeter = 96;
	t_info_header.biYPelsPerMeter = 96;

	BITMAPFILEHEADER t_file_header = {0};
	t_file_header.bfType = ('B' | ('M' << 8));
	t_file_header.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	t_file_header.bfSize = t_file_header.bfOffBits + t_info_header.biSizeImage;
	// create header

	if(fwrite(&t_file_header, sizeof(BITMAPFILEHEADER), 1, p_fw) != 1 ||
	   fwrite(&t_info_header, sizeof(BITMAPINFOHEADER), 1, p_fw) != 1) {
		fclose(p_fw);
		return false;
	}
	// write header

	int n_scanline_padd = (3 * p_imagestruct->XSize) & 3;
	// length of scanline padding (padd to 4-byte multiples)

	for(const unsigned char *p_raster = p_imagestruct->Raster +
	   ((p_imagestruct->XSize - 1) * p_imagestruct->YSize * 4); p_raster >= p_imagestruct->Raster;) {
		for(const unsigned char *p_scanline_end = p_raster + p_imagestruct->XSize * 4;
		   p_raster != p_scanline_end; p_raster += 2) {
			if(fwrite(p_raster, sizeof(unsigned char), 1, p_fw) != 1 ||
			   fwrite(++ p_raster, sizeof(unsigned char), 1, p_fw) != 1 ||
			   fwrite(++ p_raster, sizeof(unsigned char), 1, p_fw) != 1) {
				fclose(p_fw);
				return false;
			}
		}
		p_raster -= 4 * p_imagestruct->XSize * 2;
		// write scanline

		static const unsigned char p_dummy[3] = {0};
		if(n_scanline_padd && fwrite(p_dummy, n_scanline_padd, 1, p_fw) != 1) {
			fclose(p_fw);
			return false;
		}
		// write padding
	}
	// write bitmap

	fclose(p_fw);

	return true;
}

TBmp *p_Load_Jpeg(const char *p_s_filename)
{
	TBmp *p_bitmap;
	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "rb"))) {
		fprintf(stderr, "error: error opening \'%s\'\n", p_s_filename);
		return 0;
	}
	CJpegDecoder decoder;
	if(!(p_bitmap = decoder.p_Decode_JPEG(p_fr))) {
		fprintf(stderr, "error: error loading \'%s\'\n", p_s_filename);
		fclose(p_fr);
		return 0;
	}
	fclose(p_fr);

	return p_bitmap;
}

ImageStruct *p_LoadImage(const char *p_s_filename)
{
	TBmp *p_bmp = 0;
	if(strstr(p_s_filename, ".jpg") && (!(p_bmp = p_Load_Jpeg(p_s_filename))))
		return NULL;
	if(strstr(p_s_filename, ".tga")) {
		if(!(p_bmp = p_Load_TGA(p_s_filename)))
			return NULL;
		for(unsigned __int32 *p_src = p_bmp->p_buffer, *p_end = p_bmp->p_buffer +
		   (p_bmp->n_width * p_bmp->n_height); p_src != p_end; ++ p_src)
			*p_src = (*p_src & 0xff00ff00) | ((*p_src >> 16) & 0xff) | ((*p_src & 0xff) << 16);
		// RGBA to BGRA
	}
	if(!p_bmp)
		return NULL;
	// load bitmap

	ImageStruct *p_imagestruct;
	if((p_imagestruct = p_Alloc_Image(p_bmp->n_width, p_bmp->n_height, ImageRGBLinear)) == NULL) {
		delete[] p_bmp->p_buffer;
		delete p_bmp;
		return NULL;
	}
	// get an empty image

	memcpy(p_imagestruct->Raster, p_bmp->p_buffer, p_bmp->n_width *
		p_bmp->n_height * sizeof(unsigned __int32));
	delete[] p_bmp->p_buffer;
	delete p_bmp;
	// fill it with data

	return p_imagestruct; // return it
}

ImageStruct *p_CreateGauss(unsigned int n_width, unsigned int n_height)
{
	ImageStruct *p_imagestruct;
	if((p_imagestruct = p_Alloc_Image(n_width, n_height, ImageFloatLinear)) == NULL)
		return NULL;
	// get an empty image

	float *p_raster = (float*)p_imagestruct->Raster;
	float f_sum = 0;
	for(unsigned int y = 0; y < n_height; ++ y) {
		float f_dy = y - n_height * .5f;
		for(unsigned int x = 0; x < n_width; ++ x) {
			float f_dx = x - n_width * .5f;
			float f_r = (float)sqrt(f_dx * f_dx + f_dy * f_dy) / (n_width * .5f);
			// calc function argument

			p_raster[x + n_width * y] = (float)exp(-.5f * f_r * f_r);
			if(p_raster[x + n_width * y] < 0)
				p_raster[x + n_width * y] = 0;
			f_sum += p_raster[x + n_width * y];
		}
	}
	f_sum = 1 / f_sum;
	for(unsigned int i = 0; i < n_width * n_height; ++ i)
		p_raster[i] *= f_sum;
	// calculate normalized gauss kernel

	return p_imagestruct; // return it
}

const unsigned int n_histogram_width = 64;
const unsigned int n_histogram_height = 8;

void PrintHistogramTable(FILE *p_fw, unsigned int *p_histogram)
{
	unsigned int n_max = p_histogram[0];
	for(unsigned int i = 1; i < n_histogram_width; ++ i) {
		if(n_max < p_histogram[i])
			n_max = p_histogram[i];
	}
	char p_s_histo_buffer[n_histogram_width + 1];
	p_s_histo_buffer[n_histogram_width] = 0;
	for(unsigned int y = 0; y < n_histogram_height; ++ y) {
		for(unsigned int x = 0; x < n_histogram_width; ++ x) {
			unsigned int n_value = (n_max)? (unsigned int)ceil(sqrt(p_histogram[x] /
				(float)n_max) * n_histogram_height) : 0;
			p_s_histo_buffer[x] = (n_value > n_histogram_height - y - 1)? '#' : ' ';
		}
		fprintf(p_fw, "|%s|\n", p_s_histo_buffer);
	}
}

void PrintHistogram(FILE *p_fw, ImageStruct *p_image)
{
	unsigned int p_histogram[n_histogram_width];
	ImageGetHistogram(p_image, n_histogram_width, p_histogram);
	PrintHistogramTable(p_fw, p_histogram);
}

void PrintLumaHistogram(FILE *p_fw, ImageStruct *p_image)
{
	unsigned int p_histogram[n_histogram_width];
	ImageGetLuminosityHistogram(p_image, n_histogram_width, p_histogram);
	PrintHistogramTable(p_fw, p_histogram);
}

void PrintRGBHistogram(FILE *p_fw, ImageStruct *p_image)
{
	unsigned int p_histogram_r[n_histogram_width];
	unsigned int p_histogram_g[n_histogram_width];
	unsigned int p_histogram_b[n_histogram_width];
	ImageGetRGBHistogram(p_image, n_histogram_width,
		p_histogram_r, p_histogram_g, p_histogram_b);
	fprintf(p_fw, "           ---- red ----\n");
	PrintHistogramTable(p_fw, p_histogram_r);
	fprintf(p_fw, "          ---- green ----\n");
	PrintHistogramTable(p_fw, p_histogram_g);
	fprintf(p_fw, "           ---- blue ----\n");
	PrintHistogramTable(p_fw, p_histogram_b);
}

void PrintRGBAHistogram(FILE *p_fw, ImageStruct *p_image)
{
	unsigned int p_histogram_r[n_histogram_width];
	unsigned int p_histogram_g[n_histogram_width];
	unsigned int p_histogram_b[n_histogram_width];
	unsigned int p_histogram_a[n_histogram_width];
	ImageGetRGBAHistogram(p_image, n_histogram_width,
		p_histogram_r, p_histogram_g, p_histogram_b, p_histogram_a);
	fprintf(p_fw, "           ---- red ----\n");
	PrintHistogramTable(p_fw, p_histogram_r);
	fprintf(p_fw, "          ---- green ----\n");
	PrintHistogramTable(p_fw, p_histogram_g);
	fprintf(p_fw, "           ---- blue ----\n");
	PrintHistogramTable(p_fw, p_histogram_b);
	fprintf(p_fw, "          ---- alpha ----\n");
	PrintHistogramTable(p_fw, p_histogram_a);
}

#define PrintARGB(p_fw,n) do { fprintf((p_fw), "(%d, %d, %d, %d)", ((n) >> 24), \
	((n) >> 16) & 0xff, ((n) >> 8) & 0xff, (n) & 0xff); } while(0)

/*
 *		-end-of-file-
 */
