/*
								+---------------------------------+
								|                                 |
								|      ***   OpenGL2.0   ***      |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Shader2.h            |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 10-09-2006
 *
 *	removed most of berLame stuff (todo - write new shader manager and generic
 *	encapsulation class)
 *	integers, containing just true or false were changed to bool
 *	fixed minor flaws in shader binding
 *	renamed to Shader2.cpp and Shader2.h respectively to avoid confusion with similar old source
 *
 *	passed code revision at 11-10-2006
 *
 *	created inline versions of Bind, Release and SetParam* functions so the encapsulation class,
 *	aware of shader api could use them (virtual functions themselves call them so there is no
 *	code written twice); note inlines are in Shader2.inl, there is too much of them to put them
 *	here
 *
 *	passed code revision at 05-10-2007
 *
 *	code to notify current GL state of shader deletion was added to GL_Free functions
 *	destructors don't invoke GL_Free in case shader isn't compiled
 *	void GL_Free(CGLState *p_state) was renamed to void Delete(void)
 *
 */

#ifndef __GL_SHADER2_INCLUDED
#define __GL_SHADER2_INCLUDED

// shader parameter types
enum {
	shader_param_Uniform, // ARB_*_shader unifrom
	shader_param_Parameter_Env, // ARB_*_program env param
	shader_param_Parameter_Local_VP, // ARB_vertex_program local param
	shader_param_Parameter_Local_FP // ARB_fragment_program local param
};

// shader parameter content types (do not swap / modify, algorithms rely on this order)
enum {
	shader_param_Unknown,
	shader_param_Float,
	shader_param_Float2,
	shader_param_Float3,
	shader_param_Float4,
	shader_param_Int,
	shader_param_Int2,
	shader_param_Int3,
	shader_param_Int4,
	shader_param_Bool,
	shader_param_Bool2,
	shader_param_Bool3,
	shader_param_Bool4,
	shader_param_Matrix2,
	shader_param_Matrix3,
	shader_param_Matrix4,
	shader_sampler_1D,
	shader_sampler_2D,
	shader_sampler_3D,
	shader_sampler_Cube,
	shader_sampler_1D_Shadow,
	shader_sampler_2D_Shadow,
	shader_sampler_2D_Rect,
	shader_sampler_2D_Rect_Shadow
};

/**
 *	class CShaderParamTypeConverter
 *		- OpenGL returns shader parameter data types as set of constants
 *		- for better convenience, those contains are translated to simple enum
 *		  with successive values so determining the group parameter data type
 *		  belongs to no longer requires switch(), two cmp will do
 */
class CShaderParamTypeConverter {
protected:
	static const struct TTableElem {
		int n_linear_enum_value;
		GLenum n_gl_value;
	} m_p_type_table[];

public:
	/**
	 *	static int CShaderParamTypeConverter::n_EnumValue(GLenum n_gl_value)
	 *		- translate OpenGL parameter type value to enum value
	 *		  (reason to do that is faster parameter type tests as enum values
	 *		  are successive integers)
	 *		- in case unknown parameter type is passed, returns shader_param_Unknown
	 */
	static int n_EnumValue(GLenum n_gl_value);

	/**
	 *	static __inline GLenum n_OpenGLValue(int n_enum_value)
	 *		- translate enum value n_enum_value to OpenGL parameter
	 *		- returns 0 (equal to GL_NONE) on shader_param_Unknown / bad enum value
	 */
	static __inline GLenum n_OpenGLValue(int n_enum_value)
	{
		if(n_enum_value >= shader_param_Float && n_enum_value <= shader_sampler_2D_Rect_Shadow)
			return m_p_type_table[n_enum_value - (int)shader_param_Float].n_gl_value;
		return 0;
	}
};

// shader api types
enum {
	shader_api_None,
	shader_api_ARB_program,
	shader_api_ARB_shader,
	shader_api_Core_shader // OpenGL20
};

/**
 *	class CGLProgram
 *		- most common class for OpenGL vertex / fragment shader / program
 */
class CGLProgram {
protected:
	const char *m_p_s_vertex_shader;
	const char *m_p_s_fragment_shader;

public:
	CGLProgram(const char *p_s_vertex_shader, const char *p_s_fragment_shader)
		:m_p_s_vertex_shader(p_s_vertex_shader),
		m_p_s_fragment_shader(p_s_fragment_shader)
	{ }

	virtual bool Compile(CGLState *p_state, char *&p_s_info_log) = 0;
	virtual bool SetParam4fv(CGLState *p_state, int n_param_type, int n_location, const float *p_value) = 0;
	virtual void Bind(CGLState *p_state) = 0;
	virtual void Release(CGLState *p_state) = 0;
	virtual void Delete() = 0;
	virtual int n_Api() const = 0;
};

/**
 *	class CGLShader : public CGLProgram
 *		- class for OpenGL vertex / fragment shader
 */
class CGLShader : public CGLProgram {
public:
	CGLShader(const char *p_s_vertex_shader, const char *p_s_fragment_shader)
		:CGLProgram(p_s_vertex_shader, p_s_fragment_shader)
	{ }

	virtual bool BindAttribLocation(int n_location, const char *p_s_attrib_name) = 0;
	virtual bool Link(char *&p_s_info_log) = 0;
	virtual int n_Uniform_Num() const = 0;
	virtual char *p_s_Uniform_Name(int n_index, int &r_n_content_type) const = 0;
	virtual int n_GetUniformLocation(const char *p_s_uniform_name) const = 0;
	virtual bool SetParam1fv(CGLState *p_state, int n_param_type, int n_location, const float *p_value) = 0;
	virtual bool SetParam2fv(CGLState *p_state, int n_param_type, int n_location, const float *p_value) = 0;
	virtual bool SetParam3fv(CGLState *p_state, int n_param_type, int n_location, const float *p_value) = 0;
	virtual bool SetParam1i(CGLState *p_state, int n_location, int n_value) = 0;
	virtual bool SetParam2iv(CGLState *p_state, int n_location, const int *p_value) = 0;
	virtual bool SetParam3iv(CGLState *p_state, int n_location, const int *p_value) = 0;
	virtual bool SetParam4iv(CGLState *p_state, int n_location, const int *p_value) = 0;
	virtual bool SetParamMatrix2fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) = 0;
	virtual bool SetParamMatrix3fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) = 0;
	virtual bool SetParamMatrix4fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) = 0;
};

/**
 *	class CGL_ARB_program : public CGLProgram
 *		- class, encapsulating GL_ARB_vertex_program and GL_ARB_fragment_program extensions
 *		- note low-level language doesn't have vertex attrib binding neither uniform locations
 *		  (both are hard-coded in vertex / fragment program source)
 */
class CGL_ARB_program : public CGLProgram {
protected:
	bool m_b_compiled;
	bool m_b_linked;

	unsigned int m_n_vs_object;
	unsigned int m_n_fs_object;

public:
	/**
	 *	CGL_ARB_program::CGL_ARB_program(const char *p_s_vertex_shader,
	 *		const char *p_s_fragment_shader)
	 *		- default constructor
	 *		- p_s_vertex_shader and p_s_fragment_shader are strings, containing shader source
	 *		  codes, one of them can be 0
	 *		- note pointers are copied only, they must not be freed until this object is deleted
	 */
	CGL_ARB_program(const char *p_s_vertex_shader, const char *p_s_fragment_shader);

	/**
	 *	CGL_ARB_program::~CGL_ARB_program()
	 *		- default destructor
	 */
	~CGL_ARB_program();

	/**
	 *	static bool CGL_ARB_program::b_Supported(bool b_vertex_pipeline = true,
	 *		bool b_fragment_pipeline = true)
	 *		- b_vertex_pipeline and b_fragment_pipeline selects required functionality
	 *		- returns true in case all necessary OpenGL extensions are supported,
	 *		  otherwise returns false
	 */
	static bool b_Supported(bool b_vertex_pipeline = true, bool b_fragment_pipeline = true);

	/**
	 *	bool CGL_ARB_program::Compile(CGLState *p_state, char *&r_p_s_info_log)
	 *		- compile program and return info-log, which has to be freed
	 *		- if you re-compile program, it call Delete() first
	 *		- returns true on success, (still can return log with warnings) false on failure
	 */
	virtual bool Compile(CGLState *p_state, char *&r_p_s_info_log);

	/**
	 *	bool CGL_ARB_program::SetParam4fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value) const
	 *		- set float vec4 shader parameter
	 *		- p_value is pointer to 4 floats, data will be read from
	 *		- n_location is index of destination register
	 *		- n_param_type must be one of shader_param_Parameter_Env,
	 *		  shader_param_Parameter_Local_VP or shader_param_Parameter_Local_FP
	 */
	virtual bool SetParam4fv(CGLState *p_state, int n_param_type,
		int n_location, const float *p_value);

	/**
	 *	void CGL_ARB_program::Bind(CGLState *p_state)
	 *		- binds shader
	 */
	virtual void Bind(CGLState *p_state);

	/**
	 *	void CGL_ARB_program::Release(CGLState *p_state)
	 *		- un-binds shader
	 *		- used for safe switching between shaders with different api-s
	 */
	virtual void Release(CGLState *p_state);

	/**
	 *	void CGL_ARB_program::Delete()
	 *		- free open-gl resources
	 *		- if program is about to be used again, it must be re-compiled,
	 *		  re-bound vertex attribs, re-linked and re-queried for unifroms
	 */
	virtual void Delete();

	/**
	 *	int CGL_ARB_program::n_Api() const
	 *		- return which api is used (shader_api_ARB_program)
	 */
	virtual int n_Api() const;

	__inline void _Bind(CGLState *p_state) const;
	__inline void _Release(CGLState *p_state) const;

	// the following function doesn't do any parameter checking, assumes shader is bound
	__inline bool _SetParam4fv(CGLState *p_state, int n_param_type,
		int n_location, const float *p_value) const;
};

/**
 *	class CGL_ARB_shader : public CGLShader
 *		- class, encapsulating GL_ARB_vertex_shader and GL_ARB_fragment_shader extensions
 */
class CGL_ARB_shader : public CGLShader {
protected:
	bool m_b_compiled;
	bool m_b_linked;

	GLhandleARB m_n_program_object;
	GLhandleARB m_n_vs_object;
	GLhandleARB m_n_fs_object;

public:
	/**
	 *	CGL_ARB_shader::CGL_ARB_shader(const char *p_s_vertex_shader,
	 *		const char *p_s_fragment_shader)
	 *		- default constructor
	 *		- p_s_vertex_shader and p_s_fragment_shader are strings, containing shader source
	 *		  codes, one of them can be 0
	 *		- note pointers are copied only, they must not be freed until this object is deleted
	 */
	CGL_ARB_shader(const char *p_s_vertex_shader, const char *p_s_fragment_shader);

	/**
	 *	CGL_ARB_shader::~CGL_ARB_shader()
	 *		- default destructor
	 */
	~CGL_ARB_shader();

	/**
	 *	static bool CGL_ARB_shader::b_Supported(bool b_vertex_pipeline = true,
	 *		bool b_fragment_pipeline = true)
	 *		- b_vertex_pipeline and b_fragment_pipeline selects required functionality
	 *		- returns true in case all necessary OpenGL extensions are supported,
	 *		  otherwise returns false
	 */
	static bool b_Supported(bool b_vertex_pipeline = true, bool b_fragment_pipeline = true);

	/**
	 *	virutal bool CGL_ARB_shader::Compile(CGLState *p_state, char *&r_p_s_info_log)
	 *		- compile program and return info-log, which has to be freed
	 *		- if you re-compile program, it call Delete() first
	 *		- returns true on success (can still return warnings), false on failure
	 */
	virtual bool Compile(CGLState *p_state, char *&r_p_s_info_log);

	/**
	 *	virutal bool CGL_ARB_shader::BindAttribLocation(int n_location, const char *p_s_attrib_name)
	 *		- bind vertex attrib location
	 *		- can be called only if program was succesfuly compiled and not linked
	 *		- returns true on success, false on failure
	 */
	virtual bool BindAttribLocation(int n_location, const char *p_s_attrib_name);

	/**
	 *	virutal bool CGL_ARB_shader::Link(char *&r_p_s_info_log)
	 *		- link program, return info-log which has to be freed
	 *		- can be called only if program was succesfuly compiled and not linked
	 *		- returns true on success (can still return warnings), false on failure
	 */
	virtual bool Link(char *&r_p_s_info_log);

	/**
	 *	virtual int CGL_ARB_shader::n_Uniform_Num() const
	 *		- return number of uniform variables or -1 on failure
	 */
	virtual int n_Uniform_Num() const;

	/**
	 *	virtual char *CGL_ARB_shader::p_s_Uniform_Name(int n_index, int &r_n_content_type) const
	 *		- return name of uniform variable with index n_index or 0 on failure
	 *		- r_n_content_type is set to variable content type (one of shader_param_Float,
	 *		  shader_param_Float2, shader_param_Float3, shader_param_Float4, shader_param_Int,
	 *		  shader_param_Int2, shader_param_Int3, shader_param_Int4, shader_param_Bool,
	 *		  shader_param_Bool2, shader_param_Bool3, shader_param_Bool4, shader_param_Matrix2,
	 *		  shader_param_Matrix3, shader_param_Matrix4, shader_sampler_1D, shader_sampler_2D,
	 *		  shader_sampler_3D, shader_sampler_Cube, shader_sampler_1D_Shadow,
	 *		  shader_sampler_2D_Shadow, shader_sampler_2D_Rect, shader_sampler_2D_Rect_Shadow)
	 */
	virtual char *p_s_Uniform_Name(int n_index, int &r_n_content_type) const;

	/**
	 *	virtual int CGL_ARB_shader::n_GetUniformLocation(const char *p_s_uniform_name) const
	 *		- returns uniform location by uniform name
	 *		- returns -1 on failure
	 */
	virtual int n_GetUniformLocation(const char *p_s_uniform_name) const;

	/**
	 *	virtual bool CGL_ARB_shader::SetParam1fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float uniform to value of first element in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam1fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParam2fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float vec2 uniform to value of first two elements in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam2fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParam3fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float vec3 uniform to value of first three elements in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam3fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParam4fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float vec4 uniform to value of first four elements in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam4fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParam1i(CGLState *p_state, int n_location, int n_value)
	 *		- set integer uniform (sampler) to n_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam1i(CGLState *p_state, int n_location, int n_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParam2iv(CGLState *p_state, int n_location,
	 *		const int *p_value)
	 *		- set integer vec2 uniform to n_value of first two elements in p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam2iv(CGLState *p_state, int n_location, const int *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParam3iv(CGLState *p_state, int n_location,
	 *		const int *p_value)
	 *		- set integer vec3 uniform to n_value of first three elements in p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam3iv(CGLState *p_state, int n_location, const int *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParam4iv(CGLState *p_state, int n_location,
	 *		const int *p_value)
	 *		- set integer vec4 uniform to n_value of first four elements in p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam4iv(CGLState *p_state, int n_location, const int *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParamMatrix2fv(CGLState *p_state, int n_location,
	 *		bool b_transpose, const float *p_value)
	 *		- set 2x2 float matrix uniform to 2x2 matrix form p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParamMatrix2fv(CGLState *p_state, int n_location,
		bool b_transpose, const float *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParamMatrix3fv(CGLState *p_state, int n_location,
	 *		bool b_transpose, const float *p_value)
	 *		- set 3x3 float matrix uniform to 3x3 matrix form p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParamMatrix3fv(CGLState *p_state, int n_location,
		bool b_transpose, const float *p_value);

	/**
	 *	virtual bool CGL_ARB_shader::SetParamMatrix4fv(CGLState *p_state, int n_location,
	 *		bool b_transpose, const float *p_value)
	 *		- set 4x4 float matrix uniform to 4x4 matrix form p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParamMatrix4fv(CGLState *p_state, int n_location,
		bool b_transpose, const float *p_value);

	/**
	 *	virtual void CGL_ARB_shader::Bind(CGLState *p_state)
	 *		- binds shader
	 */
	virtual void Bind(CGLState *p_state);

	/**
	 *	virtual void CGL_ARB_shader::Release(CGLState *p_state)
	 *		- un-binds shader
	 *		- used for safe switching between shaders with different api-s
	 */
	virtual void Release(CGLState *p_state);

	/**
	 *	virtual void CGL_ARB_shader::Delete()
	 *		- free open-gl resources
	 *		- if program is about to be used again, it must be re-compiled,
	 *		  re-bound vertex attribs, re-linked and re-queried for unifroms
	 */
	virtual void Delete();

	/**
	 *	virtual int CGL_ARB_shader::n_Api() const
	 *		- return which api is used (shader_api_ARB_shader)
	 */
	virtual int n_Api() const;

	__inline void _Bind(CGLState *p_state) const;
	__inline void _Release(CGLState *p_state) const;

	// the following functions  doesn't do any parameter checking, assumes shader is bound
	__inline bool _SetParam1fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam2fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam3fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam4fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam1i(CGLState *p_state, int n_location, int n_value) const;
	__inline bool _SetParam2iv(CGLState *p_state, int n_location, const int *p_value) const;
	__inline bool _SetParam3iv(CGLState *p_state, int n_location, const int *p_value) const;
	__inline bool _SetParam4iv(CGLState *p_state, int n_location, const int *p_value) const;
	__inline bool _SetParamMatrix2fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) const;
	__inline bool _SetParamMatrix3fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) const;
	__inline bool _SetParamMatrix4fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) const;
};

/**
 *	class CGL_Core_shader : public CGLShader
 *		- class, encapsulating OpenGL 2.0 high-level vertex and fragment shaders
 */
class CGL_Core_shader : public CGLShader {
protected:
	bool m_b_compiled;
	bool m_b_linked;

	unsigned int m_n_program_object;
	unsigned int m_n_vs_object;
	unsigned int m_n_fs_object;

public:
	/**
	 *	CGL_Core_shader::CGL_Core_shader(const char *p_s_vertex_shader, const char *p_s_fragment_shader)
	 *		- default constructor
	 *		- p_s_vertex_shader and p_s_fragment_shader are strings, containing shader source
	 *		  codes, one of them can be 0
	 *		- note pointers are copied only, they must not be freed until this object is deleted
	 */
	CGL_Core_shader(const char *p_s_vertex_shader, const char *p_s_fragment_shader);

	/**
	 *	CGL_Core_shader::~CGL_Core_shader()
	 *		- default destructor
	 */
	~CGL_Core_shader();

	/**
	 *	static bool CGL_Core_shader::b_Supported(bool b_vertex_pipeline = true,
	 *		bool b_fragment_pipeline = true)
	 *		- b_vertex_pipeline and b_fragment_pipeline selects required functionality
	 *		- returns true in case all necessary OpenGL extensions are supported,
	 *		  otherwise returns false
	 */
	static bool b_Supported(bool b_vertex_pipeline = true, bool b_fragment_pipeline = true);

	/**
	 *	virtual bool CGL_Core_shader::Compile(CGLState *p_state, char *&r_p_s_info_log)
	 *		- compile program and return info-log, which has to be freed
	 *		- if you re-compile program, it call Delete() first
	 *		- returns true on success (can still return warnings), false on failure
	 */
	virtual bool Compile(CGLState *p_state, char *&r_p_s_info_log);

	/**
	 *	virtual bool CGL_Core_shader::BindAttribLocation(int n_location, const char *p_s_attrib_name)
	 *		- bind vertex attrib location
	 *		- can be called only if program was succesfuly compiled and not linked
	 *		- returns true on success, false on failure
	 */
	virtual bool BindAttribLocation(int n_location, const char *p_s_attrib_name);

	/**
	 *	virtual bool CGL_Core_shader::Link(char *&r_p_s_info_log)
	 *		- link program, return info-log which has to be freed
	 *		- can be called only if program was succesfuly compiled and not linked
	 *		- returns true on success (can still return warnings), false on failure
	 */
	virtual bool Link(char *&r_p_s_info_log);

	/**
	 *	virtual int CGL_Core_shader::n_Uniform_Num() const
	 *		- return number of uniform variables or -1 on failure
	 */
	virtual int n_Uniform_Num() const;

	/**
	 *	virtual char *CGL_Core_shader::p_s_Uniform_Name(int n_index, int &r_n_content_type) const
	 *		- return name of uniform variable with index n_index or 0 on failure
	 *		- r_n_content_type is set to variable content type (one of shader_param_Float,
	 *		  shader_param_Float2, shader_param_Float3, shader_param_Float4, shader_param_Int,
	 *		  shader_param_Int2, shader_param_Int3, shader_param_Int4, shader_param_Bool,
	 *		  shader_param_Bool2, shader_param_Bool3, shader_param_Bool4, shader_param_Matrix2,
	 *		  shader_param_Matrix3, shader_param_Matrix4, shader_sampler_1D, shader_sampler_2D,
	 *		  shader_sampler_3D, shader_sampler_Cube, shader_sampler_1D_Shadow,
	 *		  shader_sampler_2D_Shadow, shader_sampler_2D_Rect, shader_sampler_2D_Rect_Shadow)
	 */
	virtual char *p_s_Uniform_Name(int n_index, int &r_n_content_type) const;

	/**
	 *	virtual int CGL_Core_shader::n_GetUniformLocation(const char *p_s_uniform_name) const
	 *		- returns uniform location by uniform name
	 *		- returns -1 on failure
	 */
	virtual int n_GetUniformLocation(const char *p_s_uniform_name) const;

	/**
	 *	virtual bool CGL_Core_shader::SetParam1fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float uniform to value of first element in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam1fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParam2fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float vec2 uniform to value of first two elements in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam2fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParam3fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float vec3 uniform to value of first three element in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam3fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParam4fv(CGLState *p_state, int n_param_type,
	 *		int n_location, const float *p_value)
	 *		- set float vec4 uniform to value of first three element in p_value array
	 *		- n_param_type is safe-check only, it should be shader_param_Uniform
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam4fv(CGLState *p_state, int n_param_type, int n_location,
		const float *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParam1i(CGLState *p_state, int n_location, int n_value)
	 *		- set integer uniform (sampler) to n_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam1i(CGLState *p_state, int n_location, int n_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParam2iv(CGLState *p_state, int n_location,
	 *		const int *p_value)
	 *		- set integer vec2 uniform to n_value of first two elements in p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam2iv(CGLState *p_state, int n_location, const int *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParam3iv(CGLState *p_state, int n_location,
	 *		const int *p_value)
	 *		- set integer vec3 uniform to n_value of first three elements in p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam3iv(CGLState *p_state, int n_location, const int *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParam4iv(CGLState *p_state, int n_location,
	 *		const int *p_value)
	 *		- set integer vec4 uniform to n_value of first four elements in p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParam4iv(CGLState *p_state, int n_location, const int *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParamMatrix2fv(CGLState *p_state, int n_location,
	 *		bool b_transpose, const float *p_value)
	 *		- set 2x2 float matrix uniform to 2x2 matrix form p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParamMatrix2fv(CGLState *p_state, int n_location,
		bool b_transpose, const float *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParamMatrix3fv(CGLState *p_state, int n_location,
	 *		bool b_transpose, const float *p_value)
	 *		- set 3x3 float matrix uniform to 3x3 matrix form p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParamMatrix3fv(CGLState *p_state, int n_location,
		bool b_transpose, const float *p_value);

	/**
	 *	virtual bool CGL_Core_shader::SetParamMatrix4fv(CGLState *p_state, int n_location,
	 *		bool b_transpose, const float *p_value)
	 *		- set 4x4 float matrix uniform to 4x4 matrix form p_value
	 *		- n_location is destination register index
	 *		- returns true on success, false on failure
	 */
	virtual bool SetParamMatrix4fv(CGLState *p_state, int n_location,
		bool b_transpose, const float *p_value);

	/**
	 *	virtual void CGL_Core_shader::Bind(CGLState *p_state)
	 *		- binds shader
	 */
	virtual void Bind(CGLState *p_state);

	/**
	 *	virtual void CGL_Core_shader::Release(CGLState *p_state)
	 *		- un-binds shader
	 *		- use for safe switching between shaders with different api-s
	 */
	virtual void Release(CGLState *p_state);

	/**
	 *	virtual void CGL_Core_shader::Delete()
	 *		- free open-gl resources
	 *		- if program is about to be used again, it must be re-compiled,
	 *		  re-bound vertex attribs, re-linked and re-queried for unifroms
	 */
	virtual void Delete();

	/**
	 *	virtual int CGL_Core_shader::n_Api() const
	 *		- return which api is used (shader_api_Core_shader)
	 */
	virtual int n_Api() const;

	__inline void _Bind(CGLState *p_state) const;
	__inline void _Release(CGLState *p_state) const;

	// the following functions doesn't do any parameter checking, assumes shader is bound
	__inline bool _SetParam1fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam2fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam3fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam4fv(CGLState *p_state, /*int n_param_type,*/ int n_location, const float *p_value) const;
	__inline bool _SetParam1i(CGLState *p_state, int n_location, int n_value) const;
	__inline bool _SetParam2iv(CGLState *p_state, int n_location, const int *p_value) const;
	__inline bool _SetParam3iv(CGLState *p_state, int n_location, const int *p_value) const;
	__inline bool _SetParam4iv(CGLState *p_state, int n_location, const int *p_value) const;
	__inline bool _SetParamMatrix2fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) const;
	__inline bool _SetParamMatrix3fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) const;
	__inline bool _SetParamMatrix4fv(CGLState *p_state, int n_location, bool b_transpose, const float *p_value) const;
};

#include "Shader2.inl"
// fast inline functions

#endif //__GL_SHADER2_INCLUDED
