/*
								+---------------------------------+
								|                                 |
								|      ***   OpenGL2.0   ***      |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Texture.h            |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 01-11-2006
 *
 *	bool CGLTexture::b_State() const renamed to bool CGLTexture::b_Status() const
 *	(want to have unified naming over the whole OpenGL framework)
 *
 *	passed code revision at 02-05-2007
 *
 *	"static GLenum CGLTexture_Cube::p_cube_face[6]" is now
 *	"static const GLenum CGLTexture_Cube::p_cube_face[6]"
 *
 *	05-10-2007
 *
 *	added CGLState::NotifyDeleteTexture() to texture destructor to prevent
 *	"unbindable" textures with id of currently bound texture which was destroyed
 *
 */

#ifndef __TEXTURE_MANAGER_INCLUDED
#define __TEXTURE_MANAGER_INCLUDED

enum {
	coord_S = 0,
	coord_T = 1,
	coord_R = 2
};

/**
 *	class CGLTextureParams
 *		- class for guarding texture parameters
 */
class CGLTextureParams {
protected:
	GLenum m_n_wrap_mode[3]; // s, t, r
	GLenum m_n_minify_filter;
	GLenum m_n_magnify_filter;
	float m_p_border_color[4];
	float m_f_priority;
	float m_f_min_lod, m_f_max_lod;
	float m_f_lod_bias;
	int m_n_base_level, m_n_max_level;
	GLenum m_n_depth_mode;
	GLenum m_n_compare_mode, m_n_compare_func;
	bool m_b_generate_mipmap;

	static const float m_f_compare_epsilon/* = 1e-3f*/; // comparison epsilon

public:
	/**
	 *	CGLTextureParams::CGLTextureParams()
	 *		- default constructor
	 *		- fill itself with default texture parameters
	 */
	CGLTextureParams();

	/**
	 *	void CGLTextureParams::UpdateState(GLenum n_target = GL_TEXTURE_2D)
	 *		- get texture parameters from OpenGL
	 *		- note texture of type n_target must be bound and enabled
	 *		- OpenGL error(s) may be generated, call glGetError() afterwards
	 */
	void UpdateState(GLenum n_target = GL_TEXTURE_2D);

	/**
	 *	void CGLTextureParams::SetState(GLenum n_target = GL_TEXTURE_2D) const
	 *		- set texture parameters to OpenGL
	 *		- note texture of type n_target must be bound and enabled
	 *		- OpenGL error(s) may be generated, call glGetError() afterwards
	 */
	void SetState(GLenum n_target = GL_TEXTURE_2D) const;

	/**
	 *	void CGLTextureParams::SetState(GLenum n_target, const CGLTextureParams &r_state) const
	 *		- set texture parameters different from corresponding parameters in r_state to OpenGL
	 *		  (effect is the same as calling r_state.SetState() but equal values aren't set
	 *		  so OpenGL call overhead is lower which should result in better performance)
	 *		- note floating-point number comparison is done with epsilon = 1e-3
	 *		- note texture of type n_target must be bound and enabled
	 *		- OpenGL error(s) may be generated, call glGetError() afterwards
	 */
	void SetState(GLenum n_target, const CGLTextureParams &r_state) const;

	/**
	 *	bool CGLTextureParams::operator ==(const CGLTextureParams &r_state) const
	 *		- compare two texture parameters objects, return true if equal, otherwise false
	 *		- does no changes to OpenGL state, needs no particular state nor generates any errors
	 *		- note floating-point number comparison is done with epsilon = 1e-3
	 */
	bool operator ==(const CGLTextureParams &r_state) const;

	__inline GLenum n_Texture_Wrap_S() const { return m_n_wrap_mode[coord_S]; }
	__inline GLenum n_Texture_Wrap_T() const { return m_n_wrap_mode[coord_T]; }
	__inline GLenum n_Texture_Wrap_R() const { return m_n_wrap_mode[coord_R]; }
	__inline GLenum n_Texture_Min_Filter() const { return m_n_minify_filter; }
	__inline GLenum n_Texture_Mag_Filter() const { return m_n_magnify_filter; }
	__inline const float *p_Texture_Border_Color() const { return m_p_border_color; }
	__inline float f_Texture_Priority() const { return m_f_priority; }
	__inline float f_Texture_Min_LOD() const { return m_f_min_lod; }
	__inline float f_Texture_Max_LOD() const { return m_f_max_lod; }
	__inline int n_Texture_Base_Level() const { return m_n_base_level; }
	__inline int n_Texture_Max_Level() const { return m_n_max_level; }
	__inline float f_Texture_LOD_Bias() const { return m_f_lod_bias; }
	__inline GLenum n_Depth_Texture_Mode() const { return m_n_depth_mode; }
	__inline GLenum n_Texture_Compare_Mode() const { return m_n_compare_mode; }
	__inline GLenum n_Texture_Compare_Func() const { return m_n_compare_func; }
	__inline bool b_Generate_Mipmap() const { return m_b_generate_mipmap; }

	__inline GLenum &n_Texture_Wrap_S() { return m_n_wrap_mode[coord_S]; }
	__inline GLenum &n_Texture_Wrap_T() { return m_n_wrap_mode[coord_T]; }
	__inline GLenum &n_Texture_Wrap_R() { return m_n_wrap_mode[coord_R]; }
	__inline GLenum &n_Texture_Min_Filter() { return m_n_minify_filter; }
	__inline GLenum &n_Texture_Mag_Filter() { return m_n_magnify_filter; }
	__inline float *p_Texture_Border_Color() { return m_p_border_color; }
	__inline float &f_Texture_Priority() { return m_f_priority; }
	__inline float &f_Texture_Min_LOD() { return m_f_min_lod; }
	__inline float &f_Texture_Max_LOD() { return m_f_max_lod; }
	__inline int &n_Texture_Base_Level() { return m_n_base_level; }
	__inline int &n_Texture_Max_Level() { return m_n_max_level; }
	__inline float &f_Texture_LOD_Bias() { return m_f_lod_bias; }
	__inline GLenum &n_Depth_Texture_Mode() { return m_n_depth_mode; }
	__inline GLenum &n_Texture_Compare_Mode() { return m_n_compare_mode; }
	__inline GLenum &n_Texture_Compare_Func() { return m_n_compare_func; }
	__inline bool &b_Generate_Mipmap() { return m_b_generate_mipmap; }

	__inline void Set_Texture_Wrap_S(GLenum n_wrap_mode, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_WRAP_S, n_wrap_mode); m_n_wrap_mode[coord_S] = n_wrap_mode; }
	__inline void Set_Texture_Wrap_T(GLenum n_wrap_mode, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_WRAP_T, n_wrap_mode); m_n_wrap_mode[coord_T] = n_wrap_mode; }
	__inline void Set_Texture_Wrap_R(GLenum n_wrap_mode, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_WRAP_R, n_wrap_mode); m_n_wrap_mode[coord_R] = n_wrap_mode; }
	__inline void Set_Texture_Min_Filter(GLenum n_minify_filter, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_MIN_FILTER, n_minify_filter); m_n_minify_filter = n_minify_filter; }
	__inline void Set_Texture_Mag_Filter(GLenum n_magnify_filter, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_MAG_FILTER, n_magnify_filter); m_n_magnify_filter = n_magnify_filter; }
	__inline void Set_Texture_Border_Color(const float *p_border_color, GLenum n_target = GL_TEXTURE_2D) { glTexParameterfv(n_target, GL_TEXTURE_BORDER_COLOR, p_border_color); memcpy(m_p_border_color, p_border_color, 4 * sizeof(float)); }
	__inline void Set_Texture_Priority(float f_priority, GLenum n_target = GL_TEXTURE_2D) { glTexParameterf(n_target, GL_TEXTURE_PRIORITY, f_priority); m_f_priority = f_priority; }
	__inline void Set_Texture_Min_LOD(float f_min_lod, GLenum n_target = GL_TEXTURE_2D) { glTexParameterf(n_target, GL_TEXTURE_MIN_LOD, f_min_lod); m_f_min_lod = f_min_lod; }
	__inline void Set_Texture_Max_LOD(float f_max_lod, GLenum n_target = GL_TEXTURE_2D) { glTexParameterf(n_target, GL_TEXTURE_MAX_LOD, f_max_lod); m_f_max_lod = f_max_lod; }
	__inline void Set_Texture_Base_Level(int n_base_level, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_BASE_LEVEL, n_base_level); m_n_base_level = n_base_level; }
	__inline void Set_Texture_Max_Level(int n_max_level, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_MAX_LEVEL, n_max_level); m_n_max_level = n_max_level; }
	__inline void Set_Texture_LOD_Bias(float f_lod_bias, GLenum n_target = GL_TEXTURE_2D) { glTexParameterf(n_target, GL_TEXTURE_LOD_BIAS, f_lod_bias); m_f_lod_bias = f_lod_bias; }
	__inline void Set_Depth_Texture_Mode(GLenum n_depth_mode, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_DEPTH_TEXTURE_MODE, n_depth_mode); m_n_depth_mode = n_depth_mode; }
	__inline void Set_Texture_Compare_Mode(GLenum n_compare_mode, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_COMPARE_MODE, n_compare_mode); m_n_compare_mode = n_compare_mode; }
	__inline void Set_Texture_Compare_Func(GLenum n_compare_func, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_TEXTURE_COMPARE_FUNC, n_compare_func); m_n_compare_func = n_compare_func; }
	__inline void Set_Generate_Mipmap(bool b_generate_mipmap, GLenum n_target = GL_TEXTURE_2D) { glTexParameteri(n_target, GL_GENERATE_MIPMAP, (b_generate_mipmap)? GL_TRUE : GL_FALSE); m_b_generate_mipmap = b_generate_mipmap; }
};

class CGLTexture {
protected:
	GLuint m_n_id;
	GLenum m_n_internal_format; // GL_RGB8
	GLenum m_n_target; // GL_TEXTURE_2D
	int m_n_border_width;
	CGLTextureParams m_tex_parameters;

public:
	/**
	 *	CGLTexture::CGLTexture()
	 *		- default constructor
	 *		- creates no texture object
	 */
	CGLTexture();

	/**
	 *	CGLTexture::~CGLTexture()
	 *		- default destructor
	 *		- deletes OpenGL texture object in case it was created
	 */
	~CGLTexture();

	/**
	 *	bool CGLTexture::b_Status() const
	 *		- returns true in case texture was successfuly created and can be used for texturing
	 */
	bool b_Status() const;

	/**
	 *	void CGLTexture::Bind(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 */
	virtual void Bind(CGLState *p_state) const;

	/**
	 *	void CGLTexture::Bind_Enable(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 *		- enables application of texture target (i.e. GL_TEXTURE_2D or similar)
	 */
	virtual void Bind_Enable(CGLState *p_state) const;

	/**
	 *	void CGLTexture::Disable(CGLState *p_state) const
	 *		- disables application of texture target (i.e. GL_TEXTURE_2D or similar)
	 */
	virtual void Disable(CGLState *p_state) const;

	__inline GLenum n_Pixel_Format() const { return m_n_internal_format; }
	__inline GLenum n_Target() const { return m_n_target; }
	__inline int n_Border_Width() const { return m_n_border_width; }
	__inline const CGLTextureParams &t_Parameters() const { return m_tex_parameters; }

	/**
	 *	CGLTextureParams &CGLTexture::r_Parameters(CGLState *p_state)
	 *		- enables texture parameters to be changed
	 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
	 */
	virtual CGLTextureParams &r_Parameters(CGLState *p_state);

private:
	friend class CGLRenderBuffer_FBO;

	/**
	 *	__inline GLuint n_Id() const
	 *		- for use within CGLTexture_Friend inheritants
	 *		- do not free the texture nor change it's parameters, purpose is
	 *		  to enable access to texture id to functions needing it (such as
	 *		  framebuffer object texture binding, etc)
	 */
	__inline GLuint n_Id() const { return m_n_id; }
};

class CGLTexture_1D : public CGLTexture {
protected:
	int m_n_width;

public:
	CGLTexture_1D(CGLState *p_state, int n_width, GLenum n_internal_format,
		bool b_create_mipmaps = true, int n_border_width = 0, GLenum n_target = GL_TEXTURE_1D,
		GLenum n_format = GL_RGB, GLenum n_data_type = GL_UNSIGNED_BYTE,
		const void *p_src_data = 0);

	__inline int n_Width() const { return m_n_width; }

	/**
	 *	void CGLTexture_1D::Bind(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 */
	virtual void Bind(CGLState *p_state) const;

	/**
	 *	void CGLTexture_1D::Bind_Enable(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 *		- enables application of texture target (i.e. GL_TEXTURE_1D or similar)
	 */
	virtual void Bind_Enable(CGLState *p_state) const;

	/**
	 *	void CGLTexture_1D::Disable(CGLState *p_state) const
	 *		- disables application of texture target (i.e. GL_TEXTURE_1D or similar)
	 */
	virtual void Disable(CGLState *p_state) const;

	/**
	 *	CGLTextureParams &CGLTexture_1D::r_Parameters(CGLState *p_state)
	 *		- enables texture parameters to be changed
	 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
	 */
	virtual CGLTextureParams &r_Parameters(CGLState *p_state);
};

class CGLTexture_2D : public CGLTexture {
protected:
	int m_n_width;
	int m_n_height;

public:
	CGLTexture_2D(CGLState *p_state, int n_width, int n_height, GLenum n_internal_format,
		bool b_create_mipmaps = true, int n_border_width = 0, GLenum n_target = GL_TEXTURE_2D,
		GLenum n_format = GL_RGB, GLenum n_data_type = GL_UNSIGNED_BYTE,
		const void *p_src_data = 0);

	__inline int n_Width() const { return m_n_width; }
	__inline int n_Height() const { return m_n_height; }

	/**
	 *	void CGLTexture_2D::Bind(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 */
	virtual void Bind(CGLState *p_state) const;

	/**
	 *	void CGLTexture_2D::Bind_Enable(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 *		- enables application of texture target (i.e. GL_TEXTURE_2D or similar)
	 */
	virtual void Bind_Enable(CGLState *p_state) const;

	/**
	 *	void CGLTexture_2D::Disable(CGLState *p_state) const
	 *		- disables application of texture target (i.e. GL_TEXTURE_2D or similar)
	 */
	virtual void Disable(CGLState *p_state) const;

	/**
	 *	CGLTextureParams &CGLTexture_2D::r_Parameters(CGLState *p_state)
	 *		- enables texture parameters to be changed
	 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
	 */
	virtual CGLTextureParams &r_Parameters(CGLState *p_state);
};

class CGLTexture_3D : public CGLTexture {
protected:
	int m_n_width;
	int m_n_height;
	int m_n_depth;

public:
	CGLTexture_3D(CGLState *p_state, int n_width, int n_height, int n_depth,
		GLenum n_internal_format, bool b_create_mipmaps = true, int n_border_width = 0,
		GLenum n_target = GL_TEXTURE_3D, GLenum n_format = GL_RGB,
		GLenum n_data_type = GL_UNSIGNED_BYTE, const void *p_src_data = 0);

	__inline int n_Width() const { return m_n_width; }
	__inline int n_Height() const { return m_n_height; }
	__inline int n_Depth() const { return m_n_depth; }

	/**
	 *	void CGLTexture_3D::Bind(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 */
	virtual void Bind(CGLState *p_state) const;

	/**
	 *	void CGLTexture_3D::Bind_Enable(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 *		- enables application of texture target (i.e. GL_TEXTURE_3D or similar)
	 */
	virtual void Bind_Enable(CGLState *p_state) const;

	/**
	 *	void CGLTexture_3D::Disable(CGLState *p_state) const
	 *		- disables application of texture target (i.e. GL_TEXTURE_3D or similar)
	 */
	virtual void Disable(CGLState *p_state) const;

	/**
	 *	CGLTextureParams &CGLTexture_3D::r_Parameters(CGLState *p_state)
	 *		- enables texture parameters to be changed
	 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
	 */
	virtual CGLTextureParams &r_Parameters(CGLState *p_state);
};

class CGLTexture_Cube : public CGLTexture {
protected:
	static const GLenum p_cube_face[6];
	int m_n_width;

public:
	CGLTexture_Cube(CGLState *p_state, int n_width, GLenum n_internal_format,
		bool b_create_mipmaps = true, bool b_software_mipmaps = false, int n_border_width = 0,
		GLenum n_target = GL_TEXTURE_CUBE_MAP, GLenum n_format = GL_RGB,
		GLenum n_data_type = GL_UNSIGNED_BYTE, const void *p_src_data[6] = 0);

	__inline int n_Width() const { return m_n_width; }
	__inline int n_Height() const { return m_n_width; }
	__inline int n_Depth() const { return m_n_width; }

	/**
	 *	void CGLTexture_Cube::Bind(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 */
	virtual void Bind(CGLState *p_state) const;

	/**
	 *	void CGLTexture_Cube::Bind_Enable(CGLState *p_state) const
	 *		- binds texture object to OpenGL state as active texture in current texture unit
	 *		- enables application of texture target (i.e. GL_TEXTURE_CUBE_MAP or similar)
	 */
	virtual void Bind_Enable(CGLState *p_state) const;

	/**
	 *	void CGLTexture_Cube::Disable(CGLState *p_state) const
	 *		- disables application of texture target (i.e. GL_TEXTURE_CUBE_MAP or similar)
	 */
	virtual void Disable(CGLState *p_state) const;

	/**
	 *	CGLTextureParams &CGLTexture_Cube::r_Parameters(CGLState *p_state)
	 *		- enables texture parameters to be changed
	 *		- binds and enables texture first (necessary for OpenGL texture parameter access)
	 */
	virtual CGLTextureParams &r_Parameters(CGLState *p_state);
};

#endif //__TEXTURE_MANAGER_INCLUDED
