/*
								+----------------------------------+
								|                                  |
								|  ***  Reference rasterizer  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|          RefRaster.cpp           |
								|                                  |
								+----------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "PixelConv.h"
#include "RefRaster.h"

/**
 *	inline unsigned char clampb(int n_value)
 *		- returns n_value, saturated into range 0 trough 255
 */
inline unsigned char clampb(int n_value)
{
	return (n_value >= 0)? ((n_value <= 0xff)? (unsigned char)n_value : 0xff) : 0;
}

/**
 *	bool rrImageCopy(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- copies source image to dest image. converts pixel types and
 *		  has free bilinear resampling, same as all the others
 *		- returns true on success, false on failure
 */
bool rrImageCopy(ImageStruct *p_dest, const ImageStruct *p_src)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	if(!b_src_resample && p_dest->PixelType == p_src->PixelType &&
	   p_dest->XOffset == p_src->XOffset && p_dest->YOffset == p_src->YOffset) {
		const unsigned char *p_src_raster = p_src->Raster;
		unsigned char *p_dest_raster = p_dest->Raster;
		unsigned int n_image_size = abs(p_src->YOffset) * p_src->YSize;
		if(p_src->YOffset < 0) {
			p_src_raster += p_src->YOffset * (p_src->YSize - 1);
			p_dest_raster += p_src->YOffset * (p_src->YSize - 1);
		}
		if(p_src->XOffset < 0) {
			p_src_raster += p_src->XOffset * (p_src->XSize - 1);
			p_dest_raster += p_src->XOffset * (p_src->XSize - 1);
		}
		memcpy(p_dest_raster, p_src_raster, n_image_size);
		return true;
	}
	// in case absolutely same formats, just memcpy

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, GetPixel8(p_src, x0, y0));
				// fast gray
			} else if(n_path == 1) {
				SetPixelRGB(p_dest, x, y, GetPixelRGB(p_src, x0, y0));
				// fast RGB
			} else {
				float p_pix[4];
				GetPixel4f(p_src, x0, y0, p_pix);
				SetPixel4f(p_dest, x, y, p_pix);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageImageAdd(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image addition point operator
 */
bool rrImageImageAdd(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, clampb((int)GetPixel8(p_src0, x0, y0) +
					(int)GetPixel8(p_src1, x1, y1)));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					clampb(((n_pix0 >> 16) & 0xff) + ((n_pix1 >> 16) & 0xff)),
					clampb(((n_pix0 >> 8) & 0xff) + ((n_pix1 >> 8) & 0xff)),
					clampb((n_pix0 & 0xff) + (n_pix1 & 0xff)),
					clampb((n_pix0 >> 24) + (n_pix1 >> 24))));
				// fast RGB
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				p_pix0[0] += p_pix1[0];
				p_pix0[1] += p_pix1[1];
				p_pix0[2] += p_pix1[2];
				p_pix0[3] += p_pix1[3];
				SetPixel4f(p_dest, x, y, p_pix0);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageImageSub(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image subtraction point operator
 */
bool rrImageImageSub(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, clampb((int)GetPixel8(p_src0, x0, y0) -
					(int)GetPixel8(p_src1, x1, y1)));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					clampb(((n_pix0 >> 16) & 0xff) - ((n_pix1 >> 16) & 0xff)),
					clampb(((n_pix0 >> 8) & 0xff) - ((n_pix1 >> 8) & 0xff)),
					clampb((n_pix0 & 0xff) - (n_pix1 & 0xff)),
					clampb((n_pix0 >> 24) - (n_pix1 >> 24))));
				// fast RGB
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				p_pix0[0] -= p_pix1[0];
				p_pix0[1] -= p_pix1[1];
				p_pix0[2] -= p_pix1[2];
				p_pix0[3] -= p_pix1[3];
				SetPixel4f(p_dest, x, y, p_pix0);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

inline ImageComplexValue t_CmplxMul(ImageComplexValue x, ImageComplexValue y)
{
	ImageComplexValue t_tmp = {x.Re * y.Re - x.Im * y.Im, x.Im * y.Re + x.Re * y.Im};
	return t_tmp;
}

inline ImageComplexValue t_CmplxMulf(float x, ImageComplexValue y)
{
	ImageComplexValue t_tmp = {x * y.Re, x * y.Im};
	return t_tmp;
}

/**
 *	bool rrImageImageMult(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image multiplication point operator
 */
bool rrImageImageMult(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	if((p_src0->PixelType == ImageComplex || p_src0->PixelType == ImageComplexLinear) &&
	   p_src1->PixelType != ImageComplex && p_src1->PixelType != ImageComplexLinear) {
		const ImageStruct *p_tmp = p_src0;
		p_src0 = p_src1;
		p_src1 = p_tmp;
	}
	// prefer having complex type in p_src1 (less branches leading to smaller
	// code which better fits cache and is faster)

	char n_out_type = 2;
	if(p_dest->PixelType == ImageComplex || p_dest->PixelType == ImageComplexLinear)
		n_out_type = 0;
	else if(p_dest->PixelType == ImageFloat || p_dest->PixelType == ImageFloatLinear)
		n_out_type = 1;
	// complex, float or other

	char n_path = 4;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	else if((p_src0->PixelType == ImageComplex || p_src0->PixelType == ImageComplexLinear) &&
	   (p_src1->PixelType == ImageComplex || p_src1->PixelType == ImageComplexLinear))
		n_path = 2;
	else if((p_src0->PixelType != ImageComplex || p_src0->PixelType != ImageComplexLinear) &&
	   (p_src1->PixelType == ImageComplex || p_src1->PixelType == ImageComplexLinear))
		n_path = 3;
	// processing path (0, 1 are optimized real, 2 is cmplx * cmplx, 3 is re * cmplx)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, clampb((int)GetPixel8(p_src0, x0, y0) *
					(int)GetPixel8(p_src1, x1, y1)));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					clampb(((n_pix0 >> 16) & 0xff) * ((n_pix1 >> 16) & 0xff)),
					clampb(((n_pix0 >> 8) & 0xff) * ((n_pix1 >> 8) & 0xff)),
					clampb((n_pix0 & 0xff) * (n_pix1 & 0xff)),
					clampb((n_pix0 >> 24) * (n_pix1 >> 24))));
				// fast RGB
			} else if(n_path == 2) {
				ImageComplexValue t_src0 = GetPixelComplex(p_src0, x0, y0),
					t_src1 = GetPixelComplex(p_src1, x1, y1);
				if(!n_out_type) {
					ImageComplexValue t_dest = t_CmplxMul(t_src0, t_src1);
					SetPixelComplex(p_dest, x, y, t_dest);
				} else if(n_out_type == 1)
					SetPixelFloat(p_dest, x, y, t_src0.Re * t_src1.Re - t_src0.Im * t_src1.Im);
				else {
					ImageComplexValue t_dest = t_CmplxMul(t_src0, t_src1);
					float p_pix[4] = {t_dest.Re, t_dest.Re, t_dest.Re, t_dest.Im};
					SetPixel4f(p_dest, x, y, p_pix);
				}
				// complex by complex into complex, float or non-complex
			} else if(n_path == 3) {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				ImageComplexValue t_src1 = GetPixelComplex(p_src1, x1, y1);
				if(!n_out_type) {
					ImageComplexValue t_dest = t_CmplxMulf(p_pix0[0], t_src1);
					SetPixelComplex(p_dest, x, y, t_dest);
				} else if(n_out_type == 1)
					SetPixelFloat(p_dest, x, y, p_pix0[0] * t_src1.Re);
				else {
					ImageComplexValue t_dest = t_CmplxMulf(p_pix0[0], t_src1);
					float p_pix[4] = {t_dest.Re, t_dest.Re, t_dest.Re, t_dest.Im};
					SetPixel4f(p_dest, x, y, p_pix);
				}
				// real by complex into complex, float or non-complex
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				p_pix0[0] *= p_pix1[0];
				p_pix0[1] *= p_pix1[1];
				p_pix0[2] *= p_pix1[2];
				p_pix0[3] *= p_pix1[3];
				SetPixel4f(p_dest, x, y, p_pix0);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

inline unsigned char divideb(int a, int b)
{
	if(b)
		return clampb(a / b);
	else if(a)
		return 0xff; // 1.#INF00
	return 0; // 1.#NaN00
}

inline ImageComplexValue t_CmplxDiv(ImageComplexValue x, ImageComplexValue y)
{
/*
	re1 + jim1   re2 - jim2   re1 * re2 + im1 * im2 + j(im1 * re2 - re1 * im2)
	---------- * ---------- = ------------------------------------------------
	re2 + jim2   re2 - jim2             re2 * re2 + im2 * im2
*/
	float f = 1 / (y.Re * y.Re + y.Im * y.Im);
	ImageComplexValue t_tmp = {f * (x.Re * y.Re + x.Im * y.Im), f * (x.Im * y.Re - x.Re * y.Im)};
	return t_tmp;
}

inline ImageComplexValue t_CmplxDivf(float x, ImageComplexValue y)
{
/*
	re1 + jim1   re2 - jim2   re1 * re2 + 0 * im2 + j(0 * re2 - re1 * im2)
	---------- * ---------- = --------------------------------------------
	re2 + jim2   re2 - jim2             re2 * re2 + im2 * im2
*/
	float f = 1 / (y.Re * y.Re + y.Im * y.Im);
	ImageComplexValue t_tmp = {f * (x * y.Re), f * (- x * y.Im)};
	return t_tmp;
}

inline ImageComplexValue t_CmplxDivf2(ImageComplexValue x, float y)
{
/*
	re1 + jim1   re2 - jim2   re1 * re2 + 0 * im2 + j(0 * re2 - re1 * im2)
	---------- * ---------- = --------------------------------------------
	re2 + jim2   re2 - jim2             re2 * re2 + im2 * im2
*/
	float f = 1 / y;
	ImageComplexValue t_tmp = {f * x.Re, f * x.Im};
	return t_tmp;
}

/**
 *	bool rrImageImageDiv(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- image division point operator
 */
bool rrImageImageDiv(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_out_type = 2;
	if(p_dest->PixelType == ImageComplex || p_dest->PixelType == ImageComplexLinear)
		n_out_type = 0;
	else if(p_dest->PixelType == ImageFloat || p_dest->PixelType == ImageFloatLinear)
		n_out_type = 1;
	// complex, float or other

	char n_path = 5;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	else if((p_src0->PixelType == ImageComplex || p_src0->PixelType == ImageComplexLinear) &&
	   (p_src1->PixelType == ImageComplex || p_src1->PixelType == ImageComplexLinear))
		n_path = 2;
	else if((p_src0->PixelType != ImageComplex || p_src0->PixelType != ImageComplexLinear) &&
	   (p_src1->PixelType == ImageComplex || p_src1->PixelType == ImageComplexLinear))
		n_path = 3;
	else if((p_src0->PixelType == ImageComplex || p_src0->PixelType == ImageComplexLinear) &&
	   (p_src1->PixelType != ImageComplex || p_src1->PixelType != ImageComplexLinear))
		n_path = 4;
	// processing path (0, 1 are optimized real,
	// 2 is cmplx * cmplx, 3 is re * cmplx, 4 is cmplx * re)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, divideb(GetPixel8(p_src0, x0, y0),
					GetPixel8(p_src1, x1, y1)));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					divideb(((n_pix0 >> 16) & 0xff), ((n_pix1 >> 16) & 0xff)),
					divideb(((n_pix0 >> 8) & 0xff), ((n_pix1 >> 8) & 0xff)),
					divideb((n_pix0 & 0xff), (n_pix1 & 0xff)),
					divideb((n_pix0 >> 24), (n_pix1 >> 24))));
				// fast RGB
			} else if(n_path == 2) {
				ImageComplexValue t_src0 = GetPixelComplex(p_src0, x0, y0),
					t_src1 = GetPixelComplex(p_src1, x1, y1);
				if(!n_out_type) {
					ImageComplexValue t_dest = t_CmplxDiv(t_src0, t_src1);
					SetPixelComplex(p_dest, x, y, t_dest);
				} else if(n_out_type == 1) {
					SetPixelFloat(p_dest, x, y, (t_src0.Re * t_src1.Re + t_src0.Im * t_src1.Im) /
						(t_src1.Re * t_src1.Re + t_src1.Im * t_src1.Im));
				} else {
					ImageComplexValue t_dest = t_CmplxDiv(t_src0, t_src1);
					float p_pix[4] = {t_dest.Re, t_dest.Re, t_dest.Re, t_dest.Im};
					SetPixel4f(p_dest, x, y, p_pix);
				}
				// complex by complex into complex, float or non-complex
			} else if(n_path == 3) {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				ImageComplexValue t_src1 = GetPixelComplex(p_src1, x1, y1);
				if(!n_out_type) {
					ImageComplexValue t_dest = t_CmplxDivf(p_pix0[0], t_src1);
					SetPixelComplex(p_dest, x, y, t_dest);
				} else if(n_out_type == 1) {
					SetPixelFloat(p_dest, x, y, (p_pix0[0] * t_src1.Re) /
						(t_src1.Re * t_src1.Re + t_src1.Im * t_src1.Im));
				} else {
					ImageComplexValue t_dest = t_CmplxDivf(p_pix0[0], t_src1);
					float p_pix[4] = {t_dest.Re, t_dest.Re, t_dest.Re, t_dest.Im};
					SetPixel4f(p_dest, x, y, p_pix);
				}
				// real by complex into complex, float or non-complex
			} else if(n_path == 4) {
				ImageComplexValue t_src0 = GetPixelComplex(p_src0, x0, y0);
				float p_pix1[4];
				GetPixel4f(p_src1, x1, y1, p_pix1);
				if(!n_out_type) {
					ImageComplexValue t_dest = t_CmplxDivf2(t_src0, p_pix1[0]);
					SetPixelComplex(p_dest, x, y, t_dest);
				} else if(n_out_type == 1)
					SetPixelFloat(p_dest, x, y, t_src0.Re / p_pix1[0]);
				else {
					ImageComplexValue t_dest = t_CmplxDivf2(t_src0, p_pix1[0]);
					float p_pix[4] = {t_dest.Re, t_dest.Re, t_dest.Re, t_dest.Im};
					SetPixel4f(p_dest, x, y, p_pix);
				}
				// complex by real into complex, float or non-complex
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				p_pix0[0] /= p_pix1[0];
				p_pix0[1] /= p_pix1[1];
				p_pix0[2] /= p_pix1[2];
				p_pix0[3] /= p_pix1[3];
				SetPixel4f(p_dest, x, y, p_pix0);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

inline float minf(float a, float b)
{
	return (a > b)? b : a;
}

inline float maxf(float a, float b)
{
	return (a > b)? a : b;
}

/**
 *	bool rrImageImageBlend(ImageStruct *p_dest, const ImageStruct *p_src0,
 *		const ImageStruct *p_src1, float f_weight0, float f_weight1)
 *		- calculates following equation: p_src0 * f_weight0 + p_src1 * f_weight1
 *		- weight magnitudes are unlimited, results are clamped based on dest image pixelformat
 */
bool rrImageImageBlend(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1, float f_weight0, float f_weight1)
{
	char n_path = 4;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
	    n_path = (f_weight0 >= 0 && f_weight0 <= 1 && f_weight1 >= 0 && f_weight1 <= 1)? 0 : 2;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
	    n_path = (f_weight0 >= 0 && f_weight0 <= 1 && f_weight1 >= 0 && f_weight1 <= 1)? 1 : 3;
	// processing path (0, 1 / 2, 3 are optimized, 4 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	unsigned char n_weight0 = (unsigned char)minf(maxf(f_weight0 * 255, 0), 255);
	unsigned char n_weight1 = (unsigned char)minf(maxf(f_weight1 * 255, 0), 255);
	// fixed-point weights

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, clampb(((int)GetPixel8(p_src0, x0, y0) * n_weight0 +
					(int)GetPixel8(p_src1, x1, y1) * n_weight1) >> 8));
				// fast fixed-point gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					clampb((((n_pix0 >> 16) & 0xff) * n_weight0 + ((n_pix1 >> 16) & 0xff) * n_weight1) >> 8),
					clampb((((n_pix0 >> 8) & 0xff) * n_weight0 + ((n_pix1 >> 8) & 0xff) * n_weight1) >> 8),
					clampb(((n_pix0 & 0xff) * n_weight0 + (n_pix1 & 0xff) * n_weight1) >> 8),
					clampb(((n_pix0 >> 24) * n_weight0 + (n_pix1 >> 24) * n_weight1) >> 8)));
				// fast fixed-point RGB
			} else if(n_path == 2) {
				SetPixel8(p_dest, x, y, (unsigned char)minf(255, maxf(0, GetPixel8(p_src0, x0, y0) * f_weight0 +
					GetPixel8(p_src1, x1, y1) * f_weight1)));
				// slower gray
			} else if(n_path == 3) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					(unsigned char)minf(255, maxf(0, ((n_pix0 >> 16) & 0xff) * f_weight0 + ((n_pix1 >> 16) & 0xff) * f_weight1)),
					(unsigned char)minf(255, maxf(0, ((n_pix0 >> 8) & 0xff) * f_weight0 + ((n_pix1 >> 8) & 0xff) * f_weight1)),
					(unsigned char)minf(255, maxf(0, (n_pix0 & 0xff) * f_weight0 + (n_pix1 & 0xff) * f_weight1)),
					(unsigned char)minf(255, maxf(0, (n_pix0 >> 24) * f_weight0 + (n_pix1 >> 24) * f_weight1))));
				// slower RGB
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				float p_pixd[4];
				p_pixd[0] = p_pix0[0] * f_weight0 + p_pix1[0] * f_weight1;
				p_pixd[1] = p_pix0[1] * f_weight0 + p_pix1[1] * f_weight1;
				p_pixd[2] = p_pix0[2] * f_weight0 + p_pix1[2] * f_weight1;
				p_pixd[3] = p_pix0[3] * f_weight0 + p_pix1[3] * f_weight1;
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageImageDecal(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- alpha-blends source image 1 onto source image 0, ie:
 *		  p_src0 * (1 - p_src1_alpha) + p_src1 * p_src1_alpha
 */
bool rrImageImageDecal(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 1;
	if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 0;
	// processing path (0 is optimized, 1 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				unsigned char n_weight1 = (unsigned char)(n_pix1 >> 24);
				unsigned char n_weight0 = 0xff - n_weight1; // fixed-point weights from alpha1
				SetPixelRGB(p_dest, x, y, RGBA(
					clampb((((n_pix0 >> 16) & 0xff) * n_weight0 + ((n_pix1 >> 16) & 0xff) * n_weight1) >> 8),
					clampb((((n_pix0 >> 8) & 0xff) * n_weight0 + ((n_pix1 >> 8) & 0xff) * n_weight1) >> 8),
					clampb(((n_pix0 & 0xff) * n_weight0 + (n_pix1 & 0xff) * n_weight1) >> 8),
					clampb(((n_pix0 >> 24) * n_weight0 + (n_pix1 >> 24) * n_weight1) >> 8)));
				// fast RGBA
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				float f_weight0 = 1 - p_pix1[3], f_weight1 = p_pix1[3]; // weights from alpha1
				float p_pixd[4];
				p_pixd[0] = p_pix0[0] * f_weight0 + p_pix1[0] * f_weight1;
				p_pixd[1] = p_pix0[1] * f_weight0 + p_pix1[1] * f_weight1;
				p_pixd[2] = p_pix0[2] * f_weight0 + p_pix1[2] * f_weight1;
				p_pixd[3] = p_pix0[3] * f_weight0 + p_pix1[3] * f_weight1;
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageInvert(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise 1 - p_src
 */
bool rrImageInvert(ImageStruct *p_dest, const ImageStruct *p_src0)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, 0xff - GetPixel8(p_src0, x0, y0));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					0xff - ((n_pix0 >> 16) & 0xff),
					0xff - ((n_pix0 >> 8) & 0xff),
					0xff - (n_pix0 & 0xff),
					0xff - (n_pix0 >> 24)));
				// fast RGB
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = 1 - p_pix0[0];
				p_pixd[1] = 1 - p_pix0[1];
				p_pixd[2] = 1 - p_pix0[2];
				p_pixd[3] = 1 - p_pix0[3];
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

inline unsigned char minb(unsigned char a, unsigned char b)
{
	return (a > b)? b : a;
}

inline unsigned char maxb(unsigned char a, unsigned char b)
{
	return (a > b)? a : b;
}

/**
 *	bool rrImageImageMin(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- pointwise minimum of p_src0 and p_src1
 */
bool rrImageImageMin(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, minb(GetPixel8(p_src0, x0, y0),
					GetPixel8(p_src1, x1, y1)));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					minb((unsigned char)(n_pix0 >> 16), (unsigned char)(n_pix1 >> 16)),
					minb((unsigned char)(n_pix0 >> 8), (unsigned char)(n_pix1 >> 8)),
					minb((unsigned char)n_pix0, (unsigned char)n_pix1),
					minb((unsigned char)(n_pix0 >> 24), (unsigned char)(n_pix1 >> 24))));
				// fast RGB
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				float p_pixd[4];
				p_pixd[0] = minf(p_pix0[0], p_pix1[0]);
				p_pixd[1] = minf(p_pix0[1], p_pix1[1]);
				p_pixd[2] = minf(p_pix0[2], p_pix1[2]);
				p_pixd[3] = minf(p_pix0[3], p_pix1[3]);
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageImageMax(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- pointwise maximum of p_src0 and p_src1
 */
bool rrImageImageMax(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	bool b_src1_resample = p_dest->XSize != p_src1->XSize || p_dest->YSize != p_src1->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	unsigned int w1 = p_src0->XSize, h1 = p_src1->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			int x1, y1;
			if(!b_src1_resample)
				x1 = x, y1 = y;
			else
				x1 = (x * w1) / w, y1 = (y * h1) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, maxb(GetPixel8(p_src0, x0, y0),
					GetPixel8(p_src1, x1, y1)));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0),
					n_pix1 = GetPixelRGB(p_src1, x1, y1);
				SetPixelRGB(p_dest, x, y, RGBA(
					maxb((unsigned char)(n_pix0 >> 16), (unsigned char)(n_pix1 >> 16)),
					maxb((unsigned char)(n_pix0 >> 8), (unsigned char)(n_pix1 >> 8)),
					maxb((unsigned char)n_pix0, (unsigned char)n_pix1),
					maxb((unsigned char)(n_pix0 >> 24), (unsigned char)(n_pix1 >> 24))));
				// fast RGB
			} else {
				float p_pix0[4], p_pix1[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				GetPixel4f(p_src1, x1, y1, p_pix1);
				float p_pixd[4];
				p_pixd[0] = maxf(p_pix0[0], p_pix1[0]);
				p_pixd[1] = maxf(p_pix0[1], p_pix1[1]);
				p_pixd[2] = maxf(p_pix0[2], p_pix1[2]);
				p_pixd[3] = maxf(p_pix0[3], p_pix1[3]);
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageThresh(ImageStruct *p_dest, const ImageStruct *p_src, float f_threshold)
 *		- thresholds image to f_threshold
 *		- note threshold value is relative to image pixel-type value range, for example
 *		  threshold 0.5 corresponds with both 128 for Image8 and 32767 for Image16
 *		- floating-point images are compared without any value rescaling
 */
bool rrImageThresh(ImageStruct *p_dest, const ImageStruct *p_src0, float f_thresh)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	short n_fp_thresh = (short)maxf(minf(f_thresh * 255, -1), 256);
	// clampt threshold to be in range of ubyte +- 1

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, (GetPixel8(p_src0, x0, y0) > n_fp_thresh)? 0xff : 0);
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					((short)((n_pix0 >> 16) & 0xff) > n_fp_thresh)? 0xff : 0,
					((short)((n_pix0 >> 8) & 0xff) > n_fp_thresh)? 0xff : 0,
					((short)(n_pix0 & 0xff) > n_fp_thresh)? 0xff : 0,
					((short)(n_pix0 >> 24) > n_fp_thresh)? 0xff : 0));
				// fast RGB
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = p_pix0[0] > f_thresh;
				p_pixd[1] = p_pix0[1] > f_thresh;
				p_pixd[2] = p_pix0[2] > f_thresh;
				p_pixd[3] = p_pix0[3] > f_thresh;
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageLog(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise log(1 + p_src)
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool rrImageLog(ImageStruct *p_dest, const ImageStruct *p_src0)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	static char p_fp_log[256];
	static bool b_fp_log_ready = false;
	if(n_path != 2 && !b_fp_log_ready) {
		b_fp_log_ready = true;
		for(short i = 0; i < 256; ++ i)
			p_fp_log[i] = (unsigned char)minf(maxf((float)log(i / 255.0f) * 255, 0), 255);
	}
	// precalc fixed-point log function to speed-up

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, p_fp_log[GetPixel8(p_src0, x0, y0)]);
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					p_fp_log[(n_pix0 >> 16) & 0xff],
					p_fp_log[(n_pix0 >> 8) & 0xff],
					p_fp_log[n_pix0 & 0xff],
					p_fp_log[n_pix0 >> 24]));
				// fast RGB
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = (float)log(p_pix0[0]);
				p_pixd[1] = (float)log(p_pix0[1]);
				p_pixd[2] = (float)log(p_pix0[2]);
				p_pixd[3] = (float)log(p_pix0[3]);
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageExp(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise exp(p_src) - 1
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool rrImageExp(ImageStruct *p_dest, const ImageStruct *p_src0)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	static char p_fp_exp[256];
	static bool b_fp_exp_ready = false;
	if(n_path != 2 && !b_fp_exp_ready) {
		b_fp_exp_ready = true;
		for(short i = 0; i < 256; ++ i)
			p_fp_exp[i] = (unsigned char)minf(maxf((float)exp(i / 255.0f) * 255, 0), 255);
	}
	// precalc fixed-point log function to speed-up

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, p_fp_exp[GetPixel8(p_src0, x0, y0)]);
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					p_fp_exp[(n_pix0 >> 16) & 0xff],
					p_fp_exp[(n_pix0 >> 8) & 0xff],
					p_fp_exp[n_pix0 & 0xff],
					p_fp_exp[n_pix0 >> 24]));
				// fast RGB
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = (float)exp(p_pix0[0]);
				p_pixd[1] = (float)exp(p_pix0[1]);
				p_pixd[2] = (float)exp(p_pix0[2]);
				p_pixd[3] = (float)exp(p_pix0[3]);
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageSqrt(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise square root of p_src
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageLog
 */
bool rrImageSqrt(ImageStruct *p_dest, const ImageStruct *p_src0)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	static char p_fp_sqrt[256];
	static bool b_fp_sqrt_ready = false;
	if(n_path != 2 && !b_fp_sqrt_ready) {
		b_fp_sqrt_ready = true;
		for(short i = 0; i < 256; ++ i)
			p_fp_sqrt[i] = (unsigned char)minf(maxf((float)sqrt(i / 255.0f) * 255, 0), 255);
	}
	// precalc fixed-point log function to speed-up

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, p_fp_sqrt[GetPixel8(p_src0, x0, y0)]);
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					p_fp_sqrt[(n_pix0 >> 16) & 0xff],
					p_fp_sqrt[(n_pix0 >> 8) & 0xff],
					p_fp_sqrt[n_pix0 & 0xff],
					p_fp_sqrt[n_pix0 >> 24]));
				// fast RGB
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = (float)sqrt(p_pix0[0]);
				p_pixd[1] = (float)sqrt(p_pix0[1]);
				p_pixd[2] = (float)sqrt(p_pix0[2]);
				p_pixd[3] = (float)sqrt(p_pix0[3]);
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImagePow(ImageStruct *p_dest, const ImageStruct *p_src, float f_power)
 *		- pointwise raises p_src to f_power
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageExp
 */
bool rrImagePow(ImageStruct *p_dest, const ImageStruct *p_src0, float f_power)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	double lf_power = (double)f_power;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, (unsigned char)minf(255, maxf(0, 255 * (float)pow(GetPixel8(p_src0, x0, y0) / 255.0, lf_power))));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					(unsigned char)minf(255, maxf(0, 255 * (float)pow(((n_pix0 >> 16) & 0xff) / 255.0, lf_power))),
					(unsigned char)minf(255, maxf(0, 255 * (float)pow(((n_pix0 >> 8) & 0xff) / 255.0, lf_power))),
					(unsigned char)minf(255, maxf(0, 255 * (float)pow((n_pix0 & 0xff) / 255.0, lf_power))),
					(unsigned char)minf(255, maxf(0, 255 * (float)pow((n_pix0 >> 24) / 255.0, lf_power)))));
				// fast RGB
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = (float)pow((double)p_pix0[0], lf_power);
				p_pixd[1] = (float)pow((double)p_pix0[1], lf_power);
				p_pixd[2] = (float)pow((double)p_pix0[2], lf_power);
				p_pixd[3] = (float)pow((double)p_pix0[3], lf_power);
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageAbs(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- calculates pointwise absolute value of p_src
 */
bool rrImageAbs(ImageStruct *p_dest, const ImageStruct *p_src0)
{
	char n_path = 6;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	else if((p_dest->PixelType == Image16 || p_dest->PixelType == Image16Linear) &&
	   (p_src0->PixelType == Image16 || p_src0->PixelType == Image16Linear))
		n_path = 2;
	else if((p_dest->PixelType == Image32 || p_dest->PixelType == Image32Linear) &&
	   (p_src0->PixelType == Image32 || p_src0->PixelType == Image32Linear))
		n_path = 3;
	else if((p_dest->PixelType == ImageFloat || p_dest->PixelType == ImageFloatLinear) &&
	   (p_src0->PixelType == ImageFloat || p_src0->PixelType == ImageFloatLinear))
		n_path = 4;
	else if((p_dest->PixelType == ImageComplex || p_dest->PixelType == ImageComplexLinear) &&
	   (p_src0->PixelType == ImageComplex || p_src0->PixelType == ImageComplexLinear))
		n_path = 5;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, GetPixel8(p_src0, x0, y0)); // always positive
				// fast gray
			} else if(n_path == 1) {
				SetPixelRGB(p_dest, x, y, GetPixelRGB(p_src0, x0, y0)); // always positive
				// fast RGB
			} else if(n_path == 2) {
				SetPixel16(p_dest, x, y, abs(GetPixel16(p_src0, x0, y0)));
				// fast gray
			} else if(n_path == 3) {
				SetPixel32(p_dest, x, y, abs(GetPixel32(p_src0, x0, y0)));
				// fast gray
			} else if(n_path == 4) {
				SetPixelFloat(p_dest, x, y, (float)fabs(GetPixelFloat(p_src0, x0, y0)));
				// fast gray
			} else if(n_path == 5) {
				ImageComplexValue t_src = GetPixelComplex(p_src0, x0, y0);
				ImageComplexValue t_abs = {(float)fabs(t_src.Re), (float)fabs(t_src.Im)};
				SetPixelComplex(p_dest, x, y, t_abs);
				// fast complex
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = (float)fabs(p_pix0[0]);
				p_pixd[1] = (float)fabs(p_pix0[1]);
				p_pixd[2] = (float)fabs(p_pix0[2]);
				p_pixd[3] = (float)fabs(p_pix0[3]);
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageScale(ImageStruct *p_dest, const ImageStruct *p_src, float f_scale, float f_bias)
 *		- pointwise image scaling as p_src * f_scale + f_bias
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- scale and bias values are unlimited
 */
bool rrImageScale(ImageStruct *p_dest, const ImageStruct *p_src0, float f_scale, float f_bias)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, (unsigned char)minf(255, maxf(0, f_bias + f_scale * GetPixel8(p_src0, x0, y0))));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					(unsigned char)minf(255, maxf(0, f_bias + f_scale * ((n_pix0 >> 16) & 0xff))),
					(unsigned char)minf(255, maxf(0, f_bias + f_scale * ((n_pix0 >> 8) & 0xff))),
					(unsigned char)minf(255, maxf(0, f_bias + f_scale * (n_pix0 & 0xff))),
					(unsigned char)minf(255, maxf(0, f_bias + f_scale * (n_pix0 >> 24)))));
				// fast RGB
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = f_bias + f_scale * p_pix0[0];
				p_pixd[1] = f_bias + f_scale * p_pix0[1];
				p_pixd[2] = f_bias + f_scale * p_pix0[2];
				p_pixd[3] = f_bias + f_scale * p_pix0[3];
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageColorMatrix(ImageStruct *p_dest, const ImageStruct *p_src,
 *		const float *p_mat4x4, bool b_transpose)
 *		- treats each point of p_src as four-dimensional vector (RGBA),
 *		  multiplies it with matrix p_mat4x4 which can be applied as is or transposed
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- matrix elements values are unlimited
 *		- for example this matrix can be used to convert RGB images to YCbCr colorspace:
 *		  (note alpha channel of such an image must contain value 1 (respectively 255))
 *
 *		float p_mat4x4[16] = { // RGB to YCbCr (ITU-R BT.601)
 *			    .299f,     .587f,     .114f,   0, // R' = .299R + .587G + .114B
 *			-.168736f, -.331264f,       .5f, .5f, // G' = .5 - .168736R - .331264G - .5B
 *			      .5f, -.418688f, -.081312f, .5f, // B' = .5 + .5R - .418688G - .081312B
 *			        0,         0,         0,   1  // A' = 1
 *		};
 */
bool rrImageColorMatrix(ImageStruct *p_dest, const ImageStruct *p_src0, const float *p_mat4x4, bool b_transpose)
{
	float f_tmp_matrix[4][4];
	if(!b_transpose) {
		for(int i = 0; i < 16; ++ i) {
			char x = i & 3, y = i >> 2;
			f_tmp_matrix[y][x] = p_mat4x4[x + y * 4];
		}
	} else {
		for(int i = 0; i < 16; ++ i) {
			char x = i & 3, y = i >> 2;
			f_tmp_matrix[x][y] = p_mat4x4[x + y * 4];
		}
	}
	// copy matrix to temp matrix to fight aliasing

	char n_path = 4;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 2;
	else if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 3;
	// processing path (0, 1, 2, 3 are optimized, 4 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	if(n_path == 0 || n_path == 2) {
		for(int i = 0; i < 4; ++ i)
			f_tmp_matrix[0][i] += f_tmp_matrix[1][i] + f_tmp_matrix[2][i];
	}
	// in case source is !optimized! gray, sum up first three matrix columns

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				unsigned char n_r = GetPixel8(p_src0, x0, y0);
				SetPixel8(p_dest, x, y, (unsigned char)minf(255, maxf(0, (f_tmp_matrix[0][0] /*+ f_tmp_matrix[1][0] + f_tmp_matrix[2][0]*/) * n_r + f_tmp_matrix[3][0] * 255)));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				unsigned char n_r = (unsigned char)(n_pix0 >> 16),
							  n_g = (unsigned char)(n_pix0 >> 8),
							  n_b = (unsigned char)(n_pix0),
							  n_a = (unsigned char)(n_pix0 >> 24);

				SetPixelRGB(p_dest, x, y, RGBA(
					(unsigned char)minf(255, maxf(0, f_tmp_matrix[0][0] * n_r + f_tmp_matrix[1][0] * n_g + f_tmp_matrix[2][0] * n_b + f_tmp_matrix[3][0] * n_a)),
					(unsigned char)minf(255, maxf(0, f_tmp_matrix[0][1] * n_r + f_tmp_matrix[1][1] * n_g + f_tmp_matrix[2][1] * n_b + f_tmp_matrix[3][1] * n_a)),
					(unsigned char)minf(255, maxf(0, f_tmp_matrix[0][2] * n_r + f_tmp_matrix[1][2] * n_g + f_tmp_matrix[2][2] * n_b + f_tmp_matrix[3][2] * n_a)),
					(unsigned char)minf(255, maxf(0, f_tmp_matrix[0][3] * n_r + f_tmp_matrix[1][3] * n_g + f_tmp_matrix[2][3] * n_b + f_tmp_matrix[3][3] * n_a))));
				// fast RGB
			} else if(n_path == 2) {
				unsigned char n_r = GetPixel8(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(
					(unsigned char)minf(255, maxf(0, (f_tmp_matrix[0][0] /*+ f_tmp_matrix[1][0] + f_tmp_matrix[2][0]*/) * n_r + f_tmp_matrix[3][0] * 255)),
					(unsigned char)minf(255, maxf(0, (f_tmp_matrix[0][1] /*+ f_tmp_matrix[1][1] + f_tmp_matrix[2][1]*/) * n_r + f_tmp_matrix[3][1] * 255)),
					(unsigned char)minf(255, maxf(0, (f_tmp_matrix[0][2] /*+ f_tmp_matrix[1][2] + f_tmp_matrix[2][2]*/) * n_r + f_tmp_matrix[3][2] * 255)),
					(unsigned char)minf(255, maxf(0, (f_tmp_matrix[0][3] /*+ f_tmp_matrix[1][3] + f_tmp_matrix[2][3]*/) * n_r + f_tmp_matrix[3][3] * 255))));
				// fast gray to RGB
			} else if(n_path == 3) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				unsigned char n_r = (unsigned char)(n_pix0 >> 16),
							  n_g = (unsigned char)(n_pix0 >> 8),
							  n_b = (unsigned char)(n_pix0),
							  n_a = (unsigned char)(n_pix0 >> 24);
				SetPixel8(p_dest, x, y, (unsigned char)minf(255, maxf(0, f_tmp_matrix[0][0] * n_r + f_tmp_matrix[1][0] * n_g + f_tmp_matrix[2][0] * n_b + f_tmp_matrix[3][0] * n_a)));
				// fast RGB to gray
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				for(int i = 0; i < 4; ++ i) {
					p_pixd[i] = f_tmp_matrix[0][i] * p_pix0[0] + f_tmp_matrix[1][i] * p_pix0[1] +
						f_tmp_matrix[2][i] * p_pix0[2] + f_tmp_matrix[3][i] * p_pix0[3];
				}
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageGreyscale(ImageStruct *p_dest, const ImageStruct *p_src)
 *		- turns source image into greyscale, using std ".299, .587, .144" equation
 *		- in case dest is RGBA image, RGB are set to gray value, A is set to 1
 */
bool rrImageGreyscale(ImageStruct *p_dest, const ImageStruct *p_src0)
{
	char n_path = 4;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 1;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear))
		n_path = 2;
	else if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear))
		n_path = 3;
	// processing path (0, 1, 2, 3 are optimized, 4 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			int x0, y0;
			if(!b_src0_resample)
				x0 = x, y0 = y;
			else
				x0 = (x * w0) / w, y0 = (y * h0) / h;
			// calculate source image coordinates (exact or nearest neighbour, just in case)

			if(!n_path) {
				SetPixel8(p_dest, x, y, GetPixel8(p_src0, x0, y0));
				// fast gray
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				unsigned char n_r = (unsigned char)(n_pix0 >> 16),
							  n_g = (unsigned char)(n_pix0 >> 8),
							  n_b = (unsigned char)(n_pix0),
							  n_a = (unsigned char)(n_pix0 >> 24);

				const unsigned short n_red_weight = (unsigned short)(.299f * 0x10000);
				const unsigned short n_green_weight = (unsigned short)(.587f * 0x10000);
				const unsigned short n_blue_weight = (unsigned short)(.114f * 0x10000);
				// 0:16 fixed-point PAL weights

				unsigned char n_i = (unsigned char)((n_red_weight * (unsigned int)n_r +
					n_green_weight * (unsigned int)n_g +
					n_blue_weight * (unsigned int)n_b + 0x8000) >> 16);
				// calc intensity

				SetPixelRGB(p_dest, x, y, RGBA(n_i, n_i, n_i, n_a));
				// fast RGB
			} else if(n_path == 2) {
				unsigned char n_r = GetPixel8(p_src0, x0, y0);
				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_r, n_r, 0xff));
				// fast gray to RGB
			} else if(n_path == 3) {
				ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
				unsigned char n_r = (unsigned char)(n_pix0 >> 16),
							  n_g = (unsigned char)(n_pix0 >> 8),
							  n_b = (unsigned char)(n_pix0);

				const unsigned short n_red_weight = (unsigned short)(.299f * 0x10000);
				const unsigned short n_green_weight = (unsigned short)(.587f * 0x10000);
				const unsigned short n_blue_weight = (unsigned short)(.114f * 0x10000);
				// 0:16 fixed-point PAL weights

				unsigned char n_i = (unsigned char)((n_red_weight * (unsigned int)n_r +
					n_green_weight * (unsigned int)n_g +
					n_blue_weight * (unsigned int)n_b + 0x8000) >> 16);
				// calc intensity

				SetPixel8(p_dest, x, y, n_i);
				// fast RGB to gray
			} else {
				float p_pix0[4];
				GetPixel4f(p_src0, x0, y0, p_pix0);
				float p_pixd[4];
				p_pixd[0] = p_pixd[1] = p_pixd[2] = .299f * p_pix0[0] +
					.587f * p_pix0[1] + .114f * p_pix0[2];
				p_pixd[3] = p_pix0[3];
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageAffineTransform(ImageStruct *p_dest, const ImageStruct *p_src,
 *		float x0, float y0, float z0, float x1, float y1, float z1,
 *		float x2, float y2, float z2, float x3, float y3, float z3)
 *		- arbitrary image interpolation over convex quad given as (x0, y0, z0), (x1, y1, z1), ...
 *		- coordinates in OpenGL-ish fashion, orthographic projection
 */
bool rrImageAffineTransform(ImageStruct *p_dest, const ImageStruct *p_src,
	float x0, float y0, float z0, float x1, float y1, float z1,
	float x2, float y2, float z2, float x3, float y3, float z3)
{
	return false; // todo
}

/**
 *	bool rrImageSquareErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0xff;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						if(GetPixel8(p_src, x0, y0) < 127) {
							n_result = 0;
							break;
						}
					}
					if(!n_result)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r = 0xff,
							  n_g = 0xff,
							  n_b = 0xff,
							  n_a = 0xff;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char _n_r = (unsigned char)(n_pix0 >> 16),
									  _n_g = (unsigned char)(n_pix0 >> 8),
									  _n_b = (unsigned char)(n_pix0),
									  _n_a = (unsigned char)(n_pix0 >> 24);
						if(_n_r < 127)
							n_r = 0;
						if(_n_g < 127)
							n_g = 0;
						if(_n_b < 127)
							n_b = 0;
						if(_n_a < 127)
							n_a = 0;
					}
					if(!n_r && !n_g && !n_b && !n_a)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_g, n_b, n_a));
				// fast RGB
			} else {
				float p_pixd[4] = {1, 1, 1, 1};

				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						if(p_pix0[0] < .5f)
							p_pixd[0] = 0;
						if(p_pix0[1] < .5f)
							p_pixd[1] = 0;
						if(p_pix0[2] < .5f)
							p_pixd[2] = 0;
						if(p_pix0[3] < .5f)
							p_pixd[3] = 0;
					}
					if(p_pixd[0] == 0 && p_pixd[1] == 0 && p_pixd[2] == 0 && p_pixd[3] == 0)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageSquareDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						if(GetPixel8(p_src, x0, y0) >= 127) {
							n_result = 0xff;
							break;
						}
					}
					if(n_result)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r = 0,
							  n_g = 0,
							  n_b = 0,
							  n_a = 0;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char _n_r = (unsigned char)(n_pix0 >> 16),
									  _n_g = (unsigned char)(n_pix0 >> 8),
									  _n_b = (unsigned char)(n_pix0),
									  _n_a = (unsigned char)(n_pix0 >> 24);
						if(_n_r >= 127)
							n_r = 0xff;
						if(_n_g >= 127)
							n_g = 0xff;
						if(_n_b >= 127)
							n_b = 0xff;
						if(_n_a >= 127)
							n_a = 0xff;
					}
					if(n_r && n_g && n_b && n_a)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_g, n_b, n_a));
				// fast RGB
			} else {
				float p_pixd[4] = {0, 0, 0, 0};

				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						if(p_pix0[0] >= .5f)
							p_pixd[0] = 1;
						if(p_pix0[1] >= .5f)
							p_pixd[1] = 1;
						if(p_pix0[2] >= .5f)
							p_pixd[2] = 1;
						if(p_pix0[3] >= .5f)
							p_pixd[3] = 1;
					}
					if(p_pixd[0] == 1 && p_pixd[1] == 1 && p_pixd[2] == 1 && p_pixd[3] == 1)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageSquareOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	ImageStruct *p_temp_image = p_Alloc_Image(p_src->XSize, p_src->YSize, p_src->PixelType);
	if(p_temp_image == NULL)
		return false;
	// alloc temp image

	if(!rrImageSquareErode(p_temp_image, p_src, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	if(!rrImageSquareDilate(p_dest, p_temp_image, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	// erode then dilate

	p_temp_image->Optional.Regular.Free(p_temp_image);
	// free temp image

	return true;
}

/**
 *	bool rrImageSquareClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool rrImageSquareClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	ImageStruct *p_temp_image = p_Alloc_Image(p_src->XSize, p_src->YSize, p_src->PixelType);
	if(p_temp_image == NULL)
		return false;
	// alloc temp image

	if(!rrImageSquareDilate(p_temp_image, p_src, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	if(!rrImageSquareErode(p_dest, p_temp_image, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	// dilate then erode

	p_temp_image->Optional.Regular.Free(p_temp_image);
	// free temp image

	return true;
}

/**
 *	bool rrImageDiamondErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondErode(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	int m = n_kernel_h / 4; // chamfer size
	int n = n_kernel_h / 2;
	int o = n_kernel_w / 2;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0xff;
				for(int dy = x - n, k = 0; dy < (int)(x + n); ++ dy, ++ k) {
					int n_space = (k < n)? (n - m) - k : k - (n + m);
					n_space = (n_space < 0)? 0 : n_space;
					for(int dx = x - o + n_space; dx < (int)(x + o) - n_space; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						if(GetPixel8(p_src, x0, y0) < 127) {
							n_result = 0;
							break;
						}
					}
					if(!n_result)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r = 0xff,
							  n_g = 0xff,
							  n_b = 0xff,
							  n_a = 0xff;
				for(int dy = x - n, k = 0; dy < (int)(x + n); ++ dy, ++ k) {
					int n_space = (k < n)? (n - m) - k : k - (n + m);
					n_space = (n_space < 0)? 0 : n_space;
					for(int dx = x - o + n_space; dx < (int)(x + o) - n_space; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char _n_r = (unsigned char)(n_pix0 >> 16),
									  _n_g = (unsigned char)(n_pix0 >> 8),
									  _n_b = (unsigned char)(n_pix0),
									  _n_a = (unsigned char)(n_pix0 >> 24);
						if(_n_r < 127)
							n_r = 0;
						if(_n_g < 127)
							n_g = 0;
						if(_n_b < 127)
							n_b = 0;
						if(_n_a < 127)
							n_a = 0;
					}
					if(!n_r && !n_g && !n_b && !n_a)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_g, n_b, n_a));
				// fast RGB
			} else {
				float p_pixd[4] = {1, 1, 1, 1};

				for(int dy = x - n, k = 0; dy < (int)(x + n); ++ dy, ++ k) {
					int n_space = (k < n)? (n - m) - k : k - (n + m);
					n_space = (n_space < 0)? 0 : n_space;
					for(int dx = x - o + n_space; dx < (int)(x + o) - n_space; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						if(p_pix0[0] < .5f)
							p_pixd[0] = 0;
						if(p_pix0[1] < .5f)
							p_pixd[1] = 0;
						if(p_pix0[2] < .5f)
							p_pixd[2] = 0;
						if(p_pix0[3] < .5f)
							p_pixd[3] = 0;
					}
					if(p_pixd[0] == 0 && p_pixd[1] == 0 && p_pixd[2] == 0 && p_pixd[3] == 0)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageDiamondDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondDilate(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	int m = n_kernel_h / 4; // chamfer size
	int n = n_kernel_h / 2;
	int o = n_kernel_w / 2;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0;
				for(int dy = x - n, k = 0; dy < (int)(x + n); ++ dy, ++ k) {
					int n_space = (k < n)? (n - m) - k : k - (n + m);
					n_space = (n_space < 0)? 0 : n_space;
					for(int dx = x - o + n_space; dx < (int)(x + o) - n_space; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						if(GetPixel8(p_src, x0, y0) >= 127) {
							n_result = 0xff;
							break;
						}
					}
					if(n_result)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r = 0,
							  n_g = 0,
							  n_b = 0,
							  n_a = 0;
				for(int dy = x - n, k = 0; dy < (int)(x + n); ++ dy, ++ k) {
					int n_space = (k < n)? (n - m) - k : k - (n + m);
					n_space = (n_space < 0)? 0 : n_space;
					for(int dx = x - o + n_space; dx < (int)(x + o) - n_space; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char _n_r = (unsigned char)(n_pix0 >> 16),
									  _n_g = (unsigned char)(n_pix0 >> 8),
									  _n_b = (unsigned char)(n_pix0),
									  _n_a = (unsigned char)(n_pix0 >> 24);
						if(_n_r >= 127)
							n_r = 0xff;
						if(_n_g >= 127)
							n_g = 0xff;
						if(_n_b >= 127)
							n_b = 0xff;
						if(_n_a >= 127)
							n_a = 0xff;
					}
					if(n_r && n_g && n_b && n_a)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_g, n_b, n_a));
				// fast RGB
			} else {
				float p_pixd[4] = {0, 0, 0, 0};

				for(int dy = x - n, k = 0; dy < (int)(x + n); ++ dy, ++ k) {
					int n_space = (k < n)? (n - m) - k : k - (n + m);
					n_space = (n_space < 0)? 0 : n_space;
					for(int dx = x - o + n_space; dx < (int)(x + o) - n_space; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						if(p_pix0[0] >= .5f)
							p_pixd[0] = 1;
						if(p_pix0[1] >= .5f)
							p_pixd[1] = 1;
						if(p_pix0[2] >= .5f)
							p_pixd[2] = 1;
						if(p_pix0[3] >= .5f)
							p_pixd[3] = 1;
					}
					if(p_pixd[0] == 1 && p_pixd[1] == 1 && p_pixd[2] == 1 && p_pixd[3] == 1)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageDiamondOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondOpen(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	ImageStruct *p_temp_image = p_Alloc_Image(p_src->XSize, p_src->YSize, p_src->PixelType);
	if(p_temp_image == NULL)
		return false;
	// alloc temp image

	if(!rrImageDiamondErode(p_temp_image, p_src, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	if(!rrImageDiamondDilate(p_dest, p_temp_image, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	// erode then dilate

	p_temp_image->Optional.Regular.Free(p_temp_image);
	// free temp image

	return true;
}

/**
 *	bool rrImageDiamondClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool rrImageDiamondClose(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	ImageStruct *p_temp_image = p_Alloc_Image(p_src->XSize, p_src->YSize, p_src->PixelType);
	if(p_temp_image == NULL)
		return false;
	// alloc temp image

	if(!rrImageDiamondDilate(p_temp_image, p_src, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	if(!rrImageDiamondErode(p_dest, p_temp_image, n_kernel_w, n_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	// dilate then erode

	p_temp_image->Optional.Regular.Free(p_temp_image);
	// free temp image

	return true;
}

/**
 *	bool rrImageImageErode(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- erosion morphologic operator
 *		- structuring element is p_src1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageErode(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	int n_kernel_h = p_src1->XSize, n_kernel_w = p_src1->YSize;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0xff;
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						if(GetPixel8(p_src1, dx, dy) < 127)
							continue;
						// non-important pixel in the mask

						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						if(GetPixel8(p_src0, x0, y0) < 127) {
							n_result = 0;
							break;
						}
					}
					if(!n_result)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r = 0xff,
							  n_g = 0xff,
							  n_b = 0xff,
							  n_a = 0xff;
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix1 = GetPixelRGB(p_src1, dx, dx);
						unsigned char _n_r1 = (unsigned char)(n_pix1 >> 16),
									  _n_g1 = (unsigned char)(n_pix1 >> 8),
									  _n_b1 = (unsigned char)(n_pix1),
									  _n_a1 = (unsigned char)(n_pix1 >> 24);
						// structuring element pixel

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
						unsigned char _n_r = (unsigned char)(n_pix0 >> 16),
									  _n_g = (unsigned char)(n_pix0 >> 8),
									  _n_b = (unsigned char)(n_pix0),
									  _n_a = (unsigned char)(n_pix0 >> 24);
						// image pixel

						if(_n_r1 >= 127 && _n_r < 127)
							n_r = 0;
						if(_n_g1 >= 127 && _n_g < 127)
							n_g = 0;
						if(_n_b1 >= 127 && _n_b < 127)
							n_b = 0;
						if(_n_a1 >= 127 && _n_a < 127)
							n_a = 0;
					}
					if(!n_r && !n_g && !n_b && !n_a)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_g, n_b, n_a));
				// fast RGB
			} else {
				float p_pixd[4] = {1, 1, 1, 1};
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix1[4];
						GetPixel4f(p_src1, dx, dy, p_pix1);
						// structuring element pixel

						float p_pix0[4];
						GetPixel4f(p_src0, x0, y0, p_pix0);
						// image pixel

						if(p_pix1[0] >= .5f && p_pix0[0] < .5f)
							p_pixd[0] = 0;
						if(p_pix1[1] >= .5f && p_pix0[1] < .5f)
							p_pixd[1] = 0;
						if(p_pix1[2] >= .5f && p_pix0[2] < .5f)
							p_pixd[2] = 0;
						if(p_pix1[3] >= .5f && p_pix0[3] < .5f)
							p_pixd[3] = 0;
					}
					if(p_pixd[0] == 0 && p_pixd[1] == 0 && p_pixd[2] == 0 && p_pixd[3] == 0)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageImageDilate(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- dilation morphologic operator
 *		- structuring element is p_src1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageDilate(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	int n_kernel_h = p_src1->XSize, n_kernel_w = p_src1->YSize;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0xff;
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						if(GetPixel8(p_src1, dx, dy) < 127)
							continue;
						// non-important pixel in the mask

						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						if(GetPixel8(p_src0, x0, y0) >= 127) {
							n_result = 0xff;
							break;
						}
					}
					if(!n_result)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r = 0,
							  n_g = 0,
							  n_b = 0,
							  n_a = 0;
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix1 = GetPixelRGB(p_src1, dx, dx);
						unsigned char _n_r1 = (unsigned char)(n_pix1 >> 16),
									  _n_g1 = (unsigned char)(n_pix1 >> 8),
									  _n_b1 = (unsigned char)(n_pix1),
									  _n_a1 = (unsigned char)(n_pix1 >> 24);
						// structuring element pixel

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
						unsigned char _n_r = (unsigned char)(n_pix0 >> 16),
									  _n_g = (unsigned char)(n_pix0 >> 8),
									  _n_b = (unsigned char)(n_pix0),
									  _n_a = (unsigned char)(n_pix0 >> 24);
						// image pixel

						if(_n_r1 >= 127 && _n_r >= 127)
							n_r = 0xff;
						if(_n_g1 >= 127 && _n_g >= 127)
							n_g = 0xff;
						if(_n_b1 >= 127 && _n_b >= 127)
							n_b = 0xff;
						if(_n_a1 >= 127 && _n_a >= 127)
							n_a = 0xff;
					}
					if(n_r && n_g && n_b && n_a)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_g, n_b, n_a));
				// fast RGB
			} else {
				float p_pixd[4] = {0, 0, 0, 0};
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix1[4];
						GetPixel4f(p_src1, dx, dy, p_pix1);
						// structuring element pixel

						float p_pix0[4];
						GetPixel4f(p_src0, x0, y0, p_pix0);
						// image pixel

						if(p_pix1[0] >= .5f && p_pix0[0] >= .5f)
							p_pixd[0] = 1;
						if(p_pix1[1] >= .5f && p_pix0[1] >= .5f)
							p_pixd[1] = 1;
						if(p_pix1[2] >= .5f && p_pix0[2] >= .5f)
							p_pixd[2] = 1;
						if(p_pix1[3] >= .5f && p_pix0[3] >= .5f)
							p_pixd[3] = 1;
					}
					if(p_pixd[0] == 1 && p_pixd[1] == 1 && p_pixd[2] == 1 && p_pixd[3] == 1)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageImageOpen(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- opening morphologic operator
 *		- structuring element is p_src1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageOpen(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	ImageStruct *p_temp_image = p_Alloc_Image(p_src0->XSize, p_src0->YSize, p_src0->PixelType);
	if(p_temp_image == NULL)
		return false;
	// alloc temp image

	if(!rrImageImageErode(p_temp_image, p_src0, p_src1)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	if(!rrImageImageDilate(p_dest, p_temp_image, p_src1)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	// erode then dilate

	p_temp_image->Optional.Regular.Free(p_temp_image);
	// free temp image

	return true;
}

/**
 *	bool rrImageImageClose(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- closing morphologic operator
 *		- structuring element is p_src1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool rrImageImageClose(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	ImageStruct *p_temp_image = p_Alloc_Image(p_src0->XSize, p_src0->YSize, p_src0->PixelType);
	if(p_temp_image == NULL)
		return false;
	// alloc temp image

	if(!rrImageImageDilate(p_temp_image, p_src0, p_src1)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	if(!rrImageImageErode(p_dest, p_temp_image, p_src1)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	// dilate then erode

	p_temp_image->Optional.Regular.Free(p_temp_image);
	// free temp image

	return true;
}

/**
 *	bool rrImageImageHitMiss(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
 *		- hit-and-miss morphologic operator
 *		- structuring element is p_src1
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- structuring element is thresholded as follows: values below .25 are "zero",
 *		  values above .75 are "one", values between .25 and .75 are "don't care"
 */
bool rrImageImageHitMiss(ImageStruct *p_dest, const ImageStruct *p_src0, const ImageStruct *p_src1)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src0->PixelType == Image8 || p_src0->PixelType == Image8Linear) &&
	   (p_src1->PixelType == Image8 || p_src1->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src0->PixelType == ImageRGB || p_src0->PixelType == ImageRGBLinear) &&
	   (p_src1->PixelType == ImageRGB || p_src1->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src0->XSize || p_dest->YSize != p_src0->YSize;
	// will we be resampling?

	unsigned int w0 = p_src0->XSize, h0 = p_src0->YSize;
	// copy sizes here to avoid aliasing

	int n_kernel_h = p_src1->XSize, n_kernel_w = p_src1->YSize;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0xff;
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						if(abs(GetPixel8(p_src0, x0, y0) - GetPixel8(p_src1, dx, dy)) > 192) {
							n_result = 0;
							break;
						}
					}
					if(!n_result)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r = 0xff,
							  n_g = 0xff,
							  n_b = 0xff,
							  n_a = 0xff;
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix1 = GetPixelRGB(p_src1, dx, dx);
						unsigned char _n_r1 = (unsigned char)(n_pix1 >> 16),
									  _n_g1 = (unsigned char)(n_pix1 >> 8),
									  _n_b1 = (unsigned char)(n_pix1),
									  _n_a1 = (unsigned char)(n_pix1 >> 24);
						// structuring element pixel

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src0, x0, y0);
						unsigned char _n_r = (unsigned char)(n_pix0 >> 16),
									  _n_g = (unsigned char)(n_pix0 >> 8),
									  _n_b = (unsigned char)(n_pix0),
									  _n_a = (unsigned char)(n_pix0 >> 24);
						// image pixel

						if(abs(_n_r1 - _n_r) > 192)
							n_r = 0;
						if(abs(_n_g1 - _n_g) > 192)
							n_g = 0;
						if(abs(_n_b1 - _n_b) > 192)
							n_b = 0;
						if(abs(_n_a1 - _n_a) > 192)
							n_a = 0;
					}
					if(!n_r && !n_g && !n_b && !n_a)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixelRGB(p_dest, x, y, RGBA(n_r, n_g, n_b, n_a));
				// fast RGB
			} else {
				float p_pixd[4] = {1, 1, 1, 1};
				for(int dy = 0; dy < n_kernel_h; ++ dy) {
					for(int dx = 0; dx < n_kernel_w; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n_kernel_w / 2, y0 = dy - n_kernel_h / 2;
						else
							x0 = ((dx - n_kernel_w / 2) * w0) / w, y0 = ((dy - n_kernel_h / 2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix1[4];
						GetPixel4f(p_src1, dx, dy, p_pix1);
						// structuring element pixel

						float p_pix0[4];
						GetPixel4f(p_src0, x0, y0, p_pix0);
						// image pixel

						if(fabs(p_pix1[0] - p_pix0[0]) > .7)
							p_pixd[0] = 0;
						if(fabs(p_pix1[1] - p_pix0[1]) > .7)
							p_pixd[1] = 0;
						if(fabs(p_pix1[2] - p_pix0[2]) > .7)
							p_pixd[2] = 0;
						if(fabs(p_pix1[3] - p_pix0[3]) > .7)
							p_pixd[3] = 0;
					}
					if(p_pixd[0] == 0 && p_pixd[1] == 0 && p_pixd[2] == 0 && p_pixd[3] == 0)
						break;
				}
				// erode (brute force, could do scanline decomposition, but we have GPU)

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageMean(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- separable (two-pass) mean image filter
 *		- averaging window size is n_kernel_w  n_kernel_h pixels
 */
bool rrImageMean(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	int n_kernel_area = n_kernel_w * n_kernel_h;
	float f_inv_kernel_area = 1.0f / n_kernel_area;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned long n_result = 0;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						n_result += GetPixel8(p_src, x0, y0);
					}
				}
				// mean

				SetPixel8(p_dest, x, y, (unsigned char)/*clampb*/(n_result / n_kernel_area));
				// fast gray
			} else if(n_path == 1) {
				unsigned long n_r_sum = 0,
							  n_g_sum = 0,
							  n_b_sum = 0,
							  n_a_sum = 0;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char n_r = (unsigned char)(n_pix0 >> 16),
									  n_g = (unsigned char)(n_pix0 >> 8),
									  n_b = (unsigned char)(n_pix0),
									  n_a = (unsigned char)(n_pix0 >> 24);
						n_r_sum += n_r;
						n_g_sum += n_g;
						n_b_sum += n_b;
						n_a_sum += n_a;
					}
				}
				// mean

				SetPixelRGB(p_dest, x, y, RGBA(/*clampb*/(n_r_sum / n_kernel_area),
					/*clampb*/(n_g_sum / n_kernel_area), /*clampb*/(n_b_sum / n_kernel_area),
					/*clampb*/(n_a_sum / n_kernel_area)));
				// fast RGB
			} else {
				float p_pixd[4] = {0, 0, 0, 0};

				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						p_pixd[0] += p_pix0[0];
						p_pixd[1] += p_pix0[1];
						p_pixd[2] += p_pix0[2];
						p_pixd[3] += p_pix0[3];
					}
				}
				// mean

				p_pixd[0] *= f_inv_kernel_area;
				p_pixd[1] *= f_inv_kernel_area;
				p_pixd[2] *= f_inv_kernel_area;
				p_pixd[3] *= f_inv_kernel_area;
				// rescale

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

static int n_cmp_ubyte(const void *p_a, const void *p_b)
{
	return (int)(*(const unsigned char*)p_a - *(const unsigned char*)p_b);
}

static int n_cmp_float(const void *p_a, const void *p_b)
{
	if(*(const float*)p_a > *(const char*)p_b)
		return 1;
	if(*(const float*)p_a < *(const char*)p_b)
		return -1;
	return 0;
}

/**
 *	bool rrImageMedian(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_size)
 *		- separable (two-pass) median image filter
 *		- averaging window size is n_kernel_size  n_kernel_size pixels, the only supported
 *		  sizes are 3  3, 5  5, 7  7 and 15  15.
 */
bool rrImageMedian(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_size)
{
	if(n_kernel_size > 15)
		return false;

	unsigned char p_byte_values[4][15 * 15];
	float p_float_values[4][15 * 15];
	int n_kernel_area = n_kernel_size * n_kernel_size;
	int n_median_index = n_kernel_area / 2;
	// arrays where values will be sorted

	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	int m = n_kernel_size / 2;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				n_kernel_area = 0;
				for(int dy = x - m; dy < (int)(x + m); ++ dy) {
					for(int dx = x - m; dx < (int)(x + m); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						p_byte_values[0][n_kernel_area ++] = GetPixel8(p_src, x0, y0);
					}
				}
				// gather values

				qsort(p_byte_values[0], n_kernel_area, sizeof(char), n_cmp_ubyte);
				// sort

				SetPixel8(p_dest, x, y, p_byte_values[0][n_median_index]);
				// fast gray
			} else if(n_path == 1) {
				n_kernel_area = 0;
				for(int dy = x - m; dy < (int)(x + m); ++ dy) {
					for(int dx = x - m; dx < (int)(x + m); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char n_r = (unsigned char)(n_pix0 >> 16),
									  n_g = (unsigned char)(n_pix0 >> 8),
									  n_b = (unsigned char)(n_pix0),
									  n_a = (unsigned char)(n_pix0 >> 24);
						p_byte_values[0][n_kernel_area] = n_r;
						p_byte_values[1][n_kernel_area] = n_g;
						p_byte_values[2][n_kernel_area] = n_b;
						p_byte_values[3][n_kernel_area ++] = n_a;
					}
				}
				// gather values

				qsort(p_byte_values[0], n_kernel_area, sizeof(char), n_cmp_ubyte);
				qsort(p_byte_values[1], n_kernel_area, sizeof(char), n_cmp_ubyte);
				qsort(p_byte_values[2], n_kernel_area, sizeof(char), n_cmp_ubyte);
				qsort(p_byte_values[3], n_kernel_area, sizeof(char), n_cmp_ubyte);
				// sort

				SetPixelRGB(p_dest, x, y, RGBA(
					p_byte_values[0][n_median_index],
					p_byte_values[1][n_median_index],
					p_byte_values[2][n_median_index],
					p_byte_values[3][n_median_index]));
				// fast RGB
			} else {
				n_kernel_area = 0;
				for(int dy = x - m; dy < (int)(x + m); ++ dy) {
					for(int dx = x - m; dx < (int)(x + m); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						p_float_values[0][n_kernel_area] = p_pix0[0];
						p_float_values[1][n_kernel_area] = p_pix0[1];
						p_float_values[2][n_kernel_area] = p_pix0[2];
						p_float_values[3][n_kernel_area ++] = p_pix0[3];
					}
				}
				// gather values

				qsort(p_float_values[0], n_kernel_area, sizeof(float), n_cmp_float);
				qsort(p_float_values[1], n_kernel_area, sizeof(float), n_cmp_float);
				qsort(p_float_values[2], n_kernel_area, sizeof(float), n_cmp_float);
				qsort(p_float_values[3], n_kernel_area, sizeof(float), n_cmp_float);
				// sort

				float p_pixd[4];
				p_pixd[0] = p_float_values[0][n_median_index];
				p_pixd[1] = p_float_values[1][n_median_index];
				p_pixd[2] = p_float_values[2][n_median_index];
				p_pixd[3] = p_float_values[3][n_median_index];
				// pick median

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageLocalMin(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- finds local minimum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool rrImageLocalMin(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0xff;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						unsigned char n_pixel = GetPixel8(p_src, x0, y0);
						if(n_result > n_pixel)
							n_result = n_pixel;
					}
				}
				// min

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r_result = 0xff,
							  n_g_result = 0xff,
							  n_b_result = 0xff,
							  n_a_result = 0xff;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char n_r = (unsigned char)(n_pix0 >> 16),
									  n_g = (unsigned char)(n_pix0 >> 8),
									  n_b = (unsigned char)(n_pix0),
									  n_a = (unsigned char)(n_pix0 >> 24);
						if(n_r_result > n_r) n_r_result = n_r;
						if(n_g_result > n_g) n_g_result = n_g;
						if(n_b_result > n_b) n_b_result = n_b;
						if(n_a_result > n_a) n_a_result = n_a;
					}
				}
				// min

				SetPixelRGB(p_dest, x, y, RGBA(n_r_result, n_g_result, n_b_result, n_a_result));
				// fast RGB
			} else {
				float p_pixd[4];
				GetPixel4f(p_src, 0, 0, p_pixd);

				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						if(p_pixd[0] > p_pix0[0]) p_pixd[0] = p_pix0[0];
						if(p_pixd[1] > p_pix0[1]) p_pixd[1] = p_pix0[1];
						if(p_pixd[2] > p_pix0[2]) p_pixd[2] = p_pix0[2];
						if(p_pixd[3] > p_pix0[3]) p_pixd[3] = p_pix0[3];
					}
				}
				// min

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageLocalMax(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
 *		- finds local maximum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool rrImageLocalMax(ImageStruct *p_dest, const ImageStruct *p_src, int n_kernel_w, int n_kernel_h)
{
	char n_path = 2;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear))
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_result = 0;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						unsigned char n_pixel = GetPixel8(p_src, x0, y0);
						if(n_result < n_pixel)
							n_result = n_pixel;
					}
				}
				// min

				SetPixel8(p_dest, x, y, n_result);
				// fast gray
			} else if(n_path == 1) {
				unsigned char n_r_result = 0,
							  n_g_result = 0,
							  n_b_result = 0,
							  n_a_result = 0;
				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char n_r = (unsigned char)(n_pix0 >> 16),
									  n_g = (unsigned char)(n_pix0 >> 8),
									  n_b = (unsigned char)(n_pix0),
									  n_a = (unsigned char)(n_pix0 >> 24);
						if(n_r_result < n_r) n_r_result = n_r;
						if(n_g_result < n_g) n_g_result = n_g;
						if(n_b_result < n_b) n_b_result = n_b;
						if(n_a_result < n_a) n_a_result = n_a;
					}
				}
				// max

				SetPixelRGB(p_dest, x, y, RGBA(n_r_result, n_g_result, n_b_result, n_a_result));
				// fast RGB
			} else {
				float p_pixd[4];
				GetPixel4f(p_src, 0, 0, p_pixd);

				for(int dy = x - n_kernel_h / 2; dy < (int)(x + n_kernel_h / 2); ++ dy) {
					for(int dx = x - n_kernel_w / 2; dx < (int)(x + n_kernel_w / 2); ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx, y0 = dy;
						else
							x0 = (dx * w0) / w, y0 = (dy * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						if(p_pixd[0] < p_pix0[0]) p_pixd[0] = p_pix0[0];
						if(p_pixd[1] < p_pix0[1]) p_pixd[1] = p_pix0[1];
						if(p_pixd[2] < p_pix0[2]) p_pixd[2] = p_pix0[2];
						if(p_pixd[3] < p_pix0[3]) p_pixd[3] = p_pix0[3];
					}
				}
				// max

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageConvolveSeparable(ImageStruct *p_dest, const ImageStruct *p_src,
 *		ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
 *		- separable convolution filter
 *		- this implementation imposes strict kernel alignment
 *		  (p_kernel_h->YSize = 1, p_kernel_v->XSize = 1)
 *		- note this handles convolution with complex kernels and / or complex image
 */
bool rrImageConvolveSeparable(ImageStruct *p_dest, const ImageStruct *p_src,
	ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
{
	if(p_kernel_h->YSize != 1 || p_kernel_v->XSize != 1)
		return false;
	// this implementation imposes strict kernel alignment

	ImageStruct *p_temp_image = p_Alloc_Image(p_src->XSize, p_src->YSize, p_src->PixelType);
	if(p_temp_image == NULL)
		return false;
	// alloc temp image

	if(!rrImageConvolve2D(p_temp_image, p_src, p_kernel_h)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	if(!rrImageConvolve2D(p_dest, p_temp_image, p_kernel_v)) {
		p_temp_image->Optional.Regular.Free(p_temp_image);
		return false;
	}
	// convolve with both kernels

	p_temp_image->Optional.Regular.Free(p_temp_image);
	// free temp image

	return true;
}

inline ImageComplexValue t_CmplxAdd(ImageComplexValue x, ImageComplexValue y)
{
	ImageComplexValue t_tmp = {x.Re + y.Re, x.Im + y.Im};
	return t_tmp;
}

/**
 *	bool rrImageConvolve2D(ImageStruct *p_dest, const ImageStruct *p_src, ImageStruct *p_kernel)
 *		- two-dimensional convolution filter
 *		- note this handles convolution with complex kernel and / or complex image
 */
bool rrImageConvolve2D(ImageStruct *p_dest, const ImageStruct *p_src, ImageStruct *p_kernel)
{
	int n_path;
	if((p_dest->PixelType == Image8 || p_dest->PixelType == Image8Linear) &&
	   (p_src->PixelType == Image8 || p_src->PixelType == Image8Linear) &&
	   (p_kernel->PixelType == Image8 || p_kernel->PixelType == Image8Linear))
		n_path = 0;
	else if((p_dest->PixelType == ImageRGB || p_dest->PixelType == ImageRGBLinear) &&
	   (p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear) &&
	   (p_kernel->PixelType == ImageRGB || p_kernel->PixelType == ImageRGBLinear))
		n_path = 1;
	else if((p_src->PixelType != ImageComplex || p_src->PixelType != ImageComplexLinear) &&
	   (p_kernel->PixelType != ImageComplex || p_kernel->PixelType != ImageComplexLinear))
		n_path = 2; // real * real
	else if((p_src->PixelType == ImageComplex || p_src->PixelType == ImageComplexLinear) &&
	   (p_kernel->PixelType != ImageComplex || p_kernel->PixelType != ImageComplexLinear))
		n_path = 3; // complex * real
	else if((p_src->PixelType != ImageComplex || p_src->PixelType != ImageComplexLinear) &&
	   (p_kernel->PixelType == ImageComplex || p_kernel->PixelType == ImageComplexLinear))
		n_path = 4; // real * complex
	else if((p_src->PixelType == ImageComplex || p_src->PixelType == ImageComplexLinear) &&
	   (p_kernel->PixelType == ImageComplex || p_kernel->PixelType == ImageComplexLinear))
		n_path = 5; // complex * complex
	// processing path (0, 1 are optimized, others float/generic)

	bool b_src0_resample = p_dest->XSize != p_src->XSize || p_dest->YSize != p_src->YSize;
	// will we be resampling?

	unsigned int w0 = p_src->XSize, h0 = p_src->YSize;
	// copy sizes here to avoid aliasing

	int n = p_kernel->XSize;
	int m = p_kernel->YSize;
	int n2 = p_kernel->XSize / 2;
	int m2 = p_kernel->YSize / 2;

	for(unsigned int y = 0, h = p_dest->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_dest->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned long n_result = 0;
				for(int dy = 0; dy < m; ++ dy) {
					for(int dx = 0; dx < n; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n2, y0 = dy - m2;
						else
							x0 = ((dx - n2) * w0) / w, y0 = ((dy - m2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						n_result += GetPixel8(p_src, x0, y0) * GetPixel8(p_kernel, dx, dy);
					}
				}
				// 2d convolution

				SetPixel8(p_dest, x, y, (unsigned char)clampb(n_result >> 8));
				// fast gray
			} else if(n_path == 1) {
				unsigned long n_r_sum = 0,
							  n_g_sum = 0,
							  n_b_sum = 0,
							  n_a_sum = 0;
				for(int dy = 0; dy < m; ++ dy) {
					for(int dx = 0; dx < n; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n2, y0 = dy - m2;
						else
							x0 = ((dx - n2) * w0) / w, y0 = ((dy - m2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						ImageRGBValueRGB n_kernel_pix = GetPixelRGB(p_src, dx, dy);
						unsigned char n_kernel_r = (unsigned char)(n_kernel_pix >> 16),
									  n_kernel_g = (unsigned char)(n_kernel_pix >> 8),
									  n_kernel_b = (unsigned char)(n_kernel_pix),
									  n_kernel_a = (unsigned char)(n_kernel_pix >> 24);
						ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, x0, y0);
						unsigned char n_r = (unsigned char)(n_pix0 >> 16),
									  n_g = (unsigned char)(n_pix0 >> 8),
									  n_b = (unsigned char)(n_pix0),
									  n_a = (unsigned char)(n_pix0 >> 24);
						n_r_sum += n_r * n_kernel_r;
						n_g_sum += n_g * n_kernel_g;
						n_b_sum += n_b * n_kernel_b;
						n_a_sum += n_a * n_kernel_a;
					}
				}
				// 2d convolution

				SetPixelRGB(p_dest, x, y, RGBA(clampb(n_r_sum >> 8), clampb(n_g_sum >> 8),
					clampb(n_b_sum >> 8), clampb(n_a_sum >> 8)));
				// fast RGB
			} else if(n_path == 2) { // real * real
				float p_pixd[4] = {0, 0, 0, 0};
				for(int dy = 0; dy < m; ++ dy) {
					for(int dx = 0; dx < n; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n2, y0 = dy - m2;
						else
							x0 = ((dx - n2) * w0) / w, y0 = ((dy - m2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_pix0[4], p_kernel_pix[4];
						GetPixel4f(p_src, x0, y0, p_pix0);
						GetPixel4f(p_kernel, dx, dy, p_kernel_pix);
						p_pixd[0] += p_pix0[0] * p_kernel_pix[0];
						p_pixd[1] += p_pix0[1] * p_kernel_pix[1];
						p_pixd[2] += p_pix0[2] * p_kernel_pix[2];
						p_pixd[3] += p_pix0[3] * p_kernel_pix[3];
					}
				}
				// 2d convolution

				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			} else if(n_path == 3) { // complex * real
				ImageComplexValue t_sum = {0, 0};
				for(int dy = 0; dy < m; ++ dy) {
					for(int dx = 0; dx < n; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n2, y0 = dy - m2;
						else
							x0 = ((dx - n2) * w0) / w, y0 = ((dy - m2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_kernel_pix[4];
						GetPixel4f(p_kernel, dx, dy, p_kernel_pix);
						t_sum = t_CmplxAdd(t_sum,
							t_CmplxMulf(p_kernel_pix[0], GetPixelComplex(p_src, x0, y0)));
					}
				}
				// 2d convolution

				float p_pixd[4] = {t_sum.Re, t_sum.Re, t_sum.Re, t_sum.Im};
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			} else if(n_path == 4) { // real * complex
				ImageComplexValue t_sum = {0, 0};
				for(int dy = 0; dy < m; ++ dy) {
					for(int dx = 0; dx < n; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n2, y0 = dy - m2;
						else
							x0 = ((dx - n2) * w0) / w, y0 = ((dy - m2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						float p_image_pix[4];
						GetPixel4f(p_src, x0, y0, p_image_pix);
						t_sum = t_CmplxAdd(t_sum,
							t_CmplxMulf(p_image_pix[0], GetPixelComplex(p_kernel, dx, dy)));
					}
				}
				// 2d convolution

				float p_pixd[4] = {t_sum.Re, t_sum.Re, t_sum.Re, t_sum.Im};
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			} else /*if(n_path == 5)*/ { // complex * complex
				ImageComplexValue t_sum = {0, 0};
				for(int dy = 0; dy < m; ++ dy) {
					for(int dx = 0; dx < n; ++ dx) {
						int x0, y0;
						if(!b_src0_resample)
							x0 = dx - n2, y0 = dy - m2;
						else
							x0 = ((dx - n2) * w0) / w, y0 = ((dy - m2) * h0) / h;
						// calculate source image coordinates (exact or nearest neighbour, just in case)

						if(x0 < 0)
							x0 = 0;
						else if(x0 >= (int)w0)
							x0 = w0 - 1;
						if(y0 < 0)
							y0 = 0;
						else if(y0 >= (int)h0)
							y0 = h0 - 1;
						// clamp source coordinates

						t_sum = t_CmplxAdd(t_sum, t_CmplxMul(GetPixelComplex(p_src, x0, y0),
							GetPixelComplex(p_kernel, dx, dy)));
					}
				}
				// 2d convolution

				float p_pixd[4] = {t_sum.Re, t_sum.Re, t_sum.Re, t_sum.Im};
				SetPixel4f(p_dest, x, y, p_pixd);
				// generic float
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageGetHistogram(const ImageStruct *p_src, int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image histogram
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool rrImageGetHistogram(const ImageStruct *p_src, int n_histogram_width, unsigned int *p_histogram)
{
	int n_width_log2 = 0;
	for(int j = 1; j < n_histogram_width; j <<= 1)
		++ n_width_log2;
	if(1 << n_width_log2 != n_histogram_width)
		return false;
	// non-power-of two histogram size

	for(int i = 0; i < n_histogram_width; ++ i)
		p_histogram[i] = 0;

	char n_path = 2;
	if(p_src->PixelType == Image8 || p_src->PixelType == Image8Linear)
		n_path = 0;
	else if(p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear)
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	int n_shift8 = abs(n_width_log2 - 8);
	// shift to 8 bits

	for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_gray = GetPixel8(p_src, x, y);
				// fast gray

				if(n_width_log2 < 8)
					++ p_histogram[n_gray >> n_shift8];
				else if(n_width_log2 == 8)
					++ p_histogram[n_gray];
				if(n_width_log2 > 8)
					++ p_histogram[n_gray << n_shift8];
				// increment histogram
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix = GetPixelRGB(p_src, x, y);
				unsigned char n_r = (unsigned char)(n_pix >> 16),
							  n_g = (unsigned char)(n_pix >> 8),
							  n_b = (unsigned char)(n_pix);
				// fast RGB

				if(n_width_log2 < 8) {
					++ p_histogram[n_r >> n_shift8];
					++ p_histogram[n_g >> n_shift8];
					++ p_histogram[n_b >> n_shift8];
				} else if(n_width_log2 == 8) {
					++ p_histogram[n_r];
					++ p_histogram[n_g];
					++ p_histogram[n_b];
				} if(n_width_log2 > 8) {
					++ p_histogram[n_r << n_shift8];
					++ p_histogram[n_g << n_shift8];
					++ p_histogram[n_b << n_shift8];
				}
				// increment histogram
			} else {
				float p_pix[4];
				GetPixel4f(p_src, x, y, p_pix);
				// generic float

				++ p_histogram[(int)(minf(1, maxf(0, p_pix[0])) * n_histogram_width)];
				++ p_histogram[(int)(minf(1, maxf(0, p_pix[1])) * n_histogram_width)];
				++ p_histogram[(int)(minf(1, maxf(0, p_pix[2])) * n_histogram_width)];
				// increment histogram
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageGetLuminosityHistogram(const ImageStruct *p_src,
 *		int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image luminosity histogram (greyscale conversion applied)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool rrImageGetLuminosityHistogram(const ImageStruct *p_src,
	int n_histogram_width, unsigned int *p_histogram)
{
	int n_width_log2 = 0;
	for(int j = 1; j < n_histogram_width; j <<= 1)
		++ n_width_log2;
	if(1 << n_width_log2 != n_histogram_width)
		return false;
	// non-power-of two histogram size

	for(int i = 0; i < n_histogram_width; ++ i)
		p_histogram[i] = 0;

	char n_path = 2;
	if(p_src->PixelType == Image8 || p_src->PixelType == Image8Linear)
		n_path = 0;
	else if(p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear)
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	int n_shift8 = abs(n_width_log2 - 8);
	// shift to 8 bits

	for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_gray = GetPixel8(p_src, x, y);
				// fast gray

				if(n_width_log2 < 8)
					++ p_histogram[n_gray >> n_shift8];
				else if(n_width_log2 == 8)
					++ p_histogram[n_gray];
				if(n_width_log2 > 8)
					++ p_histogram[n_gray << n_shift8];
				// increment histogram
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix = GetPixelRGB(p_src, x, y);
				unsigned char n_r = (unsigned char)(n_pix >> 16),
							  n_g = (unsigned char)(n_pix >> 8),
							  n_b = (unsigned char)(n_pix);
				// fast RGB

				const unsigned short n_red_weight = (unsigned short)(.299f * 0x10000);
				const unsigned short n_green_weight = (unsigned short)(.587f * 0x10000);
				const unsigned short n_blue_weight = (unsigned short)(.114f * 0x10000);
				// 0:16 fixed-point PAL weights

				unsigned char n_i = (unsigned char)((n_red_weight * (unsigned int)n_r +
					n_green_weight * (unsigned int)n_g +
					n_blue_weight * (unsigned int)n_b + 0x8000) >> 16);
				// calc intensity

				if(n_width_log2 < 8)
					++ p_histogram[n_i >> n_shift8];
				else if(n_width_log2 == 8)
					++ p_histogram[n_i];
				if(n_width_log2 > 8)
					++ p_histogram[n_i << n_shift8];
				// increment histogram
			} else {
				float p_pix[4];
				GetPixel4f(p_src, x, y, p_pix);
				// generic float

				float f_grey = .299f * p_pix[0] + .587f * p_pix[1] + .114f * p_pix[2];

				++ p_histogram[(int)(minf(1, maxf(0, f_grey)) * n_histogram_width)];
				// increment histogram
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageGetRGBHistogram(const ImageStruct *p_src, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b)
 *		- retrieves RGB image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g and p_histo_b will contain histogram data, they doen't have
 *		  to be zeroed prior calling this
 */
bool rrImageGetRGBHistogram(const ImageStruct *p_src, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b)
{
	int n_width_log2 = 0;
	for(int j = 1; j < n_histogram_width; j <<= 1)
		++ n_width_log2;
	if(1 << n_width_log2 != n_histogram_width)
		return false;
	// non-power-of two histogram size

	for(int i = 0; i < n_histogram_width; ++ i) {
		p_histo_r[i] = 0;
		p_histo_g[i] = 0;
		p_histo_b[i] = 0;
	}

	char n_path = 2;
	if(p_src->PixelType == Image8 || p_src->PixelType == Image8Linear)
		n_path = 0;
	else if(p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear)
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	int n_shift8 = abs(n_width_log2 - 8);
	// shift to 8 bits

	for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_gray = GetPixel8(p_src, x, y);
				// fast gray

				if(n_width_log2 < 8) {
					int n_index = n_gray >> n_shift8;
					++ p_histo_r[n_index];
					++ p_histo_g[n_index];
					++ p_histo_b[n_index];
				} else if(n_width_log2 == 8) {
					++ p_histo_r[n_gray];
					++ p_histo_g[n_gray];
					++ p_histo_b[n_gray];
				} if(n_width_log2 > 8) {
					int n_index = n_gray << n_shift8;
					++ p_histo_r[n_index];
					++ p_histo_g[n_index];
					++ p_histo_b[n_index];
				}
				// increment histogram
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix = GetPixelRGB(p_src, x, y);
				unsigned char n_r = (unsigned char)(n_pix >> 16),
							  n_g = (unsigned char)(n_pix >> 8),
							  n_b = (unsigned char)(n_pix);
				// fast RGB

				if(n_width_log2 < 8) {
					++ p_histo_r[n_r >> n_shift8];
					++ p_histo_g[n_g >> n_shift8];
					++ p_histo_b[n_b >> n_shift8];
				} else if(n_width_log2 == 8) {
					++ p_histo_r[n_r];
					++ p_histo_g[n_g];
					++ p_histo_b[n_b];
				} if(n_width_log2 > 8) {
					++ p_histo_r[n_r << n_shift8];
					++ p_histo_g[n_g << n_shift8];
					++ p_histo_b[n_b << n_shift8];
				}
				// increment histogram
			} else {
				float p_pix[4];
				GetPixel4f(p_src, x, y, p_pix);
				// generic float

				++ p_histo_r[(int)(minf(1, maxf(0, p_pix[0])) * n_histogram_width)];
				++ p_histo_g[(int)(minf(1, maxf(0, p_pix[1])) * n_histogram_width)];
				++ p_histo_b[(int)(minf(1, maxf(0, p_pix[2])) * n_histogram_width)];
				// increment histogram
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageGetRGBAHistogram(const ImageStruct *p_src, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g,
 *		unsigned int *p_histo_b, unsigned int *p_histo_a)
 *		- retrieves RGBA image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g, p_histo_b and p_histo_a will contain histogram data, they
 *		  doen't have to be zeroed prior calling this
 */
bool rrImageGetRGBAHistogram(const ImageStruct *p_src, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g,
	unsigned int *p_histo_b, unsigned int *p_histo_a)
{
	int n_width_log2 = 0;
	for(int j = 1; j < n_histogram_width; j <<= 1)
		++ n_width_log2;
	if(1 << n_width_log2 != n_histogram_width)
		return false;
	// non-power-of two histogram size

	for(int i = 0; i < n_histogram_width; ++ i) {
		p_histo_r[i] = 0;
		p_histo_g[i] = 0;
		p_histo_b[i] = 0;
		p_histo_a[i] = 0;
	}

	char n_path = 2;
	if(p_src->PixelType == Image8 || p_src->PixelType == Image8Linear) {
		n_path = 0;
		p_histo_a[n_histogram_width - 1] = p_src->XSize * p_src->YSize; // grey has all alpha 0xff
	} else if(p_src->PixelType == ImageRGB || p_src->PixelType == ImageRGBLinear)
		n_path = 1;
	// processing path (0, 1 are optimized, 2 is float/generic)

	int n_shift8 = abs(n_width_log2 - 8);
	// shift to 8 bits

	for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
		for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
			if(!n_path) {
				unsigned char n_gray = GetPixel8(p_src, x, y);
				// fast gray

				if(n_width_log2 < 8) {
					int n_index = n_gray >> n_shift8;
					++ p_histo_r[n_index];
					++ p_histo_g[n_index];
					++ p_histo_b[n_index];
				} else if(n_width_log2 == 8) {
					++ p_histo_r[n_gray];
					++ p_histo_g[n_gray];
					++ p_histo_b[n_gray];
				} if(n_width_log2 > 8) {
					int n_index = n_gray << n_shift8;
					++ p_histo_r[n_index];
					++ p_histo_g[n_index];
					++ p_histo_b[n_index];
				}
				// increment histogram
			} else if(n_path == 1) {
				ImageRGBValueRGB n_pix = GetPixelRGB(p_src, x, y);
				unsigned char n_r = (unsigned char)(n_pix >> 16),
							  n_g = (unsigned char)(n_pix >> 8),
							  n_b = (unsigned char)(n_pix),
							  n_a = (unsigned char)(n_pix >> 24);
				// fast RGB

				if(n_width_log2 < 8) {
					++ p_histo_r[n_r >> n_shift8];
					++ p_histo_g[n_g >> n_shift8];
					++ p_histo_b[n_b >> n_shift8];
					++ p_histo_a[n_a >> n_shift8];
				} else if(n_width_log2 == 8) {
					++ p_histo_r[n_r];
					++ p_histo_g[n_g];
					++ p_histo_b[n_b];
					++ p_histo_a[n_a];
				} if(n_width_log2 > 8) {
					++ p_histo_r[n_r << n_shift8];
					++ p_histo_g[n_g << n_shift8];
					++ p_histo_b[n_b << n_shift8];
					++ p_histo_a[n_a << n_shift8];
				}
				// increment histogram
			} else {
				float p_pix[4];
				GetPixel4f(p_src, x, y, p_pix);
				// generic float

				++ p_histo_r[(int)(minf(1, maxf(0, p_pix[0])) * n_histogram_width)];
				++ p_histo_g[(int)(minf(1, maxf(0, p_pix[1])) * n_histogram_width)];
				++ p_histo_b[(int)(minf(1, maxf(0, p_pix[2])) * n_histogram_width)];
				++ p_histo_a[(int)(minf(1, maxf(0, p_pix[3])) * n_histogram_width)];
				// increment histogram
			}
			// main processing
		}
	}
	// when optimized for speed, compiler should shift conditions down and cycles up

	return true;
}

/**
 *	bool rrImageGetMinMax(const ImageStruct *p_src, void *p_minmax)
 *		- get image minimal and maximal values
 *		- values are written to p_minmax in image pixel-type format in order min, then max
 */
bool rrImageGetMinMax(const ImageStruct *p_src, void *p_minmax)
{
	switch(p_src->PixelType) {
	case Image8:
	case Image8Linear:
		{
			unsigned char *p_min = (unsigned char*)p_minmax;
			unsigned char *p_max = (unsigned char*)p_minmax + 1;

			*p_min = *p_max = GetPixel8(p_src, 0, 0);

			for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
				for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
					unsigned char n_gray = GetPixel8(p_src, x, y);
					if(n_gray > *p_max)
						*p_max = n_gray;
					if(n_gray < *p_min)
						*p_min = n_gray;
				}
			}

			return true;
		}
		break;
	case Image16:
	case Image16Linear:
		{
			Image16Value *p_min = (Image16Value*)p_minmax;
			Image16Value *p_max = (Image16Value*)p_minmax + 1;

			*p_min = *p_max = GetPixel16(p_src, 0, 0);

			for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
				for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
					Image16Value n_gray = GetPixel16(p_src, x, y);
					if(n_gray > *p_max)
						*p_max = n_gray;
					if(n_gray < *p_min)
						*p_min = n_gray;
				}
			}

			return true;
		}
		break;
	case Image32:
	case Image32Linear:
		{
			Image32Value *p_min = (Image32Value*)p_minmax;
			Image32Value *p_max = (Image32Value*)p_minmax + 1;

			*p_min = *p_max = GetPixel32(p_src, 0, 0);

			for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
				for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
					Image32Value n_gray = GetPixel32(p_src, x, y);
					if(n_gray > *p_max)
						*p_max = n_gray;
					if(n_gray < *p_min)
						*p_min = n_gray;
				}
			}

			return true;
		}
		break;
	case ImageFloat:
	case ImageFloatLinear:
		{
			float *p_min = (float*)p_minmax;
			float *p_max = (float*)p_minmax + 1;

			*p_min = *p_max = GetPixelFloat(p_src, 0, 0);

			for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
				for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
					float f_gray = GetPixelFloat(p_src, x, y);
					if(f_gray > *p_max)
						*p_max = f_gray;
					if(f_gray < *p_min)
						*p_min = f_gray;
				}
			}

			return true;
		}
		break;
	case ImageComplex:
	case ImageComplexLinear:
		{
			ImageComplexValue *p_min = (ImageComplexValue*)p_minmax;
			ImageComplexValue *p_max = (ImageComplexValue*)p_minmax + 1;

			*p_min = *p_max = GetPixelComplex(p_src, 0, 0);

			for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
				for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
					ImageComplexValue f_cmplx = GetPixelComplex(p_src, x, y);
					if(f_cmplx.Im > p_max->Im)
						p_max->Im = f_cmplx.Im;
					if(f_cmplx.Re > p_max->Re)
						p_max->Re = f_cmplx.Re;
					if(f_cmplx.Im < p_min->Im)
						p_min->Im = f_cmplx.Im;
					if(f_cmplx.Re < p_min->Re)
						p_min->Re = f_cmplx.Re;
				}
			}

			return true;
		}
		break;
	case ImageRGB:
	case ImageRGBLinear:
		{
			ImageRGBValueRGB *p_min = (ImageRGBValueRGB*)p_minmax;
			ImageRGBValueRGB *p_max = (ImageRGBValueRGB*)p_minmax + 1;

			ImageRGBValueRGB n_pix0 = GetPixelRGB(p_src, 0, 0);

			unsigned char n_min_r = (unsigned char)(n_pix0 >> 16),
						  n_min_g = (unsigned char)(n_pix0 >> 8),
						  n_min_b = (unsigned char)(n_pix0),
						  n_min_a = (unsigned char)(n_pix0 >> 24);
			unsigned char n_max_r = n_min_r,
						  n_max_g = n_min_g,
						  n_max_b = n_min_b,
						  n_max_a = n_min_a;
			// min/max values of first pixel

			for(unsigned int y = 0, h = p_src->YSize; y < h; ++ y) {
				for(unsigned int x = 0, w = p_src->XSize; x < w; ++ x) {
					ImageRGBValueRGB n_pix = GetPixelRGB(p_src, x, y);
					unsigned char n_r = (unsigned char)(n_pix >> 16),
								  n_g = (unsigned char)(n_pix >> 8),
								  n_b = (unsigned char)(n_pix),
								  n_a = (unsigned char)(n_pix >> 24);
					if(n_min_r > n_r)
						n_min_r = n_r;
					if(n_min_g > n_g)
						n_min_g = n_g;
					if(n_min_b > n_b)
						n_min_b = n_b;
					if(n_min_a > n_a)
						n_min_a = n_a;
					if(n_max_r < n_r)
						n_max_r = n_r;
					if(n_max_g < n_g)
						n_max_g = n_g;
					if(n_max_b < n_b)
						n_max_b = n_b;
					if(n_max_a < n_a)
						n_max_a = n_a;
				}
			}

			*p_min = RGBA(n_min_r, n_min_g, n_min_b, n_min_a);
			*p_max = RGBA(n_max_r, n_max_g, n_max_b, n_max_a);
			// save values

			return true;
		}
		break;
	}

	return false;
}

/*
 *		-end-of-file-
 */
